/* Translate a CNF format file to LISP format */
/* Compile: gcc -o cnf2lisp cnf2lisp.c */

/* Written by Fco. Jesus Martin (fjesus@cica.es)
   based on previous version by Tamas Badics (badics@rutcor.rutgers.edu)
   Last modification: Mon Oct 15 17:20:55 CEST 2001

Usage: ./cnf2lisp in_file
Description: 
        Translates from CNF format to LISP format.
	The "in-file" argument is the name of CNF file whitout extension.
	The extension ".cnf" is assumed for CNF file.
	It uses the same name for CNF file and LISP file.
        By default it uses the stdin and stdout.
*/
/*********************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void cnf2sat(FILE * input, FILE * output, char * name);

int main(int argc, char * argv[])
{
  int i;
  FILE * output, * input;
  char * tmp;

  tmp = (char *)calloc(100, sizeof(char));
  
  output = stdout;
  input = stdin;

  if (argc = 1) {
    sprintf(tmp,"%s.cnf",argv[1]);
    input = fopen(tmp,"r");
    if (input == NULL) {
      fprintf(stderr
	      ,"%s: Input file %s.cnf can't be opened\n",argv[0],argv[i]);
      exit(-1);
    }	
    sprintf(tmp,"%s.lisp",argv[1]);
    output = fopen(tmp,"w");
    if (output == NULL) {
      fprintf(stderr
	      ,"%s: Output file %s.lisp can't be opened\n",argv[0],argv[i]);
      exit(-1);
    }
  } else {
    if (output != stdout) fclose(output);
    if (input != stdin) fclose(input);
    
    printf("Usage: %s in_file", argv[0]);
    printf("Description:\n\tTranslates from CNF format to LISP format.\n");
    printf("\tThe <in-file> argument is the name of CNF file whitout extension.\n"); 
    printf("\tThe extension \".cnf\" is assumed for CNF file.\n");
    printf("\tIt uses the same name for CNF file and LISP file.\n");
    printf("\tBy default it uses the stdin and stdout.\n");

    exit(0);
  }
   
  cnf2sat(input, output, argv[1]);
  
  if (output != stdout) fclose(output);
  if (input != stdin) fclose(input);
  exit(0);
  
}

/*===========================================================================*/

void cnf2sat(FILE * f, FILE * output, char * name)
{
  char c;
  char * tmp;
  int i, d1, d2, new;
  
  tmp = (char *)calloc(100, sizeof(char));
  new = 1;

  while ((c = fgetc(f)) != EOF) {
    switch (c) {
    case '\n':
    case ' ':
    case '\t':
      fputc(' ',output);
      while ((c = fgetc(f)) == ' ' || c == '\t'|| c == '\n');
      ungetc(c, f);
      break;
      
    case 'c':
      fprintf(output, ";;; %c", c);
      while ((c = fgetc(f)) != '\n' && c != EOF)
	fputc(c,output);
      
      fputc('\n',output);
      if (c == EOF) {
	ungetc(c, f);
	ungetc(' ', f);
	ungetc('c', f);
      }
      break;
      
    case 'p':
      fprintf(output, ";;; %c", c);
      fscanf(f, "%s %d %d\n", tmp, &d1, &d2);
      fprintf(output, " %s %d %d\n", tmp, d1, d2);
      fprintf(output, "(defconst *%s*\n", name);
      fprintf(output, "  '( ");
      break;
      
    case '0':
      fprintf(output, ")\n    ");
      new = 1;
      break;
      
    default:
      if (new) {
	fprintf(output,"( ");
	new = 0;
      }
      fputc(c,output);
      while ((c = fgetc(f)) != ' ' && c != '\t' && c != '\n') {
	fputc(c,output);
      }
      fputc(' ',output);
      break;
    }
  }
  fprintf(output, " )\n  )\n");
  
}

/*===========================================================================*/

