;;;============================================================================
;;;
;;; Ttulo: Frmulas de Pelletier
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../0-ejemplos/pelletier")

|#

(in-package "ACL2")

;;;============================================================================
;;;
;;; 1) Frmulas de Pelletier
;;;
;;;============================================================================

;;; 1)
;;;   $(p \imp q) \sii (\no q \imp \no p)$
;;; 2)
;;;   $(\no(\no p) \sii p)$
;;; 3)
;;;   $\no(p \imp q) \imp (q \imp p)$
;;; 4)
;;;   $(\no p \imp q) \sii (\no q \imp p)$
;;; 5)
;;;   $((p \dis q) \imp (p \dis r)) \imp (p \dis (q \imp r))$
;;; 6)
;;;   $(p \dis \no p)$
;;; 7)
;;;   $(p \dis \no(\no(\no p)))$
;;; 8)
;;;   $((p \imp q) \imp p) \imp p$
;;; 9)
;;;   $[(p \dis q) \con (\no p \dis q) \con (p \dis \no q)]
;;;    \imp \no(\no p \dis \no q)$
;;; 10)
;;;   $[(q \imp r) \con (r \imp (p \con q)) \con (p \imp (q \dis r))]
;;;    \imp (p \sii q)$
;;; 11)
;;;   $(p \sii p)$
;;; 12)
;;;   $[(p \sii q) \sii r] \sii [p \sii (q \sii r)]$
;;; 13)
;;;   $[p \dis (q \con r)] \sii [(p \dis q) \con (p \dis r)]$
;;; 14)
;;;   $(p \sii q) \sii ((q \dis \no p) \con (\no q \dis p)) $
;;; 15)
;;;   $(p \imp q) \sii (\no p \dis q)$
;;; 16)
;;;   $(p \imp q) \dis (q \imp p)$
;;; 17)
;;;   $((p \con (q \imp r)) \imp s) \sii
;;;    ((\no p \dis q \dis s) \dis (\no p \dis \no r \dis s))$
;;; -)

(defconst *pelletier-1*
  '(<-> (-> p q) (-> (- q) (- p))))

(defconst *pelletier-2*
  '(<-> (- (- p)) p))

(defconst *pelletier-3*
  '(-> (- (-> p q)) (-> q p)))

(defconst *pelletier-4*
  '(<-> (-> (- p) q) (-> (- q) p)))

(defconst *pelletier-5*
  '(-> (-> (\/ p q) (\/ p r)) (\/ p (-> q r))))

(defconst *pelletier-6*
  '(\/ p (- p)))

(defconst *pelletier-7*
  '(\/ p (- (- (- p)))))

(defconst *pelletier-8*
  '(-> (-> (-> p q) p) p))

(defconst *pelletier-9*
  '(-> (& (& (\/ p q) (\/ (- p) q)) (\/ p (- q)))
       (- (\/ (- p) (- q)))))

(defconst *pelletier-10*
  '(-> (& (& (-> q r) (-> r (& p q))) (-> p (\/ q r))) (<-> p q)))

(defconst *pelletier-11*
  '(<-> p p))

(defconst *pelletier-12*
  '(<-> (<-> (<-> p q) r) (<-> p (<-> q r))))

(defconst *pelletier-13*
  '(<-> (\/ p (& q r)) (& (\/ p q) (\/ p r))))

(defconst *pelletier-14*
  '(<-> (<-> p q) (& (\/ q (- p)) (\/ (- q) p))))

(defconst *pelletier-15*
  '(<-> (-> p q) (\/ (- p) q)))

(defconst *pelletier-16*
  '(\/ (-> p q) (-> q p)))

(defconst *pelletier-17*
  '(<-> (-> (& p (-> q r)) s) (& (\/ (\/ (- p) q) s)
				 (\/ (\/ (- p) (- r)) s))))

(defconst *pelletier*
  (list
   (cons " 1) (p -> q) <-> (-q -> -p)" *pelletier-1*)
   (cons " 2) (-(-p) <-> p)" *pelletier-2*)
   (cons " 3) -(p -> q) -> (q -> p)" *pelletier-3*)
   (cons " 4) (-p -> q) <-> (-q -> p)" *pelletier-4*)
   (cons " 5) ((p | q) -> (p | r)) -> (p | (q -> r))" *pelletier-5*)
   (cons " 6) (p | -p)" *pelletier-6*)
   (cons " 7) (p | -(-(-p)))" *pelletier-7*)
   (cons " 8) ((p -> q) -> p) -> p" *pelletier-8*)
   (cons " 9) [(p | q) & (-p | q) & (p | -q)] -> -(-p | -q)" *pelletier-9*)
   (cons "10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q)"
    *pelletier-10*)
   (cons "11) (p <-> p)" *pelletier-11*)
   (cons "12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]" *pelletier-12*)
   (cons "13) [p | (q & r)] <-> [(p | q) & (p | r)]" *pelletier-13*)
   (cons "14) (p <-> q) <-> ((q | -p) & (-q | p))" *pelletier-14*)
   (cons "15) (p -> q) <-> (-p | q)" *pelletier-15*)
   (cons "16) (p -> q) | (q -> p)" *pelletier-16*)
   (cons "17) ((p & (q -> r)) -> s) <-> ((-p | q | s) & (-p | -r | s))"
    *pelletier-17*)))

;;;============================================================================