;;;============================================================================
;;;
;;; Ttulo: Problema de las N reinas
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../0-ejemplos/reinas")

|#

(in-package "ACL2")

(include-book "../3-logica/sintaxis")

;;;============================================================================
;;;
;;; 1) N reinas
;;;
;;;============================================================================

;;;   Esta funcin obtiene la variable correspondiente a la casilla (i,j) con
;;; 0<= i,j < N. Dicha variable es el valor numrico i*N+j+1

(defun casilla (i j n)
  (+ (* i n) j 1))

;;;   Lista de casillas en la misma fila

(defun casillas-fila-aux (k i j n)
  (declare (xargs :mode :program))
  (cond ((= k j) (casillas-fila-aux (+ k 1) i j n))
	((< k n) (cons (casilla i k n)
		       (casillas-fila-aux (+ k 1) i j n)))
	(t nil)))

(defun casillas-fila (i j n)
  (declare (xargs :mode :program))
  (casillas-fila-aux 0 i j n))

;;;   Lista de casillas en la misma columna

(defun casillas-columna-aux (k i j n)
  (declare (xargs :mode :program))
  (cond ((= k i) (casillas-columna-aux (+ k 1) i j n))
	((< k n) (cons (casilla k j n)
		       (casillas-columna-aux (+ k 1) i j n)))
	(t nil)))

(defun casillas-columna (i j n)
  (declare (xargs :mode :program))
  (casillas-columna-aux 0 i j n))

;;;   Ejemplos
;;; (casilla 0 1 2 5)
;;;

;;;   Lista de casillas en diagonal principal

(defun casillas-diagonal-p-aux (k i j n)
  (declare (xargs :mode :program))
  (cond ((= k 0) (casillas-diagonal-p-aux (+ k 1) i j n))
	((< k (min (- n i) (- n j)))
	 (cons (casilla (+ i k) (+ j k) n)
	       (casillas-diagonal-p-aux (+ k 1) i j n)))
	(t nil)))
	 
(defun casillas-diagonal-p (i j n)
  (declare (xargs :mode :program))
  (casillas-diagonal-p-aux (max (- 0 i) (- 0 j)) i j n))

;;;   Lista de casillas en diagonal secundaria

(defun casillas-diagonal-s-aux (k i j n)
  (declare (xargs :mode :program))
  (cond ((= k 0) (casillas-diagonal-s-aux (+ k 1) i j n))
	((< k (min (- n i) (+ j 1)))
	 (cons (casilla (+ i k) (- j k) n)
	       (casillas-diagonal-s-aux (+ k 1) i j n)))
	(t nil)))

(defun casillas-diagonal-s (i j n)
  (declare (xargs :mode :program))
  (casillas-diagonal-s-aux (max (- i) (+ 1 (- j n))) i j n))

;;;   La funcin no-hay-amenaza recibe como argumentos una fila I y una columna
;;; J y construye la frmula proposicional que representa el hecho de que si
;;; hay una reina en la casilla (I,J), entonces no debe haber ninguna otra
;;; reina en la fila I, en la columna J ni en las diagonales que se cruzan en
;;; la casilla (I,J).

(defun conjuncion-negaciones-lista (lista)
  (declare (xargs :mode :program))
  (cond ((endp (cdr lista)) (negacion (car lista)))
	(t (conjuncion (negacion (car lista))
		       (conjuncion-negaciones-lista (cdr lista))))))

(defun no-hay-amenaza (i j n)
  (declare (xargs :mode :program))
  (cond ((= n 1) nil)
	(t (implicacion (casilla i j n)
			(conjuncion-negaciones-lista
			 (append (casillas-fila i j n)
				 (casillas-columna i j n)
				 (casillas-diagonal-p i j n)
				 (casillas-diagonal-s i j n)))))))

;;;   La funcin REINA-EN-FILA recibe como argumento una fila y construye la
;;; frmula proposicional que representa el hecho de que en la fila I debe
;;; haber al menos una reina.

(defun disyuncion-lista (lista)
  (declare (xargs :mode :program))
  (cond ((endp (cdr lista)) (car lista))
	(t (disyuncion (car lista)
		       (disyuncion-lista (cdr lista))))))

(defun fila-completa-aux (k i n)
  (declare (xargs :mode :program))
  (cond ((< k n) (cons (casilla i k n)
		       (fila-completa-aux (+ k 1) i n)))
	(t nil)))

(defun fila-completa (i n)
  (declare (xargs :mode :program))
  (fila-completa-aux 0 i n))

(defun reina-en-fila (i n)
  (declare (xargs :mode :program))
  (disyuncion-lista (fila-completa i n)))

;;;   La funcin reinas-en-filas construye la frmula que afirma que en cada
;;; fila hay al menos una reina:

(defun conjuncion-lista (lista)
  (declare (xargs :mode :program))
  (cond ((endp (cdr lista)) (car lista))
	(t (conjuncion (car lista)
		       (conjuncion-lista (cdr lista))))))

(defun lista-reinas-en-filas (k n)
  (declare (xargs :mode :program))
  (cond ((< k n) (cons (reina-en-fila k n)
		       (lista-reinas-en-filas (+ k 1) n)))
	(t nil)))

(defun lista-no-hay-amenazas-fila (k i n)
  (declare (xargs :mode :program))
  (cond ((< k n) (cons (no-hay-amenaza i k n)
		       (lista-no-hay-amenazas-fila (+ k 1) i n)))
	(t nil)))

(defun lista-no-hay-amenazas-columna (k n)
  (declare (xargs :mode :program))
  (cond ((< k n) (append (lista-no-hay-amenazas-fila 0 k n)
			 (lista-no-hay-amenazas-columna (+ k 1) n)))
	(t nil)))

(defun lista-no-hay-amenazas (n)
  (declare (xargs :mode :program))
  (lista-no-hay-amenazas-columna 0 n))

(defun n-reinas (n)
  (declare (xargs :mode :program))
  (conjuncion-lista (append (lista-reinas-en-filas 0 n)
			    (lista-no-hay-amenazas n))))

(defconst *n-reinas*
  (list (cons "Reinas-2" (n-reinas 2))
	(cons "Reinas-3" (n-reinas 3))
	(cons "Reinas-4" (n-reinas 4))
	(cons "Reinas-5" (n-reinas 5))
	(cons "Reinas-6" (n-reinas 6))
	(cons "Reinas-7" (n-reinas 7))
	(cons "Reinas-8" (n-reinas 8))))

;;;============================================================================