;;;============================================================================
;;;
;;; Ttulo: Notacin uniforme
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../4-tableros/uniforme")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../3-logica/semantica")

(in-theory (disable definiciones-conectivas
		    otra-definicion-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) Clasificacin basada en la notacin uniforme
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Alfa-frmulas y Beta-frmulas
;;;
;;;----------------------------------------------------------------------------

(defun alfa-formula (F)
  (or (es-conjuncion F)
      (and (es-negacion F)
	   (or (es-disyuncion (arg1 F))
	       (es-implicacion (arg1 F))))))

(defun beta-formula (F)
  (or (es-disyuncion F)
      (es-implicacion F)
      (es-equivalencia F)
      (and (es-negacion F)
	   (or (es-conjuncion (arg1 F))
	       (es-equivalencia (arg1 F))))))

(defun componente-1 (F)
  (cond ((es-negacion F)
	 (let ((F-1 (arg1 F)))
	   (cond ((es-conjuncion F-1)
		  (negacion (arg1 F-1)))
		 ((es-disyuncion F-1)
		  (negacion (arg1 F-1)))
		 ((es-implicacion F-1)
		  (arg1 F-1))
		 ((es-equivalencia F-1)
		  (conjuncion (arg1 F-1) (negacion (arg2 F-1))))
		 (t F))))
	((es-conjuncion F)
	 (arg1 F))
	((es-disyuncion F)
	 (arg1 F))
	((es-implicacion F)
	 (negacion (arg1 F)))
	((es-equivalencia F)
	 (conjuncion (arg1 F) (arg2 F)))
	(t F)))
	 
(defun componente-2 (F)
  (cond ((es-negacion F)
	 (let ((F-1 (arg1 F)))
	   (cond ((es-conjuncion F-1)
		  (negacion (arg2 F-1)))
		 ((es-disyuncion F-1)
		  (negacion (arg2 F-1)))
		 ((es-implicacion F-1)
		  (negacion (arg2 F-1)))
		 ((es-equivalencia F-1)
		  (conjuncion (negacion (arg1 F-1)) (arg2 F-1)))
		 (t F))))
	((es-conjuncion F)
	 (arg2 F))
	((es-disyuncion F)
	 (arg2 F))
	((es-implicacion F)
	 (arg2 F))
	((es-equivalencia F)
	 (conjuncion (negacion (arg1 F)) (negacion (arg2 F))))
	(t F)))

(defthm valor-componentes-alfa-formula
  (implies (alfa-formula F)
	   (equal (valor F sigma)
		  (funcion-de-verdad-de-conjuncion
		   (valor (componente-1 F) sigma)
		   (valor (componente-2 F) sigma)))))

(defthm valor-componentes-beta-formula
  (implies (beta-formula F)
	   (equal (valor F sigma)
		  (funcion-de-verdad-de-disyuncion
		   (valor (componente-1 F) sigma)
		   (valor (componente-2 F) sigma)))))

(defthm componentes-es-proposicional
  (implies (or (alfa-formula F)
	       (beta-formula F))
	   (and (es-proposicional (componente-1 F))
		(es-proposicional (componente-2 F)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Dobles negaciones
;;;
;;;----------------------------------------------------------------------------

(defun doble-negacion (F)
  (and (es-negacion F)
       (es-negacion (arg1 F))))

(defun componente-neg-neg (F)
  (if (doble-negacion F)
      (arg1 (arg1 F))
      F))

(defthm valor-componentes-doble-negacion
  (implies (doble-negacion F)
	   (equal (valor F sigma)
		  (valor (componente-neg-neg F) sigma))))

(defthm componente-neg-neg-es-proposicional
  (implies (doble-negacion F)
	   (es-proposicional (componente-neg-neg F))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Literales
;;;
;;;----------------------------------------------------------------------------

(defun es-literal-positivo (F)
  (es-atomica F))

(defun es-literal-negativo (F)
  (and (es-negacion F)
       (es-atomica (arg1 F))))

(defun es-literal (F)
  (or (es-literal-positivo F)
      (es-literal-negativo F)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Clasificacin basada en la notacin uniforme
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 ()

 (local 
  (defthm definicion-uniforme-de-es-proposicional-aux
    (implies (and (es-negacion F)
		  (not (es-negacion (arg1 F)))
		  (not (es-disyuncion (arg1 F)))
		  (not (es-implicacion (arg1 F)))
		  (not (es-conjuncion (arg1 F)))
		  (not (es-equivalencia (arg1 F))))
	     (es-atomica (arg1 F)))
    :hints (("Goal" 
	     :in-theory (disable es-atomica otra-definicion-de-es-proposicional
				 es-proposicional)))))
 
 (local
  (in-theory (enable otra-definicion-de-es-proposicional)))

 (defthm definicion-uniforme-de-es-proposicional
   (iff (es-proposicional F)
	(or (doble-negacion F)
	    (alfa-formula F)
	    (beta-formula F)
	    (es-literal F)))
   :hints (("Goal" :in-theory (disable es-atomica)))
   :rule-classes :definition)

)

;;;============================================================================
;;;
;;; 2) Medida asociada a la notacin uniforme
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Medida uniforme
;;;
;;;----------------------------------------------------------------------------

(defun medida-uniforme (F)
  (declare (xargs :measure (numero-conectivas F)))
  (cond ((es-equivalencia F)
	 (+ 5 (medida-uniforme (arg1 F))
	      (medida-uniforme (arg2 F))))
 	((or (es-conjuncion F)
 	     (es-disyuncion F)
 	     (es-implicacion F))
 	 (+ 2 (medida-uniforme (arg1 F))
 	      (medida-uniforme (arg2 F))))
 	((es-negacion F)
 	 (+ 1 (medida-uniforme (arg1 F))))
	(t 0)))

(defthm medida-uniforme-conjuncion-lineal
  (implies (es-conjuncion F)
	   (= (+ 2 (medida-uniforme (arg1 F))
		 (medida-uniforme (arg2 F)))
	      (medida-uniforme F)))
  :rule-classes :linear)

(defthm medida-uniforme-disyuncion-lineal
  (implies (es-disyuncion F)
	   (= (+ 2 (medida-uniforme (arg1 F))
		 (medida-uniforme (arg2 F)))
	      (medida-uniforme F)))
  :rule-classes :linear)

(defthm medida-uniforme-implicacion-lineal
  (implies (es-implicacion F)
	   (= (+ 2 (medida-uniforme (arg1 F))
		 (medida-uniforme (arg2 F)))
	      (medida-uniforme F)))
  :rule-classes :linear)

(defthm medida-uniforme-equivalencia-lineal
  (implies (es-equivalencia F)
	   (= (+ 5 (medida-uniforme (arg1 F))
		 (medida-uniforme (arg2 F)))
	      (medida-uniforme F)))
  :rule-classes :linear)

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) La medida uniforme disminuye
;;;
;;;----------------------------------------------------------------------------

(defthm medida-uniforme-componente-neg-neg-disminuye
  (implies (doble-negacion F)
	   (< (medida-uniforme (componente-neg-neg F))
	      (medida-uniforme F)))
  :rule-classes :linear)

(defthm medida-uniforme-componentes-alfa-disminuye
  (implies (alfa-formula F)
	   (and (< (medida-uniforme (componente-1 F))
		   (medida-uniforme F))
		(< (medida-uniforme (componente-2 F))
		   (medida-uniforme F))))
  :rule-classes :linear
  :hints (("Goal"
	   :in-theory (disable definicion-uniforme-de-es-proposicional))))

(defthm medida-uniforme-componentes-beta-disminuye
  (implies (beta-formula F)
	   (and (< (medida-uniforme (componente-1 F))
		   (medida-uniforme F))
		(< (medida-uniforme (componente-2 F))
		   (medida-uniforme F))))
  :rule-classes :linear
  :hints (("Goal"
	   :in-theory (disable definicion-uniforme-de-es-proposicional))))

;;;   Otras propiedades de la medida uniforme

(defthm medida-uniforme-doble-negacion->0
  (implies (doble-negacion F)
	   (< 0 (medida-uniforme F)))
  :rule-classes :linear)

(defthm medida-uniforme-alfa-formula->0
  (implies (alfa-formula F)
	   (< 0 (medida-uniforme F)))
  :rule-classes :linear)

(defthm medida-uniforme-beta-formula->0
  (implies (beta-formula F)
	   (< 0 (medida-uniforme F)))
  :rule-classes :linear)

(defthm medida-uniforme-suma-componentes-alfa-disminuye
  (implies (alfa-formula F)
	   (< (+ (medida-uniforme (componente-1 F))
		 (medida-uniforme (componente-2 F)))
	      (medida-uniforme F)))
  :rule-classes :linear
  :hints (("Goal" 
	   :in-theory (disable definicion-uniforme-de-es-proposicional))))

;;;============================================================================