;;;============================================================================
;;;
;;; Ttulo: Ejemplo de "Pruebas de terminacin basadas en multiconjuntos"
;;; Parte del proyecto de tesis:
;;;    "Verificacin de sistemas de razonamiento automtico"
;;; FJMM 21-09-2000
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(defpkg "MUL" (union-eq *acl2-exports*
                        *common-lisp-symbols-from-main-lisp-package*))

(certify-book "../6-extensiones/arboles-binarios" 1)

|#

(in-package "ACL2")

;(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 1) Descripcin de la situacin
;;;
;;;============================================================================

(encapsulate
 (((i *) => *)
  ((d *) => *)
  ((base *) => *)
  ((f-base *) => *)
  ((mp-bin *) => *)
  ((rel-bin * *) => *)
  ((fn-bin *) => *)
  ((combina * *) => *)
  ((neutro) => *)
  ((medida *) => *))

;;   Definiciones y propiedades de los rboles binarios

 (local
  (defun base (TT)
    (atom TT)))

 (local
  (defun i (TT)
    (car TT)))

 (local
  (defun d (TT)
    (cdr TT)))

 (local
  (defun mp-bin (v)
    (e0-ordinalp v)))

 (local
  (defun rel-bin (v-1 v-2)
    (e0-ord-< v-1 v-2)))

 (local
  (defun fn-bin (v)
    v))

 (defthm rel-bin-es-bien-fundamentada-en-mp-bin
   (and (implies (mp-bin x)
                 (e0-ordinalp (fn-bin x)))
        (implies (and (mp-bin x)
                      (mp-bin y)
                      (rel-bin x y))
                 (e0-ord-< (fn-bin x) (fn-bin y))))
   :rule-classes :well-founded-relation)

 (local
  (defun medida (TT)
    (acl2-count TT)))

 (defthm medida-es-mp-bin
   (mp-bin (medida TT)))

 (defthm hijos-rel-bin
   (implies (not (base TT))
            (and (rel-bin (medida (i TT)) (medida TT))
                 (rel-bin (medida (d TT)) (medida TT)))))

;;   Algunas propiedades de la funcin recursiva

 (local
  (defun combina (v-1 v-2)
    (append v-1 v-2)))

 (local
  (defun neutro ()
    nil))

 (local
  (defun f-base (TT)
    (declare (ignore TT))
    1))

 (defthm combina-asociativa
   (equal (combina (combina v-1 v-2) v-3)
          (combina v-1 (combina v-2 v-3))))

 (defthm combina-neutro
   (equal (combina (neutro) v) v))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Esquema recursivo
;;;
;;;----------------------------------------------------------------------------

(defun f-recursiva (TT)
  (declare (xargs :well-founded-relation rel-bin
                  :measure (medida TT)))
  (if (base TT)
      (f-base TT)
      (combina (f-recursiva (i TT))
               (f-recursiva (d TT)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Esquema iterativo (recursivo de cola)
;;;
;;;----------------------------------------------------------------------------

(defun f-iterativa-aux (Lista-TT v)
  (declare (xargs :mode :program))
  (cond ((endp Lista-TT) v)
        ((base (car Lista-TT))
         (f-iterativa-aux (cdr Lista-TT) 
                          (combina v (f-base (car Lista-TT)))))
        (t (f-iterativa-aux (list* (i (car Lista-TT))
                                   (d (car Lista-TT))
                                   (cdr Lista-TT))
                            v))))

(defun f-iterativa (TT)
  (declare (xargs :mode :program))
  (f-iterativa-aux (list TT) (neutro)))

(include-book "../6-extensiones/defmul")

; (defmul-components rel-bin)
; => (REL-BIN REL-BIN-ES-BIEN-FUNDAMENTADA-EN-MP-BIN MP-BIN FN-BIN X Y)

(defmul (REL-BIN REL-BIN-ES-BIEN-FUNDAMENTADA-EN-MP-BIN MP-BIN FN-BIN X Y))

(defun medida-lista (Lista-TT)
  (if (endp Lista-TT)
      nil
      (cons (medida (car Lista-TT))
            (medida-lista (cdr Lista-TT)))))

(verify-termination f-iterativa-aux
  (declare (xargs :measure (medida-lista Lista-TT)
                  :well-founded-relation mul-rel-bin)))

(verify-termination f-iterativa)

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Igualdad de ambos esquemas
;;;
;;;----------------------------------------------------------------------------

(defun combina-recursiva (Lista-TT v)
  (if (endp Lista-TT)
      v
      (combina-recursiva (cdr Lista-TT)
                         (combina v (f-recursiva (car Lista-TT))))))

(defthm combina-recursiva-igual-f-iterativa
  (equal (f-iterativa-aux Lista-TT v)
         (combina-recursiva Lista-TT v)))

(defthm f-recursiva-igual-f-iterativa
  (equal (f-iterativa TT)
         (f-recursiva TT)))

;;;============================================================================