;;;============================================================================
;;;
;;; Ttulo: Ejemplo de "Teoras Genricas"
;;; Parte del proyecto de tesis:
;;;    "Verificacin de sistemas de razonamiento automtico"
;;; FJMM 21-09-2000
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../6-extensiones/insercion-ordenada")

|#

(in-package "ACL2")

(include-book "../6-extensiones/generic-theory")

;;;============================================================================
;;;
;;; 1) Funcin genrica de ordenacin por inserccin
;;;
;;;============================================================================

(def-generic-theory *insercion-ordenada*

  (encapsulate
   (((menor * *) => *)
    ((dominio *) => *))
   
   (local
    (defun menor (x y)
      (declare (ignore x y))
      t))

   (local
    (defun dominio (x)
      (declare (ignore x)) 
      t))
   
   (defthm menor-es-transitivo
     (implies (and (menor x y)
		   (menor y z)
		   (dominio x)
		   (dominio y)
		   (dominio z))
	      (menor x z)))
   
   (defthm menor-es-total
     (implies (and (not (menor x y))
		   (dominio x)
		   (dominio y))
	      (menor y x)))
   
   )

  (export
   (defun inserta-en-orden (e x)
     (if (or (endp x)
	     (menor e (first x)))
	 (cons e x)
	 (cons (first x)
	       (inserta-en-orden e (rest x)))))
   )

  (export
   (defun ordena-insercion (x)
     (if (endp x) ()
	 (inserta-en-orden (first x)
			   (ordena-insercion (rest x)))))
   )
  
  (export
   (defun lista-en-dominio (x)
     (or (endp x)
	 (and (dominio (first x))
	      (lista-en-dominio (rest x)))))
   )
  
  (export
   (defun lista-ordenada (x)
     (or (endp x)
	 (endp (rest x))
	 (and (menor (first x) (second x))
	      (lista-ordenada (rest x)))))
   )

   (defthm inserta-en-orden-preserva-el-dominio
     (implies (and (dominio e)
		   (lista-en-dominio x))
	      (lista-en-dominio (inserta-en-orden e x))))

   (defthm ordena-insercion-preserva-el-dominio
     (implies (lista-en-dominio x)
	      (lista-en-dominio (ordena-insercion x))))

   (defthm inserta-en-orden-es-ordenada
     (implies (and (lista-en-dominio x)
		   (lista-ordenada x)
		   (dominio e))
	      (lista-ordenada (inserta-en-orden e x))))

  (export
   (defthm ordena-insercion-es-ordenada
     (implies (lista-en-dominio x)
	      (lista-ordenada (ordena-insercion x))))
   )

  )

;;;----------------------------------------------------------------------------
;;;
;;; Algunas instancias
;;;
;;;----------------------------------------------------------------------------

(defun orden-enteros (x y) 
  (<= x y))

(definstance-*insercion-ordenada*
  ((dominio   integerp)
   (menor     orden-enteros))
  "-enteros")

;;; ACL2 !>(ordena-insercion-enteros '(4 3 5 6 2 8 9))
;;; (2 3 4 5 6 8 9)

(defun lista-de-enteros (l)
  (cond ((endp l) t)
	(t (and (integerp (car l))
		(lista-de-enteros (cdr l))))))

(defun orden-lexicografico-enteros (l1 l2)
  (cond ((endp l1) t)
	((endp l2) nil)
	((< (car l1) (car l2)) t)
	((= (car l1) (car l2))
	 (orden-lexicografico-enteros (cdr l1) (cdr l2)))
	(t nil)))

(defthm orden-lexicografico-enteros-es-total
  (IMPLIES (AND (NOT (ORDEN-LEXICOGRAFICO-ENTEROS X Y))
		(LISTA-DE-ENTEROS X)
		(LISTA-DE-ENTEROS Y))
	   (ORDEN-LEXICOGRAFICO-ENTEROS Y X)))

(defthm orden-lexicografico-enteros-es-transitivo
  (IMPLIES (AND (ORDEN-LEXICOGRAFICO-ENTEROS X Y)
		(ORDEN-LEXICOGRAFICO-ENTEROS Y Z)
		(LISTA-DE-ENTEROS X)
		(LISTA-DE-ENTEROS Y)
		(LISTA-DE-ENTEROS Z))
	   (ORDEN-LEXICOGRAFICO-ENTEROS X Z)))

(definstance-*insercion-ordenada*
  ((dominio   lista-de-enteros)
   (menor     orden-lexicografico-enteros))
  "-lista-enteros")

;;; ACL2 !>(ordena-insercion-lista-enteros '((1 5) (4 5) (2 3 4) (1 2 3)))
;;; ((1 2 3) (1 5) (2 3 4) (4 5))

;;;============================================================================

