#| Cdigo para la certificacin:

(certify-book "../6-extensiones/listas")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; Extensin conjuntiva de una propiedad.
;;;
;;;============================================================================

;;;   Supongamos que disponemos de una propiedad $P$ definida sobre cierto tipo
;;; de objetos. Extendemos dicha propiedad de manera conjuntiva a listas
;;; $P-lst$, de manera que una lista de objetos tiene la propiedad $P-lst$ si
;;; todos sus elementos tienen dicha propiedad.

;;;   Supongamos la existencia de cierta lista lst para la que se ha probado
;;; que sus elementos (de manera individual) tienen la propiedad $P$, entonces
;;; se puede demostrar que la lista lst tiene la propiedad $P-lst$.

(encapsulate
 (((lista-strategy-prop *) => *)
  ((lista-strategy-hyp) => *)
  ((lista-strategy-lst) => *))
 
 (local
  (defun lista-strategy-prop (x)
    (atom x)))

 (local
  (defun lista-strategy-hyp ()
    t))
 
 (local
  (defun lista-strategy-lst ()
    nil))

 (defthm member-equal-lista-prop
   (implies (and (lista-strategy-hyp)
		 (member-equal e (lista-strategy-lst)))
     (lista-strategy-prop e))))

(defun lista-strategy-prop-lista (lista)
  (declare (xargs :guard (true-listp lista)))
  (if (endp lista)
      t
      (and (lista-strategy-prop (car lista))
	   (lista-strategy-prop-lista (cdr lista)))))

(encapsulate
 ()

 (local
  (defun sublista-strategy (lst1 lst2)
    (declare (xargs :guard (and (true-listp lst1)
				(true-listp lst2))))
    (if (endp lst1)
	t
	(and (member-equal (car lst1) lst2)
	     (sublista-strategy (cdr lst1) lst2)))))

 (local
  (defthm sublista-strategy-cons
    (implies (sublista-strategy lst1 lst2)
      (sublista-strategy lst1 (cons e lst2)))))

 (local
  (defthm sublista-strategy-reflexiva
    (sublista-strategy lista lista)))

 (local
  (defthm sublista-strategy-lista-strategy-prop-lista
    (implies (and (lista-strategy-hyp)
		  (sublista-strategy lista (lista-strategy-lst)))
      (lista-strategy-prop-lista lista))))

 (defthm lista-strategy-prop-lista-strategy-lst
   (implies (lista-strategy-hyp)
     (lista-strategy-prop-lista (lista-strategy-lst)))))

;;;----------------------------------------------------------------------------
;;;
;;; Computed-hint
;;;
;;;----------------------------------------------------------------------------

(defun components-strategy-lst (form)
  (declare (xargs :mode :program))

  (case-match form
	      
	      (('IMPLIES form-hyp form-lista)
	       (declare (ignore form-lista))
	       form-hyp)
	      
	      (& t)))

(defun defun-defstrategy-lst-hint (form lista prop prop-lst)
  (declare (xargs :mode :program))

  `(:use (:functional-instance 
	  lista-strategy-prop-lista-strategy-lst
	  (lista-strategy-hyp 
	   (lambda () ,(components-strategy-lst (first form))))
	  (lista-strategy-lst (lambda () ,lista))
          (lista-strategy-prop ,prop)
          (lista-strategy-prop-lista ,prop-lst))))

(defmacro extension-conjuntiva-a-listas (lista prop prop-lst)
  `(defun-defstrategy-lst-hint clause 
     (quote ,lista) (quote ,prop) (quote ,prop-lst)))

;;;============================================================================
;;; 
;;; Ejemplo de uso del computed-hint
;;;
;;;============================================================================

#|

(include-book "../6-extensiones/defmul")

; (defmul-components e0-ord-<)
; => (E0-ORD-< NIL E0-ORDINALP E0-ORD-<-FN NIL NIL)

(defmul (E0-ORD-< NIL E0-ORDINALP E0-ORD-<-FN NIL NIL))

(encapsulate
 (((obj1) => *)
  ((obj1-lst) => *)
  ((obj1-medida *) => *))

 (local 
  (defun obj1 ()
    0))

 (local
  (defun obj1-lst ()
    nil))
 
 (local
  (defun obj1-medida (x)
    (declare (ignore x))
    0))

 (defthm e0-ordinalp-obj1-medida
   (e0-ordinalp (obj1-medida objeto)))

 (defthm e0-ord-<-obj1-lst
   (implies (member-equal obj2 (obj1-lst))
     (e0-ord-< (obj1-medida obj2) (obj1-medida (obj1))))))

(defun obj1-medida-lista (obj-lst)
  (cond ((endp obj-lst) nil)
	(t (cons (obj1-medida (car obj-lst)) 
		 (obj1-medida-lista (cdr obj-lst))))))

(local
 (defun e0-ord-<-bigger (ol o)
   (if (endp ol)
       t
       (and (e0-ord-< (car ol) o)
	    (e0-ord-<-bigger (cdr ol) o)))))

(local
 (defthm e0-ord-<-bigger-not-member
   (implies (e0-ord-<-bigger ol o)
     (not (member-equal o ol)))))

(local
 (defthm e0-ord-<-bigger-forall-exists-e0-ord-<-bigger
   (implies (e0-ord-<-bigger ol o)
     (forall-exists-e0-ord-<-bigger ol (list o)))))

(defmacro show-hint (hint &optional marker)
  (cond
   ((and (consp hint)
         (stringp (car hint)))
    hint)
   (t
    `(let ((marker ,marker)
           (ans ,(if (symbolp hint)
                     `(,hint id clause world)
                   hint)))
       (if ans
           (prog2$
            (cw "~%***** Computed Hint~#0~[~/ (from hint ~x1)~]~%~x2~%~%"
                (if (null marker) 0 1)
                marker
                (cons (string-for-tilde-@-clause-id-phrase id)
                      ans))
            ans)
         nil)))))

(defthm e0-ord-<-bigger-obj-medida
  (e0-ord-<-bigger (obj1-medida-lista (obj1-lst))
		   (obj1-medida (obj1)))
  :hints ((show-hint
	   (extension-conjuntiva-a-listas
	     (obj1-lst)
	     (lambda (elt)
	       (e0-ord-< (obj1-medida elt)
			 (obj1-medida (obj1))))
	     (lambda (lst)
		(e0-ord-<-bigger (obj1-medida-lista lst)
				 (obj1-medida (obj1))))
	     )
	   2)))
|#

;;;============================================================================
;;;
;;; Extensin disjuntiva de una propiedad.
;;;
;;;============================================================================

;;;   Supongamos que disponemos de una propiedad $P$ definida sobre cierto tipo
;;; de objetos. Extendemos dicha propiedad de manera disjuntiva a listas
;;; $P-lst$, de manera que una lista de objetos tiene la propiedad $P-lst$ si
;;; alguno de sus elementos tienen dicha propiedad.

;;;   Supongamos la existencia de cierta lista lst para la que se ha probado
;;; (constructivamente) que existe un elemento que tienen la propiedad $P$,
;;; entonces se puede demostrar que la lista lst tiene la propiedad $P-lst$.

;;;============================================================================
