;;; multiconjuntos.lisp
;;; Libro sobre multiconjuntos
;;; Created: 13-7-00  Last revision: 25-7-00
;;;============================================================================

#| Cdigo para la certificacin:

(defpkg "MUL" (union-eq *acl2-exports*
			*common-lisp-symbols-from-main-lisp-package*))

(certify-book "../6-extensiones/multiconjuntos" 1)

|#

(in-package "MUL")

;;; Estamos utilizando "equal" para la igualdad.

;;; Eliminacin de un elemento de un multiconjunto.

(defun remove-one (elt lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) lista)
        ((equal (car lista) elt) (cdr lista))
        (t (cons (car lista) (remove-one elt (cdr lista))))))

; (defun remove-one (x m)
;   (declare (xargs :guard (true-listp m)))
;   (cond ((atom m) m)
; 	((equal x (car m)) (cdr m))
; 	(t (cons (car m) (remove-one x (cdr m))))))

(defthm theorem-1
  (implies (not (equal x (car m)))
    (equal (car (remove-one x m)) (car m))))

;;; Tiempo:  0.02 segundos (prueba: 0.01, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:ELIM CAR-CDR-ELIM)
;;;      (:EXECUTABLE-COUNTERPART EQUAL)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE DEFAULT-CAR)

(defthm theorem-2
  (implies (and (member-equal x m)
		(not (equal x y)))
    (member-equal x (remove-one y m))))

;;; Tiempo:  0.06 segundos (prueba: 0.01, impresion: 0.04, otros: 0.01)
;;; Casos Tot: 6
;;; Usa: (:DEFINITION ATOM)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm theorem-3
  (implies (not (member-equal x m))
    (not (member-equal x (remove-one y m)))))

;;; Tiempo:  0.05 segundos (prueba: 0.04, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 4
;;; Usa: (:DEFINITION ATOM)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm theorem-4
  (equal (remove-one x (remove-one y m))
	 (remove-one y (remove-one x m))))

;;; Tiempo:  0.08 segundos (prueba: 0.04, impresion: 0.02, otros: 0.02)
;;; Casos Tot: 5
;;; Usa: (:DEFINITION ATOM)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)

;;; Igualdad entre multiconjuntos.

(defun multiset-equal (m1 m2)
  (declare (xargs :guard (and (true-listp m1) (true-listp m2))))
  (cond ((endp m1) (endp m2))
	((member-equal (car m1) m2)
	 (multiset-equal (cdr m1) (remove-one (car m1) m2)))
	(t nil)))

;;; La igualdad entre multiconjuntos es reflexiva:

(defthm multiset-equal-reflexive 
  (multiset-equal m m))

;;; Tiempo:  0.03 segundos (prueba: 0.01, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:TYPE-PRESCRIPTION MULTISET-EQUAL)

(defthm multiset-equal-remove-one
  (implies (and (member-equal x m1)
		(member-equal x m2)
		(multiset-equal (remove-one x m1)
				(remove-one x m2)))
    (multiset-equal m1 m2)))

;;; Tiempo:  0.25 segundos (prueba: 0.23, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 15
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE THEOREM-2)
;;;      (:REWRITE THEOREM-3)
;;;      (:REWRITE THEOREM-4)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-EQUAL)

(defthm multiset-equal-symmetric
  (implies (multiset-equal m1 m2)
    (multiset-equal m2 m1)))

;;; Tiempo:  0.05 segundos (prueba: 0.02, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 4
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-REMOVE-ONE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-EQUAL)

(defun multiset-subp (m1 m2)
  (declare (xargs :guard (and (true-listp m1) (true-listp m2))))
  (cond ((endp m1) t)
	((member-equal (car m1) m2)
	 (multiset-subp (cdr m1) (remove-one (car m1) m2)))
	(t nil)))

(defthm multiset-subp-remove-one
  (implies (and (member-equal x m1)
		(member-equal x m2))
    (equal (multiset-subp (remove-one x m1)
			  (remove-one x m2))
	   (multiset-subp m1 m2))))

;;; Tiempo:  0.25 segundos (prueba: 0.22, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 14
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE THEOREM-2)
;;;      (:REWRITE THEOREM-3)
;;;      (:REWRITE THEOREM-4)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm multiset-subp-member-equal
  (implies (and (multiset-subp m1 m2)
		(member-equal x m1))
    (member-equal x m2)))

;;; Tiempo:  0.06 segundos (prueba: 0.02, impresion: 0.02, otros: 0.02)
;;; Casos Tot: 5
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE THEOREM-3)

(defthm multiset-subp-transitive 
  (implies (and (multiset-subp m1 m2)
		(multiset-subp m2 m3))
    (multiset-subp m1 m3)))

;;; Tiempo:  0.07 segundos (prueba: 0.04, impresion: 0.01, otros: 0.02)
;;; Casos Tot: 5
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-SUBP-MEMBER-EQUAL)
;;;      (:REWRITE MULTISET-SUBP-REMOVE-ONE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-equal-implies-subp
  (implies (multiset-equal m1 m2)
    (multiset-subp m1 m2)))

;;; Tiempo:  0.06 segundos (prueba: 0.01, impresion: 0.04, otros: 0.01)
;;; Casos Tot: 4
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm tecnico
  (implies (and (consp m2)
		(member-equal (car m2) m1))
    (equal (multiset-subp (remove-one (car m2) m1)
			  (cdr m2))
	   (multiset-subp m1 m2)))
  :hints (("Goal" :use
	   (:instance multiset-subp-remove-one
		      (x (car m2))))))

;;; Tiempo:  0.07 segundos (prueba: 0.05, impresion: 0.00, otros: 0.02)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm subp-theorem-1
  (implies (and (multiset-subp m2 m1)
		(multiset-subp m1 m2))
    (multiset-equal m1 m2))
  :rule-classes :type-prescription)

;;; Tiempo:  0.12 segundos (prueba: 0.08, impresion: 0.03, otros: 0.01)
;;; Casos Tot: 5
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-EQUAL-REFLEXIVE)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE MULTISET-SUBP-MEMBER-EQUAL)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:REWRITE TECNICO)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm another-definition-multiset-equal
  (iff (multiset-equal m1 m2)
       (and (multiset-subp m1 m2)
	    (multiset-subp m2 m1)))
  :rule-classes :definition)

;;; Tiempo:  0.20 segundos (prueba: 0.03, impresion: 0.00, otros: 0.17)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IFF)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:TYPE-PRESCRIPTION MULTISET-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)
;;;      (:TYPE-PRESCRIPTION SUBP-THEOREM-1)

(defequiv multiset-equal)

;;; Tiempo:  0.03 segundos (prueba: 0.02, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:REWRITE MULTISET-EQUAL-REFLEXIVE)
;;;      (:DEFINITION IMPLIES)
;;;      (:COMPOUND-RECOGNIZER BOOLEANP-COMPOUND-RECOGNIZER)
;;;      (:TYPE-PRESCRIPTION MULTISET-EQUAL)
;;;      (:TYPE-PRESCRIPTION SUBP-THEOREM-1)
;;;      (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)

;;;============================================================================
;;;
;;; Pertenencia a multiconjuntos.
;;;
;;;============================================================================

(defun member-multiset (x m)
  (declare (xargs :guard (true-listp m)))
  (member-equal x m))

(defcong multiset-equal iff (member-multiset x m) 2)

;;; Tiempo:  0.03 segundos (prueba: 0.01, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION MEMBER-MULTISET)
;;;      (:DEFINITION IFF)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:REWRITE MULTISET-SUBP-MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defcong multiset-equal iff (multiset-subp m1 m2) 1)

;;; Tiempo:  0.19 segundos (prueba: 0.18, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IFF)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defcong multiset-equal iff (multiset-subp m1 m2) 2)

;;; Tiempo:  0.07 segundos (prueba: 0.06, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IFF)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

;;;============================================================================
;;;
;;; Unin de multiconjuntos.
;;;
;;;============================================================================

(defun union-multiset (m1 m2)
  (declare (xargs :guard (and (true-listp m1) (true-listp m2))))
  (cond ((endp m1) m2)
	(t (cons (car m1) (union-multiset (cdr m1) m2)))))

(in-theory (disable another-definition-multiset-equal))

(defcong multiset-equal multiset-equal (union-multiset m1 m2) 2)

;;; Tiempo:  0.07 segundos (prueba: 0.05, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:REWRITE CDR-CONS)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:REWRITE CAR-CONS)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION ENDP)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:DEFINITION UNION-MULTISET)

(defthm multiset-subp-union-multiset-1
  (multiset-subp m1
		 (union-multiset m1 m2)))

;;; Tiempo:  0.09 segundos (prueba: 0.06, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:DEFINITION UNION-MULTISET)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-subp-cons-2
  (multiset-subp m2
		 (cons x m2)))

;;; Tiempo:  0.32 segundos (prueba: 0.29, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CDR-ELIM)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-subp-union-multiset-2
  (multiset-subp m2
		 (union-multiset m1 m2)))

;;; Tiempo:  0.08 segundos (prueba: 0.05, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION UNION-MULTISET)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-CONS-2)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(local
 (defthm member-equal-car-union-multiset-2
   (implies (consp m1)
     (member-equal (car m1) (union-multiset m2 m1)))))

;;; Tiempo:  0.03 segundos (prueba: 0.00, impresion: 0.01, otros: 0.02)
;;; Casos Tot: 2
;;; Usa: (:ELIM CAR-CDR-ELIM)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE CAR-CONS)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:DEFINITION UNION-MULTISET)

(defthm multiset-subp-cons-remove-one
  (implies (member-equal x m1)
    (multiset-subp (cons y (remove-one x m1))
		   (remove-one x (cons y m1)))))

;;; Tiempo:  0.18 segundos (prueba: 0.16, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 4
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(local
 (defthm member-equal-union-multiset
   (implies (member-equal x m1)
     (member-equal x (union-multiset m2 m1)))))

;;; Tiempo:  0.02 segundos (prueba: 0.01, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE CAR-CONS)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION ENDP)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:DEFINITION UNION-MULTISET)

(defthm multiset-subp-union-multiset-remove-one
  (implies (member-equal x m1)
    (multiset-subp (union-multiset m2 (remove-one x m1))
		   (remove-one x (union-multiset m2 m1)))))

;;; Tiempo:  0.83 segundos (prueba: 0.78, impresion: 0.03, otros: 0.02)
;;; Casos Tot: 6
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:DEFINITION UNION-MULTISET)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MEMBER-EQUAL-UNION-MULTISET)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-subp-union-multiset-remove-one-instance
  (implies (consp m1)
    (multiset-subp (union-multiset m2 (cdr m1))
		   (remove-one (car m1) (union-multiset m2 m1))))
  :hints (("Goal" :use (:instance multiset-subp-union-multiset-remove-one
				  (x (car m1))))))

;;; Tiempo:  0.07 segundos (prueba: 0.05, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-subp-union-multiset
  (multiset-subp (union-multiset m1 m2)
		 (union-multiset m2 m1)))

;;; Tiempo:  0.15 segundos (prueba: 0.10, impresion: 0.03, otros: 0.02)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION UNION-MULTISET)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MEMBER-EQUAL-UNION-MULTISET)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:REWRITE MULTISET-SUBP-UNION-MULTISET-1)
;;;      (:REWRITE MULTISET-SUBP-UNION-MULTISET-REMOVE-ONE-INSTANCE)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm union-multiset-conmutative
  (multiset-equal (union-multiset m2 m1)
		  (union-multiset m1 m2))
  :hints (("Goal" :in-theory (enable another-definition-multiset-equal))))

;;; Tiempo:  0.05 segundos (prueba: 0.01, impresion: 0.00, otros: 0.04)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:REWRITE MULTISET-SUBP-UNION-MULTISET)

(defcong multiset-equal multiset-equal (union-multiset m2 m1) 1
  :hints (("Goal" :in-theory (disable union-multiset multiset-equal))))

;;; Tiempo:  0.19 segundos (prueba: 0.01, impresion: 0.00, otros: 0.18)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:CONGRUENCE MULTISET-EQUAL-IMPLIES-MULTISET-EQUAL-UNION-MULTISET-2)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:REWRITE UNION-MULTISET-CONMUTATIVE)

(defthm another-definition-union-multiset
  (equal (union-multiset m1 m2)
	 (append m1 m2))
  :rule-classes :definition)

;;; Tiempo:  0.03 segundos (prueba: 0.01, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION BINARY-APPEND)
;;;      (:DEFINITION ENDP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION UNION-MULTISET)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)

(in-theory (disable another-definition-union-multiset))

(defthm union-multiset-associative
  (multiset-equal (union-multiset m1 (union-multiset m2 m3))
		  (union-multiset (union-multiset m1 m2) m3))
  :hints (("Goal" :in-theory (enable another-definition-union-multiset))))

;;; Tiempo:  0.11 segundos (prueba: 0.05, impresion: 0.00, otros: 0.06)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION ANOTHER-DEFINITION-UNION-MULTISET)
;;;      (:DEFINITION BINARY-APPEND)
;;;      (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)

;;; Este resultado es falso, la unin de multiconjuntos NO es el menor conjunto
;;; que contiene a los que se unen:
;
; (defthm union-multiset-least
;   (implies (and (multiset-subp m1 m3)
;                 (multiset-subp m2 m3))
;     (multiset-subp (union-multiset m1 m2) m3)))
;

;
; m1 <= '(1 2 3)
; m2 <= '(3 4 5)
; (union-multiset m1 m2) <= '(1 2 3 3 4 5)
; m3 <= '(1 2 3 4 5)
;
;;; En este caso se tiene (multiset-subp m1 m3) y (multiset-subp m2 m3), pero
;;; no se tiene (multiset-subp (union-multiset m1 m2) m3).

;;;============================================================================
;;;
;;; Diferencia de multiconjuntos.
;;;
;;;============================================================================

(defun multiset-diff (m n)
  (declare (xargs :guard (and (true-listp m) (true-listp n))))
  (if (atom n)
      m
      (multiset-diff (remove-one (car n) m) (cdr n))))

(defcong multiset-equal multiset-equal (multiset-diff m1 m2) 2)

;;; Tiempo:  0.43 segundos (prueba: 0.34, impresion: 0.08, otros: 0.01)
;;; Casos Tot: 9
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:ELIM CAR-CDR-ELIM)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE THEOREM-4)
;;;      (:REWRITE CAR-CONS)
;;;      (:DEFINITION MULTISET-DIFF)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION ATOM)
;;;      (:DEFINITION MEMBER-EQUAL)

(defcong multiset-equal multiset-equal (remove-one x m) 2)

;;; Tiempo:  0.13 segundos (prueba: 0.08, impresion: 0.04, otros: 0.01)
;;; Casos Tot: 4
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:REWRITE THEOREM-4)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE THEOREM-2)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:REWRITE CAR-CONS)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION ATOM)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:DEFINITION REMOVE-ONE)

(defcong multiset-equal multiset-equal (multiset-diff m1 m2) 1)

;;; Tiempo:  0.07 segundos (prueba: 0.04, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:CONGRUENCE MULTISET-EQUAL-IMPLIES-MULTISET-EQUAL-REMOVE-ONE-2)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION ATOM)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:DEFINITION MULTISET-DIFF)

(defthm multiset-diff-union-multiset-1-multiset-equal
  (multiset-equal (multiset-diff (union-multiset l1 l2)
				 (union-multiset l1 l3))
		  (multiset-diff l2 l3)))

;;; Tiempo:  0.21 segundos (prueba: 0.20, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 2
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MULTISET-DIFF)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:DEFINITION UNION-MULTISET)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)

(defthm multiset-diff-union-multiset-2-multiset-equal
  (multiset-equal (multiset-diff (union-multiset l2 l1)
				 (union-multiset l3 l1))
		  (multiset-diff l2 l3))
  :hints (("Goal" :in-theory (disable union-multiset))))

;;; Tiempo:  0.06 segundos (prueba: 0.01, impresion: 0.00, otros: 0.05)
;;; Casos Tot: 0
;;; Usa: (:CONGRUENCE MULTISET-EQUAL-IMPLIES-MULTISET-EQUAL-MULTISET-DIFF-1)
;;;      (:CONGRUENCE MULTISET-EQUAL-IMPLIES-MULTISET-EQUAL-MULTISET-DIFF-2)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-DIFF-UNION-MULTISET-1-MULTISET-EQUAL)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE UNION-MULTISET-CONMUTATIVE)

;;;============================================================================
;;;
;;; Interseccin de multiconjuntos.
;;;
;;;============================================================================

(defun multiset-inter (m n)
  (declare (xargs :guard (and (true-listp m) (true-listp n))))
  (cond ((endp m) nil)
	((member-equal (car m) n)
	 (cons (car m)
	       (multiset-inter (cdr m) (remove-one (car m) n))))
	(t (multiset-inter (cdr m) n))))

(local
 (defthm multiset-inter-samples
   (and 
    (multiset-equal (multiset-inter '(1 2 3) '(2 4 5)) '(2))
    (multiset-equal (multiset-inter '(1 2 2 3 3 3) '(3 3 2 4)) '(3 2 3)))
   :rule-classes nil))

(defthm multiset-inter-reflexive
  (multiset-equal (multiset-inter m m) m))

;;; Tiempo:  0.06 segundos (prueba: 0.03, impresion: 0.01, otros: 0.02)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:EXECUTABLE-COUNTERPART CONSP)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)

(defthm multiset-inter-remove-one-member
  (implies (and (member-equal x m1)
		(member-equal x m2))
    (multiset-equal (multiset-inter (remove-one x m1)
				    (remove-one x m2))
		    (remove-one x (multiset-inter m1 m2)))))

;;; Tiempo:  0.77 segundos (prueba: 0.67, impresion: 0.09, otros: 0.01)
;;; Casos Tot: 15
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE THEOREM-2)
;;;      (:REWRITE THEOREM-3)
;;;      (:REWRITE THEOREM-4)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)

(defthm member-equal-multiset-inter
  (implies (and (member-equal x m1)
		(member-equal x m2))
    (member-equal x (multiset-inter m1 m2))))

;;; Tiempo:  0.13 segundos (prueba: 0.10, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 11
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE THEOREM-2)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)

(defthm multiset-inter-remove-one-not-member-2
  (implies (and (member-equal x m1)
		(not (member-equal x m2)))
    (multiset-equal (multiset-inter (remove-one x m1) m2)
		    (multiset-inter m1 m2))))
  
;;; Tiempo:  0.40 segundos (prueba: 0.34, impresion: 0.04, otros: 0.02)
;;; Casos Tot: 10
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-REFLEXIVE)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE THEOREM-3)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)

(defcong multiset-equal multiset-equal (multiset-inter m1 m2) 1)

;;; Tiempo:  0.34 segundos (prueba: 0.26, impresion: 0.07, otros: 0.01)
;;; Casos Tot: 5
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION NOT)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:EXECUTABLE-COUNTERPART MULTISET-EQUAL)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MEMBER-EQUAL-MULTISET-INTER)
;;;      (:REWRITE MULTISET-INTER-REMOVE-ONE-MEMBER)
;;;      (:REWRITE MULTISET-INTER-REMOVE-ONE-NOT-MEMBER-2)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)

(defthm multiset-inter-multiset-subp-1
  (multiset-subp (multiset-inter m1 m2) m1))

;;; Tiempo:  0.40 segundos (prueba: 0.37, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:ELIM CAR-CDR-ELIM)
;;;      (:EXECUTABLE-COUNTERPART CONSP)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE MULTISET-SUBP-CONS-2)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-inter-multiset-subp-2
  (multiset-subp (multiset-inter m1 m2) m2))

;;; Tiempo:  0.16 segundos (prueba: 0.14, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:EXECUTABLE-COUNTERPART CONSP)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-inter-remove-one-not-member-1
  (implies (and (not (member-equal x m1))
		(member-equal x m2))
    (multiset-equal (multiset-inter m1 (remove-one x m2))
		    (multiset-inter m1 m2)))
  :hints (("Goal" :in-theory (enable another-definition-multiset-equal))))

;;; Tiempo:  0.42 segundos (prueba: 0.34, impresion: 0.04, otros: 0.04)
;;; Casos Tot: 7
;;; Usa: (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EXECUTABLE-COUNTERPART MULTISET-EQUAL)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE THEOREM-2)
;;;      (:REWRITE THEOREM-3)
;;;      (:REWRITE THEOREM-4)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)
;;;      (:TYPE-PRESCRIPTION SUBP-THEOREM-1)

(encapsulate
 ()
 
 (local
  (defthm multiset-inter-remove-one-not-member-1-instance
    (implies (and (consp m2)
		  (not (member-equal (car m2) m1)))
      (multiset-equal (multiset-inter m1 (cdr m2))
		      (multiset-inter m1 m2)))
    :hints (("Goal" :use
	     (:instance multiset-inter-remove-one-not-member-1
			(x (car m2)))))))

;;; Tiempo:  0.04 segundos (prueba: 0.01, impresion: 0.00, otros: 0.03)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EXECUTABLE-COUNTERPART NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)

 (local
  (defthm multiset-inter-remove-one-member-instance
    (implies (and (consp m2)
		  (member-equal (car m2) m1))
      (multiset-equal (multiset-inter (remove-one (car m2) m1)
				      (cdr m2))
		      (remove-one (car m2) (multiset-inter m1 m2))))
    :hints (("Goal" :use
	     (:instance multiset-inter-remove-one-member
			(x (car m2)))))))

;;; Tiempo:  0.04 segundos (prueba: 0.03, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

 (defthm multiset-inter-conmutative
   (multiset-equal (multiset-inter m2 m1)
		   (multiset-inter m1 m2)))

;;; Tiempo:  0.14 segundos (prueba: 0.11, impresion: 0.03, otros: 0.00)
;;; Casos Tot: 7
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-INTER)
;;;      (:DEFINITION NOT)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:EXECUTABLE-COUNTERPART CONSP)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MEMBER-EQUAL-MULTISET-INTER)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE MULTISET-INTER-REMOVE-ONE-MEMBER-INSTANCE)
;;;      (:REWRITE MULTISET-INTER-REMOVE-ONE-NOT-MEMBER-1-INSTANCE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-INTER)

 )

(defcong multiset-equal multiset-equal (multiset-inter m2 m1) 2
  :hints (("Goal" :in-theory (disable union-multiset multiset-equal))))

;;; Tiempo:  0.05 segundos (prueba: 0.01, impresion: 0.00, otros: 0.04)
;;; Casos Tot: 0
;;; Usa: (:CONGRUENCE MULTISET-EQUAL-IMPLIES-MULTISET-EQUAL-MULTISET-INTER-1)
;;;      (:DEFINITION IMPLIES)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE MULTISET-INTER-CONMUTATIVE)

(defthm multiset-inter-greatest
  (implies (and (multiset-subp m1 m2)
		(multiset-subp m1 m3))
    (multiset-subp m1 (multiset-inter m2 m3))))

;;; Tiempo:  0.80 segundos (prueba: 0.74, impresion: 0.05, otros: 0.01)
;;; Casos Tot: 5
;;; Usa: (:CONGRUENCE MULTISET-EQUAL-IMPLIES-IFF-MULTISET-SUBP-2)
;;;      (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MEMBER-EQUAL-MULTISET-INTER)
;;;      (:REWRITE MULTISET-INTER-REMOVE-ONE-MEMBER)
;;;      (:REWRITE MULTISET-SUBP-MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-subp-multiset-inter
  (implies (multiset-subp m2 m3)
    (multiset-subp (multiset-inter m1 m2)
		   m3))
  :hints (("Goal" :in-theory (disable multiset-subp-transitive)
	   :use (:instance multiset-subp-transitive
			   (m1 (multiset-inter m1 m2))))))

;;; Tiempo:  0.05 segundos (prueba: 0.00, impresion: 0.01, otros: 0.04)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION NOT)
;;;      (:REWRITE MULTISET-INTER-MULTISET-SUBP-2)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-inter-associative
  (multiset-equal (multiset-inter m1 (multiset-inter m2 m3))
		  (multiset-inter (multiset-inter m1 m2) m3))
  :hints (("Goal" :in-theory (enable another-definition-multiset-equal))))

;;; Tiempo:  0.23 segundos (prueba: 0.19, impresion: 0.00, otros: 0.04)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-INTER-CONMUTATIVE)
;;;      (:REWRITE MULTISET-INTER-GREATEST)
;;;      (:REWRITE MULTISET-INTER-MULTISET-SUBP-1)
;;;      (:REWRITE MULTISET-SUBP-MULTISET-INTER)

;;;============================================================================
;;;
;;; Unin "mnima" de multiconjuntos.
;;;
;;;============================================================================

;;; El menor multiconjunto que contiene a dos dados:

(defun multiset-unimin (m1 m2)
  (declare (xargs :guard (and (true-listp m1) (true-listp m2))))
  (cond ((endp m1) m2)
	((member-equal (car m1) m2)
	 (cons (car m1)
	       (multiset-unimin (cdr m1) (remove-one (car m1) m2))))
	(t (cons (car m1) (multiset-unimin (cdr m1) m2)))))

(local
 (defthm multiset-unimin-samples
   (and 
    (multiset-equal (multiset-unimin '(1 2 3) '(2 4 5)) '(1 2 3 4 5))
    (multiset-equal (multiset-unimin '(1 2 2 3 3 3) '(3 3 2 4)) '(1 2 2 3 3 3 4)))
   :rule-classes nil))

(defthm multiset-unimin-reflexive
  (multiset-equal (multiset-unimin m m) m))

;;; Tiempo:  0.07 segundos (prueba: 0.03, impresion: 0.02, otros: 0.02)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)

(defthm multiset-unimin-remove-one-member
  (implies (and (member-equal x m1)
		(member-equal x m2))
    (multiset-equal (multiset-unimin (remove-one x m1)
				     (remove-one x m2))
		    (remove-one x (multiset-unimin m1 m2)))))

;;; Tiempo:  0.63 segundos (prueba: 0.53, impresion: 0.09, otros: 0.01)
;;; Casos Tot: 15
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE THEOREM-2)
;;;      (:REWRITE THEOREM-3)
;;;      (:REWRITE THEOREM-4)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm member-equal-multiset-unimin-1
  (implies (member-equal x m1)
    (member-equal x (multiset-unimin m1 m2))))

;;; Tiempo:  0.11 segundos (prueba: 0.07, impresion: 0.03, otros: 0.01)
;;; Casos Tot: 15
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm member-equal-multiset-unimin-2
  (implies (member-equal x m2)
    (member-equal x (multiset-unimin m1 m2))))

;;; Tiempo:  0.08 segundos (prueba: 0.03, impresion: 0.04, otros: 0.01)
;;; Casos Tot: 6
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE THEOREM-2)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm multiset-unimin-remove-one-not-member-2
  (implies (and (member-equal x m1)
		(not (member-equal x m2)))
    (multiset-equal (multiset-unimin (remove-one x m1) m2)
		    (remove-one x (multiset-unimin m1 m2)))))

;;; Tiempo:  0.62 segundos (prueba: 0.42, impresion: 0.03, otros: 0.17)
;;; Casos Tot: 10
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-REFLEXIVE)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE THEOREM-3)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defcong multiset-equal multiset-equal (multiset-unimin m1 m2) 1)

;;; Tiempo:  0.40 segundos (prueba: 0.35, impresion: 0.04, otros: 0.01)
;;; Casos Tot: 5
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MEMBER-EQUAL-MULTISET-UNIMIN-1)
;;;      (:REWRITE MEMBER-EQUAL-MULTISET-UNIMIN-2)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE MULTISET-UNIMIN-REMOVE-ONE-MEMBER)
;;;      (:REWRITE MULTISET-UNIMIN-REMOVE-ONE-NOT-MEMBER-2)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm multiset-subp-multiset-unimin-1
  (multiset-subp m1 (multiset-unimin m1 m2)))

;;; Tiempo:  0.13 segundos (prueba: 0.10, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-subp-not-member-1-remove-one-2
  (implies (and (multiset-subp m1 m2)
		(not (member-equal x m1)))
    (multiset-subp m1 (remove-one x m2))))

;;; Tiempo:  0.47 segundos (prueba: 0.42, impresion: 0.03, otros: 0.02)
;;; Casos Tot: 5
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE THEOREM-2)
;;;      (:REWRITE THEOREM-4)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-unimin-least
  (implies (and (multiset-subp m1 m3)
                (multiset-subp m2 m3))
    (multiset-subp (multiset-unimin m1 m2) m3)))

;;; Tiempo:  0.93 segundos (prueba: 0.84, impresion: 0.07, otros: 0.02)
;;; Casos Tot: 7
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MULTISET-SUBP)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-MEMBER-EQUAL)
;;;      (:REWRITE MULTISET-SUBP-NOT-MEMBER-1-REMOVE-ONE-2)
;;;      (:REWRITE MULTISET-SUBP-REMOVE-ONE)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-unimin-remove-one-not-member-1
  (implies (and (not (member-equal x m1))
		(member-equal x m2))
    (multiset-equal (multiset-unimin m1 (remove-one x m2))
		    (remove-one x (multiset-unimin m1 m2)))))

;;; Tiempo:  0.43 segundos (prueba: 0.38, impresion: 0.04, otros: 0.01)
;;; Casos Tot: 6
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE THEOREM-2)
;;;      (:REWRITE THEOREM-3)
;;;      (:REWRITE THEOREM-4)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

(defthm multiset-subp-remove-one-cons
  (implies (and (member-equal x m1)
		(multiset-subp (remove-one x m1) m2))
    (multiset-subp m1 (cons x m2)))
  :hints (("Goal" :use (:instance multiset-subp-remove-one
				  (m2 (cons x m2))))))

;;; Tiempo:  0.10 segundos (prueba: 0.06, impresion: 0.00, otros: 0.04)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EXECUTABLE-COUNTERPART EQUAL)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-subp-multiset-unimin-2
  (multiset-subp m2 (multiset-unimin m1 m2)))

;;; Tiempo:  0.19 segundos (prueba: 0.16, impresion: 0.02, otros: 0.01)
;;; Casos Tot: 3
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-CONS-2)
;;;      (:REWRITE MULTISET-SUBP-REMOVE-ONE-CONS)
;;;      (:REWRITE MULTISET-SUBP-TRANSITIVE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(encapsulate
 ()

 (local
  (defthm multiset-unimin-remove-one-not-member-1-instance
    (implies (and (consp m2)
		  (not (member-equal (car m2) m1)))
      (multiset-equal (multiset-unimin m1 (cdr m2))
		      (remove-one (car m2) (multiset-unimin m1 m2))))
    :hints (("Goal" :use
	     (:instance multiset-unimin-remove-one-not-member-1
			(x (car m2)))))))

;;; Tiempo:  0.04 segundos (prueba: 0.02, impresion: 0.01, otros: 0.01)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EXECUTABLE-COUNTERPART NOT)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)

 (local
  (defthm multiset-unimin-remove-one-member-instance
    (implies (and (consp m2)
		  (member-equal (car m2) m1))
      (multiset-equal (multiset-unimin (remove-one (car m2) m1)
				       (cdr m2))
		      (remove-one (car m2) (multiset-unimin m1 m2))))
    :hints (("Goal" :use
	     (:instance multiset-unimin-remove-one-member
			(x (car m2)))))))

;;; Tiempo:  0.04 segundos (prueba: 0.03, impresion: 0.00, otros: 0.01)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

 (defthm multiset-unimin-conmutative
   (multiset-equal (multiset-unimin m2 m1)
		   (multiset-unimin m1 m2)))
 
;;; Tiempo:  0.20 segundos (prueba: 0.14, impresion: 0.05, otros: 0.01)
;;; Casos Tot: 7
;;; Usa: (:DEFINITION ENDP)
;;;      (:DEFINITION MEMBER-EQUAL)
;;;      (:DEFINITION MULTISET-EQUAL)
;;;      (:DEFINITION MULTISET-UNIMIN)
;;;      (:DEFINITION NOT)
;;;      (:DEFINITION REMOVE-ONE)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE CAR-CONS)
;;;      (:REWRITE CDR-CONS)
;;;      (:REWRITE MEMBER-EQUAL-MULTISET-UNIMIN-2)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE MULTISET-UNIMIN-REMOVE-ONE-MEMBER-INSTANCE)
;;;      (:REWRITE MULTISET-UNIMIN-REMOVE-ONE-NOT-MEMBER-1-INSTANCE)
;;;      (:TYPE-PRESCRIPTION MEMBER-EQUAL)

 )

(defcong multiset-equal multiset-equal (multiset-unimin m2 m1) 2
  :hints (("Goal" :in-theory (disable multiset-equal))))

;;; Tiempo:  0.05 segundos (prueba: 0.01, impresion: 0.01, otros: 0.03)
;;; Casos Tot: 0
;;; Usa: (:CONGRUENCE MULTISET-EQUAL-IMPLIES-MULTISET-EQUAL-MULTISET-UNIMIN-1)
;;;      (:DEFINITION IMPLIES)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-SYMMETRIC)
;;;      (:REWRITE MULTISET-UNIMIN-CONMUTATIVE)

(defthm multiset-subp-multiset-unimin
  (implies (multiset-subp m1 m2)
    (multiset-subp m1 (multiset-unimin m3 m2)))
  :hints (("Goal" :in-theory (disable multiset-subp-transitive)
	   :use (:instance multiset-subp-transitive
			   (m3 (multiset-unimin m3 m2))))))

;;; Tiempo:  0.05 segundos (prueba: 0.01, impresion: 0.00, otros: 0.04)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION IMPLIES)
;;;      (:DEFINITION NOT)
;;;      (:REWRITE MULTISET-SUBP-MULTISET-UNIMIN-2)
;;;      (:TYPE-PRESCRIPTION MULTISET-SUBP)

(defthm multiset-unimin-associative
  (multiset-equal (multiset-unimin m1 (multiset-unimin m2 m3))
		  (multiset-unimin (multiset-unimin m1 m2) m3))
  :hints (("Goal" :in-theory (enable another-definition-multiset-equal))))

;;; Tiempo:  0.39 segundos (prueba: 0.35, impresion: 0.00, otros: 0.04)
;;; Casos Tot: 0
;;; Usa: (:DEFINITION ANOTHER-DEFINITION-MULTISET-EQUAL)
;;;      (:EQUIVALENCE MULTISET-EQUAL-IS-AN-EQUIVALENCE)
;;;      (:FAKE-RUNE-FOR-TYPE-SET NIL)
;;;      (:REWRITE MULTISET-EQUAL-IMPLIES-SUBP)
;;;      (:REWRITE MULTISET-SUBP-MULTISET-UNIMIN)
;;;      (:REWRITE MULTISET-SUBP-MULTISET-UNIMIN-1)
;;;      (:REWRITE MULTISET-UNIMIN-CONMUTATIVE)
;;;      (:REWRITE MULTISET-UNIMIN-LEAST)

;;;============================================================================