;;;============================================================================
;;;
;;; Ttulo: Sistemas de transformacin proposicionales
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(in-package "ACL2")

;(set-verify-guards-eagerness 2)

(defpkg "MUL" (union-eq *acl2-exports*
			*common-lisp-symbols-from-main-lisp-package*))

(certify-book "../7-generico/SAT-generico" 1)

|#

(in-package "ACL2")

;;;============================================================================
;;;
;;; 0) Resumen del trabajo
;;;
;;;============================================================================

;;;   En este libro se desarrollar un algoritmo genrico para comprobar la
;;; satisfacibilidad de una frmula, basndose en un conjunto de reglas de
;;; expansin.

;;;============================================================================
;;;
;;; 1) Introduccin
;;;
;;;============================================================================

(include-book "../3-logica/tablas-de-verdad")

(in-theory (disable SAT-tablas-de-verdad MOD-tablas-de-verdad))

(include-book "../6-extensiones/generic-theory")

(include-book "../6-extensiones/defmul")

;;;============================================================================
;;;
;;; 2) Un proceso genrico de prueba de satisfacibilidad.
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Formalizacin en ACL2
;;;
;;;----------------------------------------------------------------------------

(def-generic-theory *sat-generico*

  (encapsulate 
   (((gen-objeto *) => *)
    ((gen-asignacion-distinguida * *) => *)
    ((gen-regla-computacion *) => *)
    ((gen-representacion *) => *)
    ((gen-medida *) => *)
    ((gen-modelo *) => *)
    ((gen-seleccion *) => *))
   
   (local
    (defun gen-objeto (F)
      (es-proposicional F)))
   
   (local
    (defun gen-asignacion-distinguida (sigma O)
      (modelo sigma O)))
   
   (local
    (defun gen-regla-computacion (O)
      (if (SAT-tablas-de-verdad O)
	  t
	  nil)))
   
   (local
    (defun gen-representacion (F)
      F))
   
   (local
    (defun gen-modelo (O)
      (MOD-tablas-de-verdad O)))
   
   (local
    (defun gen-medida (O)
      (len O)))
   
   (local
    (defun gen-seleccion (lista-O)
      (car lista-O)))

   (defthm gen-representacion-es-gen-objeto
     (implies (es-proposicional F)
	      (gen-objeto (gen-representacion F))))

   (defthm member-gen-regla-computacion-es-gen-objeto
     (implies (and (gen-objeto O-1)
		   (member-equal O-2 (gen-regla-computacion O-1)))
	      (gen-objeto O-2)))

   (defthm gen-medida-es-e0-ordinalp
     (e0-ordinalp (gen-medida O)))

   (defthm P1
     (implies (member-equal O-2 (gen-regla-computacion O-1))
	      (e0-ord-< (gen-medida O-2) 
			(gen-medida O-1))))

   (defthm P2
     (implies (es-proposicional F)
	      (iff (gen-asignacion-distinguida sigma (gen-representacion F))
		   (modelo sigma F))))

   (defun gen-asignacion-distinguida-lista (sigma lista-O)
     (cond ((endp lista-O) nil)
	   (t (or (gen-asignacion-distinguida sigma (car lista-O))
		  (gen-asignacion-distinguida-lista sigma (cdr lista-O))))))

   (defthm P3
     (implies (and (gen-objeto O)
		   (not (equal (gen-regla-computacion O) t)))
	      (iff (gen-asignacion-distinguida-lista 
		    sigma (gen-regla-computacion O))
		   (gen-asignacion-distinguida sigma O)))
     :hints (("Goal" :in-theory (disable modelo))))

   (defthm P4
     (implies (and (gen-objeto O)
		   (equal (gen-regla-computacion O) t))
	      (gen-asignacion-distinguida (gen-modelo O) O))
     :hints (("Goal" :in-theory (disable modelo))))

   (defthm gen-seleccion-member
     (implies (consp lista-O)
	      (member-equal (gen-seleccion lista-O) lista-O)))

   )

;;;============================================================================
;;;
;;; 3) El algoritmo genrico de satisfacibilidad
;;;
;;;============================================================================

  (export
   (defun gen-medida-lista (lista-O)
     (if (endp lista-O)
	 nil
	 (cons (gen-medida (car lista-O))
	       (gen-medida-lista (cdr lista-O)))))
   )

; (defmul-components e0-ord-<)
; => (E0-ORD-< NIL E0-ORDINALP E0-ORD-<-FN NIL NIL)

  (defmul (E0-ORD-< NIL E0-ORDINALP E0-ORD-<-FN NIL NIL))
  
  (defun elimina-una (elt lista)
    (declare (xargs :guard (true-listp lista)))
    (cond ((endp lista) lista)
	  ((equal (car lista) elt) (cdr lista))
	  (t (cons (car lista) (elimina-una elt (cdr lista))))))
  
  (local
   (include-book "../6-extensiones/listas"))

  (local
   (defthm gen-medida-lista-cons
     (equal (gen-medida-lista (cons O lista-O))
	    (cons (gen-medida O) (gen-medida-lista lista-O)))))
  
  (local
   (defthm gen-medida-lista-cons-elimina-una
     (implies (member-equal O lista-O)
	      (mul::multiset-equal
	       (gen-medida-lista lista-O)
	       (cons (gen-medida O) 
		     (gen-medida-lista (elimina-una O lista-O)))))
     :rule-classes nil))
  
  (local
   (defthm gen-medida-lista-cons-elimina-una-seleccion
     (implies (consp lista-O)
	      (mul::multiset-equal
	       (gen-medida-lista lista-O)
	       (cons (gen-medida (gen-seleccion lista-O))
		     (gen-medida-lista 
		      (elimina-una (gen-seleccion lista-O) lista-O)))))
     :hints (("Goal" :use (:instance gen-medida-lista-cons-elimina-una
				     (O (gen-seleccion lista-O)))))))
  
  (local
   (defthm gen-medida-lista-append
     (equal (gen-medida-lista (append lista-O-1 lista-O-2))
	    (append (gen-medida-lista lista-O-1) 
		    (gen-medida-lista lista-O-2)))))
  
  (local
   (defthm mul::multiset-diff-append-cons
     (mul::multiset-equal (mul::multiset-diff (append l1 l2) (cons e l2))
			  (mul::multiset-diff l1 (list e)))
     :hints (("Goal"
	      :in-theory (disable mul::multiset-diff-append-2-multiset-equal)
	      :use (:instance mul::multiset-diff-append-2-multiset-equal
			      (mul::m1 l1)
			      (mul::m2 l2)
			      (mul::m3 (list e)))))))
  
  (local
   (defthm mul::multiset-diff-cons-append
     (mul::multiset-equal (mul::multiset-diff (cons e l2) (append l1 l2))
			  (mul::multiset-diff (list e) l1))
     :hints (("Goal"
	      :in-theory (disable mul::multiset-diff-append-2-multiset-equal)
	      :use (:instance mul::multiset-diff-append-2-multiset-equal
			      (mul::m3 l1)
			      (mul::m2 l2)
			      (mul::m1 (list e)))))))
  
  (local
   (defun e0-ord-<-bigger (ol o)
     (if (endp ol)
	 t
	 (and (e0-ord-< (car ol) o)
	      (e0-ord-<-bigger (cdr ol) o)))))
  
  (local
   (defthm e0-ord-<-bigger-not-member
     (implies (e0-ord-<-bigger ol o)
	      (not (member-equal o ol)))))
  
  (local
   (defthm e0-ord-<-bigger-forall-exists-e0-ord-<-bigger
     (implies (e0-ord-<-bigger ol o)
	      (forall-exists-e0-ord-<-bigger ol (list o)))))
  
  (local
   (defthm e0-ord-<-bigger-gen-medida
     (e0-ord-<-bigger (gen-medida-lista (gen-regla-computacion O))
		      (gen-medida O))
     :hints ((extension-conjuntiva-a-listas
	      (gen-regla-computacion O)
	      (lambda (elt)
		(e0-ord-< (gen-medida elt)
			  (gen-medida O)))
	      (lambda (lst)
		(e0-ord-<-bigger (gen-medida-lista lst)
				 (gen-medida O)))))))
  
  (local
   (defthm forall-exists-e0-ord-<-bigger-gen-regla-computacion
     (forall-exists-e0-ord-<-bigger 
      (gen-medida-lista (gen-regla-computacion O))
      (list (gen-medida O)))))
  
  (local
   (defthm not-member-equal-gen-medida-gen-regla-computacion
     (not (member-equal (gen-medida O)
			(gen-medida-lista (gen-regla-computacion O))))))
  
  (export
   (defthm OBJ-SAT-generico-terminacion
     (let* ((O (gen-seleccion lista-O))
	    (resto (elimina-una (gen-seleccion lista-O) lista-O))
	    (expansion (gen-regla-computacion O)))
       (implies (and (consp lista-O)
		     (not (equal expansion t)))
		(mul-e0-ord-< (gen-medida-lista (append expansion resto))
			      (gen-medida-lista lista-O))))
     :hints (("Goal" :in-theory (disable mul::multiset-diff))))
   )

  (export
   (defthm e0-ordinalp-true-listp-gen-medida-lista
     (e0-ordinalp-true-listp (gen-medida-lista lista-O)))
   )

  (local
   (in-theory (disable gen-medida-lista-append
		       gen-medida-lista-cons-elimina-una-seleccion)))

  (export
   (defun OBJ-SAT-generico (lista-O)
     (declare (xargs :measure (gen-medida-lista lista-O)
		     :well-founded-relation mul-e0-ord-<
		     :hints (("Goal" :in-theory 
			      (disable elimina-una mul-e0-ord-<)))))
     (if (endp lista-O)
	 nil
	 (let* ((O (gen-seleccion lista-O))
		(resto (elimina-una (gen-seleccion lista-O) lista-O))
		(expansion (gen-regla-computacion O)))
	   (cond ((equal expansion t)
		  (list O))
		 (t (OBJ-SAT-generico (append expansion resto)))))))
   )
  
  (export
   (defun SAT-generico (F)
     (OBJ-SAT-generico (list (gen-representacion F))))
   )

  (export
   (defun DAT-generico (F)
     (not (SAT-generico (negacion F))))
   )

  (export
   (defun mod-generico (F)
     (gen-modelo (car (SAT-generico F))))
   )

;;;============================================================================
;;;
;;; 4) Resultados de correccin y completitud
;;;
;;;============================================================================

  (export
   (defun gen-objeto-lista (lista-O)
     (cond ((endp lista-O) t)
	   (t (and (gen-objeto (car lista-O))
		   (gen-objeto-lista (cdr lista-O))))))
   )

  (local
   (defthm gen-objeto-lista-append 
     (iff (gen-objeto-lista (append lista-O-1 lista-O-2))
	  (and (gen-objeto-lista lista-O-1)
	       (gen-objeto-lista lista-O-2)))))
  
  (local
   (defthm gen-objeto-lista-elimina-una
     (implies (gen-objeto-lista lista-O)
	      (gen-objeto-lista (elimina-una O lista-O)))))

  (local
   (defthm gen-objeto-lista-member-equal
     (implies (and (gen-objeto-lista lista-O)
		   (member-equal O lista-O))
	      (gen-objeto O))))

  (local
   (defthm gen-objeto-gen-seleccion
     (implies (and (consp lista-O)
		   (gen-objeto-lista lista-O))
	      (gen-objeto (gen-seleccion lista-O)))))
 
  (local
   (defthm gen-objeto-lista-gen-regla-computacion
     (implies (gen-objeto O)
	      (gen-objeto-lista (gen-regla-computacion O)))
     :hints ((extension-conjuntiva-a-listas
	      (gen-regla-computacion O)
	      gen-objeto
	      gen-objeto-lista))))
 
  (local
   (defthm gen-asignacion-distinguida-lista-append
     (iff (gen-asignacion-distinguida-lista sigma (append lista-O-1 lista-O-2))
	  (or (gen-asignacion-distinguida-lista sigma lista-O-1)
	      (gen-asignacion-distinguida-lista sigma lista-O-2)))))

  (local
   (defthm gen-asignacion-distinguida-lista-elimina-una-1
     (implies (and (gen-asignacion-distinguida-lista sigma lista-O)
		   (not (gen-asignacion-distinguida sigma O)))
	      (gen-asignacion-distinguida-lista sigma 
						(elimina-una O lista-O)))))

  (local
   (defthm gen-asignacion-distinguida-lista-elimina-una-2
     (implies (gen-asignacion-distinguida-lista sigma (elimina-una O lista-O))
	      (gen-asignacion-distinguida-lista sigma lista-O))))
 
  (local
   (defthm gen-asignacion-distinguida-lista-member-equal
     (implies (and (gen-asignacion-distinguida sigma O)
		   (member-equal O lista-O))
	      (gen-asignacion-distinguida-lista sigma lista-O))))

  (local
   (defthm gen-asignacion-distinguida-lista-gen-seleccion
     (implies (and (gen-asignacion-distinguida sigma (gen-seleccion lista-O))
		   (consp lista-O))
	      (gen-asignacion-distinguida-lista sigma lista-O))))

;;;----------------------------------------------------------------------------
;;;
;;; 4.1) Teoremas de correccin y completitud de "OBJ-SAT-generico"
;;;
;;;----------------------------------------------------------------------------

  (export
   (defthm correccion-OBJ-SAT-generico
     (implies (and (gen-objeto-lista lista-O)
		   (OBJ-SAT-generico lista-O))
	      (gen-asignacion-distinguida-lista 
	       (gen-modelo (car (OBJ-SAT-generico lista-O)))
	       lista-O))
     :rule-classes nil)
   )

  (export
   (defthm completitud-OBJ-SAT-generico
     (implies (and (gen-objeto-lista lista-O)
		   (gen-asignacion-distinguida-lista sigma lista-O))
	      (OBJ-SAT-generico lista-O))
     :rule-classes nil)
   )

;;;----------------------------------------------------------------------------
;;;
;;; 4.2) Teoremas de correccin y completitud de "SAT-generico"
;;;
;;;----------------------------------------------------------------------------

  (export
   (defthm correccion-SAT-generico
     (implies (and (es-proposicional F)
		   (SAT-generico F))
	      (modelo (mod-generico F) F))
     :hints (("Goal"
	      :use (:instance correccion-OBJ-SAT-generico
			      (lista-O (list (gen-representacion F)))))))
   )

  (export
   (defthm completitud-SAT-generico
     (implies (and (es-proposicional F)
		   (modelo sigma F))
	      (SAT-generico F))
     :hints (("Goal"
	      :use (:instance completitud-OBJ-SAT-generico
			      (lista-O (list (gen-representacion F)))))))
   )

;;;----------------------------------------------------------------------------
;;;
;;; 4.3) Teoremas de correccin y completitud de "DAT-generico"
;;;
;;;----------------------------------------------------------------------------

  (export
   (defthm correccion-DAT-generico
     (implies (and (es-proposicional F)
		   (DAT-generico F))
	      (modelo sigma F))
     :hints (("Goal"
	      :use (:instance completitud-OBJ-SAT-generico
			      (lista-O (list (gen-representacion 
					      (negacion F))))))))
   )

  (export
   (defthm completitud-DAT-generico
     (implies (and (es-proposicional F)
		   (not (DAT-generico F)))
	      (not (modelo (mod-generico (negacion F)) F)))
     :hints (("Goal"
	      :use (:instance correccion-OBJ-SAT-generico
			      (lista-O (list (gen-representacion 
					      (negacion F))))))))
   )

  )

;;;============================================================================