;;;============================================================================
;;;
;;; Ttulo: Un STP exitoso basado en tableros semnticos
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(in-package "ACL2")

(certify-book "../7-generico/SAT-tableros-K")

|#

(in-package "ACL2")

(include-book "../7-generico/SAT-generico-K")

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

;;;   SAT proposicional basado en reglas asociadas a tableros semnticos
;;; generado como una instancia del modelo genrico desarrollado en
;;; "SAT-generico".

;;;============================================================================
;;;
;;; 1) STP exitoso basado en tableros semnticos (I)
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Objetos proposicionales (tab-objeto)
;;;
;;;----------------------------------------------------------------------------

(defun tab-objeto (rama) ; = es-rama-tablero
  (or (endp rama)
      (and (es-proposicional (car rama))
	   (tab-objeto (cdr rama)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Asignaciones distinguidas (tab-asignacion-distinguida)
;;;
;;;----------------------------------------------------------------------------

(defun tab-asignacion-distinguida-K (sigma rama) ; = modelo-rama
  (cond ((endp rama) t)
	(t (and (modelo-K sigma (car rama))
		(tab-asignacion-distinguida-K sigma (cdr rama))))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Regla de computacin (tab-regla-computacion-1)
;;;
;;;----------------------------------------------------------------------------

(defun rama-cerrada (rama)
  (cond ((endp rama) nil)
	((member-equal (complementario (car rama)) (cdr rama)) t)
	(t (rama-cerrada (cdr rama)))))

(include-book "../4-tableros/uniforme-K")

(encapsulate
 (((seleccion *) => *))

 (local
  (defun seleccion (rama)
    (declare (xargs :guard (true-listp rama)))
    (cond ((endp rama) nil)
	  ((es-literal (car rama)) (seleccion (cdr rama)))
	  (t (car rama)))))

 (local (in-theory (disable es-atomica (:type-prescription es-atomica)
			    es-literal (:type-prescription es-negacion)
			    (:type-prescription es-literal)
			    unicidad-es-negacion)))

 (defthm si-existe-no-literal-seleccion-pertenece
   (implies (seleccion rama)
	    (member-equal (seleccion rama) rama)))
 
 (defthm si-existe-no-literal-seleccion-no-es-literal
   (implies (seleccion rama)
	    (not (es-literal (seleccion rama)))))
 
 (defthm si-no-seleccion-miembros-literales
   (implies (and (not (seleccion rama))
		 (tab-objeto rama)
		 (member-equal F rama))
	    (es-literal F)))

 )

;;;   Funcin para aadir un F a una lista, si es que no estaba:

(defun aade-elemento (F lista)
  (if (member-equal F lista)
      lista
      (cons F lista)))

;;;   Finalmente podemos construir la funcin de expansin basada en las reglas
;;; anteriores:

(defun tab-regla-computacion-1 (rama)
  (let ((F (seleccion rama)))
    (cond ((doble-negacion F)
	   (list (aade-elemento (componente-neg-neg F)
				 (elimina-una F rama))))
	  ((alfa-formula F)
	   (list (aade-elemento (componente-1 F)
				 (aade-elemento (componente-2 F)
						 (elimina-una F rama)))))
	  ((beta-formula F)
	   (list (aade-elemento (componente-1 F)
				 (elimina-una F rama))
		 (aade-elemento (componente-2 F)
				 (elimina-una F rama))))
	  ((rama-cerrada rama)
	   nil)
	  (t t))))

(local
 (defthm tab-objeto-elimina-una
   (implies (tab-objeto l)
	    (tab-objeto (elimina-una x l)))))

(local
 (defthm tab-objeto-aade-elemento
   (implies (and (tab-objeto rama)
		 (es-proposicional F))
	    (tab-objeto (aade-elemento F rama)))))

(defthm member-tab-regla-computacion-1-es-tab-objeto
  (implies (and (tab-objeto rama-1)
		(member-equal rama-2 (tab-regla-computacion-1 rama-1)))
	   (tab-objeto rama-2))
  :hints (("Goal"
	   :in-theory (disable definicion-uniforme-de-es-proposicional
			       aade-elemento elimina-una))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Funcin de representacin (tab-representacion)
;;;
;;;----------------------------------------------------------------------------

(defun tab-representacion (F)
  (list F))

(defthm tab-representacion-es-tab-objeto
  (implies (es-proposicional F)
	   (tab-objeto (tab-representacion F)))
  :hints (("Goal"
	   :in-theory (disable definiciones-conectivas))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Funcin de medida (tab-medida)
;;;
;;;----------------------------------------------------------------------------

(defun tab-medida (rama) ; = medida-uniforme-rama
  (cond ((endp rama) 0)
	(t (+ (medida-uniforme (car rama))
	      (tab-medida (cdr rama))))))

(defthm tab-medida-es-e0-ordinalp
  (e0-ordinalp (tab-medida rama)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.6) Funcin modelo (tab-modelo)
;;;
;;;----------------------------------------------------------------------------

(defun tab-modelo-K (lista-L) ; = genera-modelo-literales
  (cond ((endp lista-L) nil)
	((es-literal-positivo (car lista-L))
	 (asume-valor (car lista-L) *V*
		      (tab-modelo-K (cdr lista-L))))
        ((es-literal-negativo (car lista-L))
         (asume-valor (arg1 (car lista-L)) *F*
                      (tab-modelo-K (cdr lista-L))))
	(t (tab-modelo-K (cdr lista-L)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.7) Propiedad P1
;;;
;;;----------------------------------------------------------------------------

(in-theory (disable alfa-formula beta-formula doble-negacion
		    componente-1 componente-2 componente-neg-neg))

(defthm tab-regla-computacion-1-seleccion-existe
  (implies (consp (tab-regla-computacion-1 rama))
	   (seleccion rama)))

(local
 (defthm tab-medida-aux-elimina-una-formula
   (implies (member-equal F rama)
	    (= (tab-medida rama)
	       (+ (medida-uniforme F)
		  (tab-medida (elimina-una F rama)))))
   :rule-classes :linear))

(defthm P1-tab
  (implies (member-equal rama-2 (tab-regla-computacion-1 rama-1))
	   (e0-ord-< (tab-medida rama-2)
		     (tab-medida rama-1))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.8) Propiedad P2
;;;
;;;----------------------------------------------------------------------------

(defthm P2-tab
  (implies (es-proposicional F)
	   (iff (tab-asignacion-distinguida-K sigma (tab-representacion F))
		(modelo-K sigma F))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.9) Propiedad P3
;;;
;;;----------------------------------------------------------------------------

(local
 (defthm modelo-K-not-modelo-K-complementario
   (implies (modelo-K sigma (complementario F))
	    (not (modelo-K sigma F)))
   :hints (("Goal" 
	    :in-theory (disable definicion-uniforme-de-es-proposicional)))))

(encapsulate
 ()

 (local
  (defthm rama-cerrada-no-tiene-modelos-aux1
    (implies (and (not (modelo-K sigma F))
		  (member-equal F rama))
	     (not (tab-asignacion-distinguida-K sigma rama)))))

 (local
  (defthm rama-cerrada-no-tiene-modelos-aux2
    (implies (and (modelo-K sigma F)
		  (member-equal (complementario F) rama))
	     (not (tab-asignacion-distinguida-K sigma rama)))
    :hints (("Goal" :in-theory (disable complementario modelo-K)))))

 (local
  (defthm tiene-formulas-complementarios-not-tab-asignacion-distinguida-K
    (implies (rama-cerrada rama)
	     (not (tab-asignacion-distinguida-K sigma rama)))
    :hints (("Goal" :in-theory (disable complementario modelo-K)))))

 (defthm rama-cerrada-no-tiene-modelos ; = rama-cerrada-no-tiene-modelos
   (implies (rama-cerrada rama)
	    (not (tab-asignacion-distinguida-K sigma rama))))

 )

(local
 (defthm tab-asignacion-distinguida-K-aade-elemento
   (iff (tab-asignacion-distinguida-K sigma (aade-elemento F rama))
	(and (modelo-K sigma F)
	     (tab-asignacion-distinguida-K sigma rama)))))

(local
 (in-theory (disable aade-elemento)))
  
;;   Las reglas de expansin preservan el conjunto de modelos: 

(local
 (defthm regla-doble-negacion-preserva-modelos-aux
   (implies (and (member-equal F lista)
		 (doble-negacion F))
	    (iff (tab-asignacion-distinguida-K
		  sigma (aade-elemento (componente-neg-neg F)
					(elimina-una F lista)))
		 (tab-asignacion-distinguida-K sigma lista)))))

(local
 (defthm regla-formula-alfa-preserva-modelos-aux
   (implies (and (member-equal F rama)
		 (alfa-formula F))
	    (iff (tab-asignacion-distinguida-K
		  sigma (aade-elemento (componente-1 F)
					(aade-elemento (componente-2 F)
							(elimina-una F rama))))
		 (tab-asignacion-distinguida-K sigma rama)))))

(local
 (defthm regla-formula-beta-preserva-modelos-aux
   (implies (and (beta-formula F)
		 (member-equal F rama))
	    (iff (tab-asignacion-distinguida-K sigma rama)
		 (or (tab-asignacion-distinguida-K
		      sigma (aade-elemento (componente-1 F)
					    (elimina-una F rama)))
		     (tab-asignacion-distinguida-K
		      sigma (aade-elemento (componente-2 F)
					    (elimina-una F rama))))))))

;;;   La regla de computacin preserva el conjunto de modelos:

(defun tab-asignacion-distinguida-K-lista (sigma tablero)
  (cond ((atom tablero) nil)
	(t (or (tab-asignacion-distinguida-K sigma (car tablero))
	       (tab-asignacion-distinguida-K-lista sigma (cdr tablero))))))

(defthm P3-tab
  (implies (and (tab-objeto rama)
		(not (equal (tab-regla-computacion-1 rama) t)))
	   (iff (tab-asignacion-distinguida-K-lista 
		 sigma (tab-regla-computacion-1 rama))
		(tab-asignacion-distinguida-K sigma rama)))
  :hints (("Goal" :in-theory (disable modelo-K))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.10) Propiedad P4
;;;
;;;----------------------------------------------------------------------------

(local
 (defthm tab-modelo-K-es-modelo-literales-rama-cerrada
   (implies (and (es-literal L)
		 (not (rama-cerrada rama))
		 (member-equal L rama))
	    (modelo-K (tab-modelo-K rama) L))
   :hints (("Goal"
	    :in-theory (enable es-valor-de-verdad)
	    :induct (member-equal L rama)))))

(local
 (defthm tab-modelo-K-es-modelo-literales-rama-expansion-1-t
   (implies (and (equal (tab-regla-computacion-1 rama) t)
		 (es-literal L)
		 (member-equal L rama))
	    (modelo-K (tab-modelo-K rama) L))
   :hints (("Goal"
	    :in-theory (disable modelo-K)))))
 
(local
 (defthm tab-objeto-es-proposicional
   (implies (and (tab-objeto rama)
		 (member-equal F rama))
	    (es-proposicional F))))

(local
 (defthm tab-objeto-seleccion-es-proposicional
   (implies (and (tab-objeto rama)
		 (seleccion rama))
	    (es-proposicional (seleccion rama)))
   :rule-classes nil))

(local
 (in-theory (disable tab-objeto-es-proposicional)))

(local
 (defthm tab-objeto-expansion-1-t->not-seleccion
   (implies (and (equal (tab-regla-computacion-1 rama) t)
		 (tab-objeto rama))
	    (not (seleccion rama)))
   :hints (("Goal"
	    :use tab-objeto-seleccion-es-proposicional))))

(local
 (defthm member-equal-lista-de-literales-es-literal-1
   (implies (and (tab-objeto rama)
		 (equal (tab-regla-computacion-1 rama) t)
		 (member-equal F rama))
	    (es-literal F))
   :hints (("Goal"
	    :use si-no-seleccion-miembros-literales
	    :in-theory (disable es-literal
				si-no-seleccion-miembros-literales)))))

(local 
 (in-theory (disable tab-regla-computacion-1 modelo-K)))

(include-book "../6-extensiones/listas")

(defthm P4-tab
  (implies (and (tab-objeto rama)
		(equal (tab-regla-computacion-1 rama) t))
	   (tab-asignacion-distinguida-K (tab-modelo-K rama) rama))
  :hints ((extension-conjuntiva-a-listas
	   rama
	   (lambda (x)
	     (modelo-K (tab-modelo-K rama) x))
	   (lambda (lst)
	     (tab-asignacion-distinguida-K (tab-modelo-K rama) lst)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.11) Instancia
;;;
;;;----------------------------------------------------------------------------

;;;   A continuacin instanciamos la teora (definiciones y teoremas) del
;;; demostrador sat genrico con la siguiente correspondencia entre las
;;; funciones:

;;; gen-objeto                       -- tab-objeto
;;; gen-representacion               -- tab-representacion
;;; gen-asignacion-distinguida       -- tab-asignacion-distinguida-K
;;; gen-asignacion-distinguida-lista -- tab-asignacion-distinguida-K-lista
;;; gen-regla-computacion            -- tab-regla-computacion-1
;;; gen-seleccion                    -- car
;;; gen-medida                       -- tab-medida
;;; gen-modelo                       -- tab-modelo-K

;;;   Falta por demostrar la propiedad asociada a la funcin obj-seleccion,
;;; que aplicada a un lista de ramas no vaca devuelva un elemento de dicha
;;; lista.  No es necesario establecerla explcitamente dado la simplicidad de
;;; la funcin que usamos en su lugar.

(in-theory (disable tab-regla-computacion-1
		    (:type-prescription TAB-REGLA-COMPUTACION-1)
		    tab-representacion))

(definstance-*sat-generico*
  ((gen-objeto                       tab-objeto)
   (gen-representacion               tab-representacion)
   (gen-asignacion-distinguida       tab-asignacion-distinguida-K)
   (gen-asignacion-distinguida-lista tab-asignacion-distinguida-K-lista)
   (gen-regla-computacion            tab-regla-computacion-1)
   (gen-seleccion                    car)
   (gen-medida                       tab-medida)
   (gen-modelo                       tab-modelo-K))
  "-tableros-1")

;;;============================================================================
;;;
;;; 2) STP exitoso basado en tableros semnticos (II)
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Regla de computacion (tab-regla-computacion-2)
;;;
;;;----------------------------------------------------------------------------

(defun tab-regla-computacion-2 (rama)
  (if (rama-cerrada rama)
      nil
      (let ((F (seleccion rama)))
	(cond ((doble-negacion F)
	       (list (aade-elemento (componente-neg-neg F)
				     (elimina-una F rama))))
	      ((alfa-formula F)
	       (list (aade-elemento 
		      (componente-1 F)
		      (aade-elemento (componente-2 F)
				      (elimina-una F rama)))))
	      ((beta-formula F)
	       (list (aade-elemento (componente-1 F)
				     (elimina-una F rama))
		     (aade-elemento (componente-2 F)
				     (elimina-una F rama))))
	      (t t)))))

(defthm member-tab-regla-computacion-2-tab-objeto
  (implies (and (tab-objeto rama-1)
		(member-equal rama-2 (tab-regla-computacion-2 rama-1)))
	   (tab-objeto rama-2))
  :hints (("Goal"
	   :in-theory (disable definicion-uniforme-de-es-proposicional))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) Propiedad P1
;;;
;;;----------------------------------------------------------------------------

(defthm tab-regla-computacion-2-seleccion-existe
  (implies (consp (tab-regla-computacion-2 rama))
	   (seleccion rama)))

(defthm P1-tab-2
  (implies (member-equal rama-2 (tab-regla-computacion-2 rama-1))
	   (e0-ord-< (tab-medida rama-2)
		     (tab-medida rama-1)))
  :hints (("Goal"
	   :in-theory (enable aade-elemento))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.3) Propiedad P3
;;;
;;;----------------------------------------------------------------------------

(defthm P3-tab-2
  (implies (and (tab-objeto rama)
		(not (equal (tab-regla-computacion-2 rama) t)))
	   (iff (tab-asignacion-distinguida-K-lista 
		 sigma (tab-regla-computacion-2 rama))
		(tab-asignacion-distinguida-K sigma rama)))
  :hints (("Goal" 
	   :cases ((rama-cerrada rama)))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.4) Propiedad P4
;;;
;;;----------------------------------------------------------------------------

(local
 (defthm tab-modelo-K-es-modelo-literales-rama-expansion-2-t
   (implies (and (equal (tab-regla-computacion-2 rama) t)
		 (es-literal L)
		 (member-equal L rama))
	    (modelo-K (tab-modelo-K rama) L))))

(local
 (defthm tab-objeto-expansion-2-t->not-seleccion
   (implies (and (equal (tab-regla-computacion-2 rama) t)
		 (tab-objeto rama))
	    (not (seleccion rama)))
   :hints (("Goal"
	    :use tab-objeto-seleccion-es-proposicional))))

(local
 (defthm member-equal-lista-de-literales-es-literal-2
   (implies (and (tab-objeto rama)
		 (equal (tab-regla-computacion-2 rama) t)
		 (member-equal F rama))
	    (es-literal F))
   :hints (("Goal"
	    :use si-no-seleccion-miembros-literales
	    :in-theory (disable es-literal
				si-no-seleccion-miembros-literales)))))

(defthm P4-tab-2
  (implies (and (tab-objeto rama)
		(equal (tab-regla-computacion-2 rama) t))
	   (tab-asignacion-distinguida-K (tab-modelo-K rama) rama))
  :hints ((extension-conjuntiva-a-listas
	   rama
	   (lambda (elt)
	     (modelo-K (tab-modelo-K rama) elt))
	   (lambda (lst)
	     (tab-asignacion-distinguida-K (tab-modelo-K rama) lst)))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.5) Instancia
;;;
;;;----------------------------------------------------------------------------

;;;   A continuacin instanciamos la teora (definiciones y teoremas) del
;;; demostrador sat genrico con la siguiente correspondencia entre las
;;; funciones:

;;; gen-objeto                       -- tab-objeto
;;; gen-representacion               -- tab-representacion
;;; gen-asignacion-distinguida       -- tab-asignacion-distinguida-K
;;; gen-asignacion-distinguida-lista -- tab-asignacion-distinguida-K-lista
;;; gen-regla-computacion            -- tab-regla-computacion-2
;;; gen-seleccion                    -- car
;;; gen-medida                       -- tab-medida
;;; gen-modelo                       -- tab-modelo-K

;;;   Falta por demostrar la propiedad asociada a la funcin obj-seleccion,
;;; que aplicada a un lista de ramas no vaca devuelva un elemento de dicha
;;; lista.  No es necesario establecerla explcitamente dado la simplicidad de
;;; la funcin que usamos en su lugar.

(in-theory (disable tab-regla-computacion-2
		    (:type-prescription TAB-REGLA-COMPUTACION-2)))

(definstance-*sat-generico*
  ((gen-objeto                       tab-objeto)
   (gen-representacion               tab-representacion)
   (gen-asignacion-distinguida       tab-asignacion-distinguida-K)
   (gen-asignacion-distinguida-lista tab-asignacion-distinguida-K-lista)
   (gen-regla-computacion            tab-regla-computacion-2)
   (gen-seleccion                    car)
   (gen-medida                       tab-medida)
   (gen-modelo                       tab-modelo-K))
  "-tableros-2")

;;;============================================================================
;;;
;;; 3) STP exitoso basado en tableros semnticos (III)
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Objetos proposicionales (tab-objeto-3)
;;;
;;;----------------------------------------------------------------------------

(defun tab-objeto-3 (rama)
  (and (tab-objeto rama)
       (not (rama-cerrada rama))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.2) Regla de computacin (tab-regla-computacion-3)
;;;
;;;----------------------------------------------------------------------------

(defun aade-no-comp (F rama)
  (cond ((member-equal (complementario F) rama) nil)
	(t (list (aade-elemento F rama)))))

(defun aade-no-comp2 (F-1 F-2 rama)
  (cond ((member-equal (complementario F-2) rama) nil)
	(t (aade-no-comp F-1 (aade-elemento F-2 rama)))))

(defun tab-regla-computacion-3 (rama)
  (let* ((F (seleccion rama))
	 (n-rama (elimina-una F rama)))
    (cond ((doble-negacion F)
	   (aade-no-comp (componente-neg-neg F) n-rama))
	  ((alfa-formula F)
	   (aade-no-comp2 (componente-1 F)
			   (componente-2 F) n-rama))
	  ((beta-formula F)
	   (append (aade-no-comp (componente-1 F) n-rama)
		   (aade-no-comp (componente-2 F) n-rama)))
	  (t t))))

(defthm member-equal-elimina-una
  (implies (member-equal F-1 (elimina-una F-2 rama))
	   (member-equal F-1 rama)))

(defthm rama-cerrada-elimina-una
  (implies (not (rama-cerrada rama))
	   (not (rama-cerrada (elimina-una F rama)))))

; (in-theory (enable aade-elemento))

(defthm rama-cerrada-aade-no-comp
  (implies (and (member-equal rama-2 (aade-no-comp F rama-1))
		(not (rama-cerrada rama-1)))
	   (not (rama-cerrada rama-2)))
  :hints (("Goal" :in-theory (enable aade-elemento))))

(defthm rama-cerrada-aade-no-comp2
  (implies (and (member-equal rama-2 (aade-no-comp2 F-1 F-2 rama-1))
		(not (rama-cerrada rama-1)))
	   (not (rama-cerrada rama-2)))
  :hints (("Goal" :in-theory (enable aade-elemento))))

(defthm tab-objeto-aade-no-comp
  (implies (and (member-equal rama-2 (aade-no-comp F rama-1))
		(tab-objeto rama-1)
		(es-proposicional F))
	   (tab-objeto rama-2)))

(defthm tab-objeto-aade-no-comp2
  (implies (and (member-equal rama-2 (aade-no-comp2 F-1 F-2 rama-1))
		(tab-objeto rama-1)
		(es-proposicional F-1)
		(es-proposicional F-2))
	   (tab-objeto rama-2))
  :hints (("Goal" 
	   :in-theory (disable complementario
			       definicion-uniforme-de-es-proposicional))))

(defthm member-equal-append
  (iff (member-equal elt (append lst-1 lst-2))
       (or (member-equal elt lst-1)
	   (member-equal elt lst-2))))

(defthm member-tab-regla-computacion-3-es-tab-objeto-3
  (implies (and (tab-objeto-3 rama-1)
		(member-equal rama-2 (tab-regla-computacion-3 rama-1)))
	   (tab-objeto-3 rama-2))
  :hints (("Goal"
	   :in-theory (disable aade-no-comp aade-no-comp2
			       definicion-uniforme-de-es-proposicional))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.3) Funcin de representacin (tab-representacion)
;;;
;;;----------------------------------------------------------------------------

(defthm tab-representacion-es-tab-objeto-3
  (implies (es-proposicional F)
	   (tab-objeto-3 (tab-representacion F)))
  :hints (("Goal"
	   :in-theory (enable rama-cerrada tab-representacion))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.4) Propiedad P1
;;;
;;;----------------------------------------------------------------------------

(defthm tab-medida-aade-no-comp
  (implies (member-equal rama-1 (aade-no-comp f rama-2))
	   (<= (tab-medida rama-1)
	       (+ (medida-uniforme f)
		  (tab-medida rama-2))))
  :hints (("Goal" :in-theory (enable aade-elemento)))
  :rule-classes :linear)

(defthm tab-medida-aade-no-comp2-2
  (implies (member-equal rama-1 (aade-no-comp2 f1 f2 rama-2))
	   (<= (tab-medida rama-1)
	       (+ (medida-uniforme f1)
		  (medida-uniforme f2)
		  (tab-medida rama-2))))
  :hints (("Goal" :in-theory (enable aade-elemento)))  
  :rule-classes :linear)

(local
 (defthm doble-negacion-seleccion->seleccion
   (implies (doble-negacion (seleccion lista))
	    (seleccion lista))))

(local
 (defthm alfa-formula-seleccion->seleccion
   (implies (alfa-formula (seleccion rama))
	    (seleccion rama))))

(local
 (defthm beta-formula-seleccion->seleccion
   (implies (beta-formula (seleccion rama))
	    (seleccion rama))))

(defthm P1-tab-3
  (implies (member-equal rama-2 (tab-regla-computacion-3 rama-1))
	   (e0-ord-< (tab-medida rama-2)
		     (tab-medida rama-1)))
  :hints (("Goal"
  	   :in-theory (disable aade-no-comp aade-no-comp2))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.5) Propiedad P3
;;;
;;;----------------------------------------------------------------------------

(local
 (defthm tab-asignacion-distinguida-K-member
   (implies (and (tab-asignacion-distinguida-K sigma rama)
		 (member-equal F rama))
	    (modelo-K sigma F))))

(local
 (defthm tab-asignacion-distinguida-K-member-complementario
   (implies (and (tab-asignacion-distinguida-K sigma rama)
		 (member-equal (complementario F) rama))
	    (not (modelo-K sigma F)))))

(local
 (defthm tab-asignacion-distinguida-K-lista-aade-no-comp
   (iff (tab-asignacion-distinguida-K-lista sigma (aade-no-comp F rama-1))
	(and (modelo-K sigma F)
	     (tab-asignacion-distinguida-K sigma rama-1)))
   :hints (("Goal" :in-theory (disable complementario)))))

(local
 (defthm tab-asignacion-distinguida-K-lista-aade-no-comp2
   (iff (tab-asignacion-distinguida-K-lista 
	 sigma (aade-no-comp2 F-1 F-2 rama-1))
	(and (modelo-K sigma F-1)
	     (modelo-K sigma F-2)
	     (tab-asignacion-distinguida-K sigma rama-1)))
   :hints (("Goal" :in-theory (disable aade-no-comp complementario)))))

(local
 (defthm tab-asignacion-distinguida-K-elimina-una-1
   (implies (tab-asignacion-distinguida-K sigma lista)
	    (tab-asignacion-distinguida-K sigma (elimina-una F lista)))))

(local
 (defthm tab-asignacion-distinguida-K-elimina-una-2
   (implies (and (tab-asignacion-distinguida-K sigma (elimina-una F lista))
		 (modelo-K sigma F))
	    (tab-asignacion-distinguida-K sigma lista))))
 
;;   La regla de expansin para dobles negaciones preserva el conjunto de
;; modelos: 

(local
 (in-theory (disable aade-no-comp aade-no-comp2)))

(local
 (defthm regla-doble-negacion-3-preserva-modelos-aux
   (implies (and (member-equal F lista)
		 (doble-negacion F))
	    (iff (tab-asignacion-distinguida-K-lista 
		  sigma (aade-no-comp (componente-neg-neg F)
				       (elimina-una F lista)))
		 (tab-asignacion-distinguida-K sigma lista)))
   :hints (("Goal" :in-theory (enable modelo-K)))))

(local
 (defthm regla-formula-alfa-3-preserva-modelos-aux
   (implies (and (member-equal F rama)
		 (alfa-formula F))
	    (iff (tab-asignacion-distinguida-K-lista
		  sigma (aade-no-comp2 (componente-1 F)
					(componente-2 F)
					(elimina-una F rama)))
		 (tab-asignacion-distinguida-K sigma rama)))
   :hints (("Goal" :in-theory (enable modelo-K)))))

(local
 (defthm regla-formula-beta-3-preserva-modelos-aux
   (implies (and (beta-formula F)
		 (member-equal F rama))
	    (iff (tab-asignacion-distinguida-K sigma rama)
		 (or (tab-asignacion-distinguida-K-lista
		      sigma (aade-no-comp (componente-1 F)
					   (elimina-una F rama)))
		     (tab-asignacion-distinguida-K-lista
		      sigma (aade-no-comp (componente-2 F)
					   (elimina-una F rama))))))
   :hints (("Goal" :in-theory (enable modelo-K)))))

(defthm tab-asignacion-distinguida-K-lista-append
  (iff (tab-asignacion-distinguida-K-lista sigma (append lista-1 lista-2))
       (or (tab-asignacion-distinguida-K-lista sigma lista-1)
	   (tab-asignacion-distinguida-K-lista sigma lista-2))))

(local
 (in-theory
  (disable tab-asignacion-distinguida-K-lista-aade-no-comp
	   tab-asignacion-distinguida-K-lista-aade-no-comp2)))

(defthm P3-tab-3
  (implies (and (tab-objeto-3 rama)
		(not (equal (tab-regla-computacion-3 rama) t)))
	   (iff (tab-asignacion-distinguida-K-lista
		 sigma (tab-regla-computacion-3 rama))
		(tab-asignacion-distinguida-K sigma rama)))
  :hints (("Goal" 
 	   :in-theory (DISABLE COMPLEMENTARIO))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.6) Propiedad P4
;;;
;;;----------------------------------------------------------------------------

(local
 (defthm tab-modelo-K-es-modelo-literales-rama-expansion-3-t
   (implies (and (equal (tab-regla-computacion-3 rama) t)
		 (tab-objeto-3 rama)
		 (es-literal L)
		 (member-equal L rama))
	    (modelo-K (tab-modelo-K rama) L))))

(local
 (defthm tab-objeto-expansion-3-t->not-seleccion
   (implies (and (equal (tab-regla-computacion-3 rama) t)
		 (tab-objeto-3 rama))
	    (not (seleccion rama)))
   :hints (("Goal"
	    :use tab-objeto-seleccion-es-proposicional))))

(local
 (defthm member-equal-tab-objeto-3-es-literal-2
   (implies (and (equal (tab-regla-computacion-3 rama) t)
		 (tab-objeto-3 rama)
		 (member-equal F rama))
	    (es-literal F))
   :hints (("Goal"
	    :use si-no-seleccion-miembros-literales
	    :in-theory (disable es-literal
				si-no-seleccion-miembros-literales)))))

(in-theory (disable tab-regla-computacion-3
		    (:type-prescription TAB-REGLA-COMPUTACION-3)
		    tab-objeto-3))

(defthm P4-tab-3
  (implies (and (tab-objeto-3 rama)
		(equal (tab-regla-computacion-3 rama) t))
	   (tab-asignacion-distinguida-K (tab-modelo-K rama) rama))
  :hints ((extension-conjuntiva-a-listas
	   rama
 	   (lambda (x)
 	     (modelo-K (tab-modelo-K rama) x))
 	   (lambda (lst)
 	     (tab-asignacion-distinguida-K (tab-modelo-K rama) lst)))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.7) Instancia
;;;
;;;----------------------------------------------------------------------------

;;;   A continuacin instanciamos la teora (definiciones y teoremas) del
;;; demostrador sat genrico con la siguiente correspondencia entre las
;;; funciones:

;;; gen-objeto                       -- tab-objeto-3
;;; gen-representacion               -- tab-representacion
;;; gen-asignacion-distinguida       -- tab-asignacion-distinguida-K
;;; gen-asignacion-distinguida-lista -- tab-asignacion-distinguida-K-lista
;;; gen-regla-computacion            -- tab-regla-computacion-3
;;; gen-seleccion                    -- car
;;; gen-medida                       -- tab-medida
;;; gen-modelo                       -- tab-modelo-K

;;;   Falta por demostrar la propiedad asociada a la funcin obj-seleccion,
;;; que aplicada a un lista de ramas no vaca devuelva un elemento de dicha
;;; lista.  No es necesario establecerla explcitamente dado la simplicidad de
;;; la funcin que usamos en su lugar.

(definstance-*sat-generico*
  ((gen-objeto                       tab-objeto-3)
   (gen-representacion               tab-representacion)
   (gen-asignacion-distinguida       tab-asignacion-distinguida-K)
   (gen-asignacion-distinguida-lista tab-asignacion-distinguida-K-lista)
   (gen-regla-computacion            tab-regla-computacion-3)
   (gen-seleccion                    car)
   (gen-medida                       tab-medida)
   (gen-modelo                       tab-modelo-K))
  "-tableros-3")

;;;============================================================================
;;;
;;; 10) Ejemplos de uso
;;;
;;;============================================================================

#|

;;;   Un ejemplo de utilizacin de este ATP es el siguiente:

(include-book "SAT-tableros")

;;;   Necesitamos una funcin de seleccin concreta, para ello indicamos que se
;;; va a permitir redefiniciones en el sistema.

(set-ld-redefinition-action '(:warn . :overwrite) state)

(defun seleccion (lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) nil)
        ((es-literal (car lista)) (seleccion (cdr lista)))
        (t (car lista))))

;;;   Cargamos el archivo de evaluacin de ejemplos:

(value :q)

(compile 'seleccion)

(load "../2-ejemplos/ejemplos.lisp")

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(lp)

(include-book "../2-ejemplos/pelletier")

(value :q)

(evalua-ejemplos 'atp-generico-tableros-1 *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.010 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.010 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.010 |
;;; +---------------------------------------------------------------+--------+

(evalua-ejemplos 'atp-generico-tableros-2 *pelletier*)

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.000 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.010 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.000 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.000 |
;;; +---------------------------------------------------------------+--------+

(evalua-ejemplos 'atp-generico-tableros-3 *pelletier*)

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.000 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.000 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.010 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.000 |
;;; +---------------------------------------------------------------+--------+

;;;   Cargamos el archivo con las frmulas de Urquhart y evaluamos los
;;; ejemplos: 

(lp)

(include-book "../2-ejemplos/urquhart")

(value :q)

(evalua-ejemplos 'atp-generico-tableros-1 *urquhart-v* t)

;;;   Los resultados obtenidos en tiempo con las frmulas de Urquhart son los
;;; siguientes:

;;; +---------+------------+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.010 |
;;; | U_8     |      0.000 |
;;; | U_9     |      0.000 |
;;; | U_10    |      0.000 |
;;; | U_11    |      0.010 |
;;; | U_12    |      0.000 |
;;; | U_13    |      0.010 |
;;; | U_14    |      0.000 |
;;; | U_15    |      0.010 |
;;; | U_16    |      0.010 |
;;; | U_17    |      0.010 |
;;; | U_18    |      0.010 |
;;; | U_19    |      0.010 |
;;; | U_20    |      0.010 |
;;; +---------+------------+

(evalua-ejemplos 'atp-generico-tableros-2 *urquhart-v*)

;;; +---------+------------+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.000 |
;;; | U_8     |      0.000 |
;;; | U_9     |      0.000 |
;;; | U_10    |      0.010 |
;;; | U_11    |      0.000 |
;;; | U_12    |      0.010 |
;;; | U_13    |      0.000 |
;;; | U_14    |      0.010 |
;;; | U_15    |      0.010 |
;;; | U_16    |      0.000 |
;;; | U_17    |      0.010 |
;;; | U_18    |      0.020 |
;;; | U_19    |      0.010 |
;;; | U_20    |      0.010 |
;;; +---------+------------+

(evalua-ejemplos 'atp-generico-tableros-3 *urquhart-v* t)

;;; +---------+------------+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.000 |
;;; | U_8     |      0.000 |
;;; | U_9     |      0.010 |
;;; | U_10    |      0.000 |
;;; | U_11    |      0.000 |
;;; | U_12    |      0.010 |
;;; | U_13    |      0.000 |
;;; | U_14    |      0.010 |
;;; | U_15    |      0.010 |
;;; | U_16    |      0.010 |
;;; | U_17    |      0.010 |
;;; | U_18    |      0.010 |
;;; | U_19    |      0.010 |
;;; | U_20    |      0.010 |
;;; +---------+------------+

(evalua-ejemplos 'atp-generico-tableros-1 *urquhart-f* t)

(evalua-ejemplos 'atp-generico-tableros-2 *urquhart-f* t)

(evalua-ejemplos 'atp-generico-tableros-3 *urquhart-f* t)

|#

;;;============================================================================
