;;;============================================================================
;;;
;;; Ttulo: Un STP exitoso basado en Davis & Putnam
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(in-package "ACL2")

(certify-book "../8-davisputnam/SAT-davisputnam")

|#

(in-package "ACL2")

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

;;;   SAT proposicional basado en el mtodo de Davis y Putnam generado como
;;; una instancia del modelo genrico desarrollado en "SAT-generico".

(include-book "../8-davisputnam/forma-clausal")

(include-book "../7-generico/SAT-generico")

(in-theory (disable forma-clausal))

;(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 1) STP exitoso basado en bifurcacin
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Objetos proposicionales (d&p-objeto)
;;;
;;;----------------------------------------------------------------------------

(defun tiene-literales-complementarios (C)
  (declare (xargs :guard (true-listp C)))
  (cond ((endp C) nil)
        ((member-equal (complementario (car C)) (cdr C)) t)
        (t (tiene-literales-complementarios (cdr C)))))

(defun literal-en-forma-clausal (L Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
	(t (or (member-equal L (car Fc))
	       (literal-en-forma-clausal L (cdr Fc))))))

(defun lista-literales-no-esta-en-forma-clausal (lista-L Fc)
  (declare (xargs :guard (and (true-listp lista-L)
			      (true-list-listp Fc))))
  (cond ((endp lista-L) t)
	(t (and (not (literal-en-forma-clausal (car lista-L) Fc))
		(not (literal-en-forma-clausal 
		      (complementario (car lista-L)) Fc))
		(lista-literales-no-esta-en-forma-clausal (cdr lista-L) Fc)))))

(defun d&p-objeto (S-M)
  (declare (xargs :guard (and (true-listp S-M)
			      (true-list-listp (car S-M)))))
  (and (es-forma-clausal (car S-M))
       (es-clausula (cdr S-M))
       (not (tiene-literales-complementarios (cdr S-M)))
       (lista-literales-no-esta-en-forma-clausal (cdr S-M) (car S-M))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Asignaciones distinguidas (d&p-asignacion-distinguida)
;;;
;;;----------------------------------------------------------------------------

(defun modelo-lista-literales (sigma lista-L)
  (declare (xargs :guard (and (alistp sigma)
			      (true-listp lista-L))))
  (cond ((endp lista-L) t)
	((modelo sigma (car lista-L))
	 (modelo-lista-literales sigma (cdr lista-L)))
	(t nil)))  

(defun d&p-asignacion-distinguida (sigma S-M)
  (declare (xargs :guard (and (alistp sigma)
			      (true-listp S-M)
			      (true-list-listp (car S-M)))))
  (and (modelo-forma-clausal sigma (car S-M))
       (modelo-lista-literales sigma (cdr S-M))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Regla de computacin (d&p-regla-computacion)
;;;
;;;----------------------------------------------------------------------------

(defun contiene-clausula-vacia (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
	((consp (car Fc))
	 (contiene-clausula-vacia (cdr Fc)))
	(t t)))

(encapsulate
 (((seleccion *) => *))
 
 (local
  (defun seleccion (Fc)
    (declare (xargs :guard (true-list-listp Fc)))
    (cond ((endp Fc) nil)
	  ((endp (car Fc))
	   (seleccion (cdr Fc)))
	  (t (car (car Fc))))))

 (defthm seleccion-esta-en-forma-clausal
   (implies (seleccion Fc)
	    (literal-en-forma-clausal (seleccion Fc) Fc)))

 (defthm seleccion-es-literal
   (implies (and (es-forma-clausal Fc)
		 (seleccion Fc))
	    (es-literal (seleccion Fc))))

 (defthm seleccion-existe
   (implies (and (es-forma-clausal Fc)
		 (consp Fc)
		 (consp (car Fc)))
	    (seleccion Fc)))

 )

(encapsulate
 ()

 (defun elimina-literal (L C)
   (declare (xargs :guard (true-listp C)))
   (cond ((endp C) C)
	 ((equal L (car C))
	  (elimina-literal L (cdr C)))
	 (t (cons (car C) (elimina-literal L (cdr C))))))
 
 (local
  (defthm member-equal-elimina-literal
    (iff (member-equal L-1 (elimina-literal L-2 C))
	 (and (member-equal L-1 C)
	      (not (equal L-1 L-2))))))
  
 (local
  (defthm elimina-literal-es-clausula
    (implies (es-clausula C)
	     (es-clausula (elimina-literal L C)))))

 (local
  (defthm elimina-literal-true-listp
    (implies (true-listp C)
	     (true-listp (elimina-literal L C)))))
  
 (defun reduce-forma-clausal-literal (L Fc)
   (declare (xargs :guard (true-list-listp Fc)))
   (cond ((endp Fc) nil)
	 ((member-equal L (car Fc))
	  (reduce-forma-clausal-literal L (cdr Fc)))
	 (t (cons (elimina-literal (complementario L) (car Fc))
		  (reduce-forma-clausal-literal L (cdr Fc))))))
 
 (defun reduce-forma-clausal-literal-c (L Fc)
   (declare (xargs :guard (true-list-listp Fc)))
   (cond ((endp Fc) nil)
	 ((member-equal (complementario L) (car Fc))
	  (reduce-forma-clausal-literal-c L (cdr Fc)))
	 (t (cons (elimina-literal L (car Fc))
		  (reduce-forma-clausal-literal-c L (cdr Fc))))))

;;;   La reduccin por un literal mantiene la propiedad de ser forma clausal:

 (defthm es-forma-clausal-reduce-forma-clausal-literal
   (implies (es-forma-clausal Fc)
	    (es-forma-clausal (reduce-forma-clausal-literal L Fc))))

 (defthm es-forma-clausal-reduce-forma-clausal-literal-C
   (implies (es-forma-clausal Fc)
	    (es-forma-clausal (reduce-forma-clausal-literal-c L Fc))))

;;;   La reduccin por un literal mantiene la propiedad de ser lista propia de
;;; listas propias:

 (local
  (defthm true-list-listp-reduce-forma-clausal-literal
    (implies (true-list-listp Fc)
	     (true-list-listp (reduce-forma-clausal-literal L Fc)))))

 (local
  (defthm true-list-listp-reduce-forma-clausal-literal-c
    (implies (true-list-listp Fc)
	     (true-list-listp (reduce-forma-clausal-literal-c L Fc)))))
  
 (defun d&p-regla-computacion (S-M)
   (declare (xargs :guard (and (true-listp S-M)
			       (true-list-listp (car S-M)))))
   (let ((Fc (car S-M))
	 (lista-L (cdr S-M)))
     (if (contiene-clausula-vacia Fc)
	 nil
	 (let ((L (seleccion Fc)))
	   (if L
	       (list (cons (reduce-forma-clausal-literal L Fc)
			   (cons L lista-L))
		     (cons (reduce-forma-clausal-literal-c L Fc)
			   (cons (complementario L) lista-L)))
	       t)))))

 (defthm complementario-literal-es-literal
   (implies (es-literal L)
	    (es-literal (complementario L))))
  
 (defthm complementario-complementario-literal
   (implies (es-literal L)
	    (equal (complementario (complementario L)) L)))

 (in-theory (disable complementario))

 (defthm lista-literales-no-esta-en-forma-clausal-1
   (implies (and (lista-literales-no-esta-en-forma-clausal lista-L Fc)
		 (literal-en-forma-clausal L Fc))
	    (not (member-equal L lista-L))))

 (in-theory (disable es-literal))
  
 (defthm lista-literales-no-esta-en-forma-clausal-2
   (implies (and (lista-literales-no-esta-en-forma-clausal lista-L Fc)
		 (literal-en-forma-clausal L Fc)
		 (es-literal L))
	    (not (member-equal (complementario L) lista-L))))

 (local
  (defthm literal-en-forma-clausal-reduce
    (implies (not (literal-en-forma-clausal L-1 Fc))
	     (not (literal-en-forma-clausal 
		   L-1 (reduce-forma-clausal-literal L-2 Fc))))))

 (local
  (defthm literal-en-forma-clausal-reduce-c
    (implies (not (literal-en-forma-clausal L-1 Fc))
	     (not (literal-en-forma-clausal 
		   L-1 (reduce-forma-clausal-literal-c L-2 Fc))))))

 (defthm lista-literales-no-esta-en-forma-clausal-reduce
   (implies (and (lista-literales-no-esta-en-forma-clausal lista-L Fc)
		 (literal-en-forma-clausal L Fc))
	    (lista-literales-no-esta-en-forma-clausal
	     (cons L lista-L)
	     (reduce-forma-clausal-literal L Fc))))

 (local
  (defthm complementario-literal-reducido-no-esta-en-forma-clausal
    (not (literal-en-forma-clausal (complementario L)
				   (reduce-forma-clausal-literal-c L Fc)))))

 (local
  (defthm literal-reducido-no-esta-en-forma-clausal
    (not (literal-en-forma-clausal L (reduce-forma-clausal-literal-c L Fc)))))

 (defthm lista-literales-no-esta-en-forma-clausal-reduce-C
   (implies (and (lista-literales-no-esta-en-forma-clausal lista-C Fc)
		 (literal-en-forma-clausal L Fc)
		 (es-literal L))
	    (lista-literales-no-esta-en-forma-clausal
	     (cons (complementario L) lista-C)
	     (reduce-forma-clausal-literal-c L Fc))))

;  (local
;   (in-theory (disable lista-literales-no-esta-en-forma-clausal)))
  
 (defthm d&p-objeto-d&p-regla-computacion
   (implies (and (d&p-objeto S-M-1)
		 (member-equal S-M-2 (d&p-regla-computacion S-M-1)))
	    (d&p-objeto S-M-2)))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Funcin de representacin (d&p-representacion)
;;;
;;;----------------------------------------------------------------------------

(defun d&p-representacion (F)
  (list (forma-clausal F)))

(defthm d&p-representacion-es-d&p-objeto
  (implies (es-proposicional F)
	   (d&p-objeto (d&p-representacion F))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Funcin de medida (d&p-medida)
;;;
;;;----------------------------------------------------------------------------

;;;   Necesitamos demostrar que la reduccin por un literal genera objetos ms
;;; pequeos que el original, para poder definir una funcin recursiva basada
;;; en dicha reduccin:

(defun medida-forma-clausal (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) 0)
	(t (+ (len (car Fc)) 
	      (medida-forma-clausal (cdr Fc))))))

(defun d&p-medida (S-M)
  (declare (xargs :guard (and (true-listp S-M)
			      (true-list-listp (car S-M)))))
  (medida-forma-clausal (car S-M)))

(defthm e0-ordinalp-d&p-media
  (e0-ordinalp (d&p-medida secuente)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.6) Funcin modelo (d&p-modelo)
;;;
;;;----------------------------------------------------------------------------

(defun modelo-literales (lista-L)
  (declare (xargs :guard (true-listp lista-L)))
  (cond ((endp lista-L) nil)
	(t (asume-literal (car lista-L)
			  (modelo-literales (cdr lista-L))))))

(defun d&p-modelo (S-M)
  (declare (xargs :guard (and (true-listp S-M)
			      (true-list-listp (car S-M)))))  
  (modelo-literales (cdr S-M)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.7) Propiedad P1
;;;
;;;----------------------------------------------------------------------------

(defthm reduce-forma-clausal-literal-disminuye-medida
  (implies (literal-en-forma-clausal L Fc)
	   (< (medida-forma-clausal (reduce-forma-clausal-literal L Fc))
	      (medida-forma-clausal Fc))))

(defthm reduce-forma-clausal-literal-c-disminuye-medida
  (implies (literal-en-forma-clausal L Fc)
	   (< (medida-forma-clausal (reduce-forma-clausal-literal-c L Fc))
	      (medida-forma-clausal Fc))))

(defthm P1-d&p
  (implies (member-equal S-M-2 (d&p-regla-computacion S-M-1))
	   (e0-ord-< (d&p-medida S-M-2)
		     (d&p-medida S-M-1))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.8) Propiedad P2
;;;
;;;----------------------------------------------------------------------------

(defthm P2-d&p
  (implies (es-proposicional F)
	   (iff (d&p-asignacion-distinguida sigma (d&p-representacion F))
		(modelo sigma F))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.9) Propiedad P3
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 ()

 (defthm contiene-clausula-vacia-es-inconsistente
   (implies (contiene-clausula-vacia Fc)
	    (not (modelo-forma-clausal sigma Fc))))

 (in-theory (disable modelo-forma-clausal-forma-clausal
		     valor-forma-clausal-forma-clausal))

 (defun true-car-list-listp (lista)
   (declare (xargs :verify-guards t))
   (cond ((atom lista) (eq lista nil))
	 (t (and (true-listp (car lista))
		 (true-list-listp (caar lista))
		 (true-car-list-listp (cdr lista))))))

 (defun d&p-asignacion-distinguida-lista (sigma d&p-obj-lst)
   (declare (xargs :guard (and (alistp sigma)
			       (true-car-list-listp d&p-obj-lst))))
   (cond ((endp d&p-obj-lst) nil)
	 (t (or (d&p-asignacion-distinguida sigma (car d&p-obj-lst))
		(d&p-asignacion-distinguida-lista sigma (cdr d&p-obj-lst))))))

 (local
  (defthm modelo-clausula-elimina-literal-falso
    (implies (and (es-literal L)
		  (not (modelo sigma L)))
	     (iff (modelo-clausula sigma (elimina-literal L C)) 
		  (modelo-clausula sigma C)))))

 (local
  (defthm modelo-clausula-elimina-complementario-literal-verdadero
    (implies (and (es-literal L)
		  (modelo sigma L))
	     (iff (modelo-clausula 
		   sigma (elimina-literal (complementario L) C))
		  (modelo-clausula sigma C)))
    :hints (("Goal" :in-theory (enable modelo)
	     :use (:instance modelo-clausula-elimina-literal-falso
			     (L (complementario L)))))))

 (defthm reduce-forma-clausal-literal-correcto
   (implies (and (es-literal L)
		 (modelo sigma L))
	    (iff (modelo-forma-clausal 
		  sigma (reduce-forma-clausal-literal L Fc))
		 (modelo-forma-clausal sigma Fc)))
   :hints (("Goal" :in-theory (disable complementario valor))))

 (defthm reduce-forma-clausal-literal-c-correcto
   (implies (and (es-literal L)
		 (modelo sigma (complementario L)))
	    (iff (modelo-forma-clausal 
		  sigma (reduce-forma-clausal-literal-c L Fc))
		 (modelo-forma-clausal sigma Fc)))
   :hints (("Goal" :in-theory (disable complementario valor)
	    :induct (reduce-forma-clausal-literal-c L Fc))))

 (defthm reduce-forma-clausal-literal-c-correcto-aux
   (implies (and (modelo-forma-clausal 
		  sigma (reduce-forma-clausal-literal L Fc))
		 (modelo-forma-clausal 
		  sigma (reduce-forma-clausal-literal-c L Fc))
		 (es-literal L))
	    (modelo-forma-clausal sigma Fc))
   :hints (("Goal" 
	    :in-theory (disable modelo-forma-clausal)
	    :cases ((modelo sigma L)))))

 (defthm reduce-forma-clausal-literal-verdadero
   (implies (and (not (modelo-forma-clausal 
		       sigma (reduce-forma-clausal-literal L Fc)))
		 (modelo-forma-clausal sigma Fc)
		 (es-literal L))
	    (modelo-forma-clausal 
	     sigma (reduce-forma-clausal-literal-c L Fc)))
   :hints (("Goal" 
	    :in-theory (disable complementario es-literal 
				modelo-forma-clausal)
	    :cases ((modelo sigma L)))))

 (defthm P3-d&p
   (implies (and (d&p-objeto S-M)
		 (not (equal (d&p-regla-computacion S-M) t)))
	    (iff (d&p-asignacion-distinguida-lista 
		  sigma (d&p-regla-computacion S-M))
		 (d&p-asignacion-distinguida sigma S-M)))
   :hints (("Goal" 
	    :in-theory (disable complementario es-literal
				modelo-forma-clausal))))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.10) Propiedad P4
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 ()

 (local
  (defthm modelo-literales-funciona-literal
    (implies (and (es-literal L)
		  (es-clausula lista-L)
		  (not (tiene-literales-complementarios lista-L))
		  (member-equal L lista-L))
	     (modelo (modelo-literales lista-L) L))
    :hints (("Goal"
	     :in-theory (disable literal-mantiene-valor))
	    ("Subgoal *1/6''"
	     :use (:instance literal-mantiene-valor 
			     (L-2 L)
			     (L-1 (car lista-L))
			     (sigma (modelo-literales (cdr lista-L)))))
	    )))

 (defthm forma-clausal-vacia-es-valida
   (implies (and (es-forma-clausal Fc)
		 (not (seleccion Fc))
		 (not (contiene-clausula-vacia Fc)))
	    (modelo-forma-clausal sigma Fc)))

 (local
  (defthm member-clausula-literal
    (implies (and (es-clausula C)
		  (member-equal L C))
	     (es-literal L))))

 (local
  (include-book "../6-extensiones/listas"))

 (local
  (in-theory (disable modelo)))

 (defthm modelo-literales-funciona
   (implies (and (es-clausula C)
		 (not (tiene-literales-complementarios C)))
	    (modelo-lista-literales (modelo-literales C) C))
   :hints ((extension-conjuntiva-a-listas
	    C
	    (lambda (x)
	      (modelo (modelo-literales C) x))
	    (lambda (lst)
	      (modelo-lista-literales (modelo-literales C) lst)))))

 (defthm P4-d&p
   (implies (and (d&p-objeto S-M)
		 (equal (d&p-regla-computacion S-M) t))
	    (d&p-asignacion-distinguida (d&p-modelo S-M) S-M)))

 )
 
;;;----------------------------------------------------------------------------
;;;
;;; 1.11) Instancia
;;;
;;;----------------------------------------------------------------------------

;;;   A continuacin instanciamos la teora (definiciones y teoremas) del
;;; demostrador sat genrico con la siguiente correspondencia entre las
;;; funciones:

;;;   gen-objeto                       -- d&p-objeto
;;;   gen-representacion               -- d&p-representacion
;;;   gen-asignacion-distinguida       -- d&p-asignacion-distinguida
;;;   gen-asignacion-distinguida-lista -- d&p-asignacion-distinguida-lista
;;;   gen-regla-computacion            -- d&p-regla-computacion
;;;   gen-seleccion                    -- car
;;;   gen-medida                       -- d&p-medida
;;;   gen-modelo                       -- d&p-modelo

;;;   Falta por demostrar la propiedad asociada a la funcin obj-seleccion,
;;; que aplicada a un lista de ramas no vaca devuelva un elemento de dicha
;;; lista. No es necesario establecerla explcitamente dado la simplicidad de
;;; la funcin que usamos en su lugar.

(encapsulate
 ()

 (local
  (in-theory (disable d&p-objeto
		      d&p-representacion
		      d&p-asignacion-distinguida
		      d&p-regla-computacion
		      d&p-medida
		      (:type-prescription d&p-medida)
		      d&p-modelo)))

 (definstance-*sat-generico*
   ((gen-objeto                       d&p-objeto)
    (gen-representacion               d&p-representacion)
    (gen-asignacion-distinguida       d&p-asignacion-distinguida)
    (gen-asignacion-distinguida-lista d&p-asignacion-distinguida-lista)
    (gen-regla-computacion            d&p-regla-computacion)
    (gen-seleccion                    car)
    (gen-medida                       d&p-medida)
    (gen-modelo                       d&p-modelo))
   "-d&p-1")

 )

;;;============================================================================
;;;
;;; 2) STP exitoso basado en bifurcacion y clusulas unitarias
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Regla de computacin (d&p-regla-computacion-2)
;;;
;;;----------------------------------------------------------------------------

;;;   Caracterizacin de las clsulas unitarias:

(defun clausula-unitaria (C)
  (declare (xargs :guard (true-listp C)))
  (and (consp C)
       (not (consp (elimina-literal (car C) C)))))

;;;   Bsqueda de una clusula unitaria:

(defun busca-clausula-unitaria (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
	((clausula-unitaria (car Fc))
	 (car Fc))
	(t (busca-clausula-unitaria (cdr Fc)))))

(defthm literal-de-clausula-unitaria-esta-en-forma-clausal
  (implies (busca-clausula-unitaria Fc)
	   (literal-en-forma-clausal (car (busca-clausula-unitaria Fc)) Fc)))

(defthm literal-de-clausula-unitaria-es-literal
  (implies (and (busca-clausula-unitaria Fc)
		(es-forma-clausal Fc))
	   (es-literal (car (busca-clausula-unitaria Fc)))))

;;;   Nueva regla de computacin

(defun d&p-regla-computacion-2 (S-M)
  (declare (xargs :guard (and (true-listp S-M)
			      (true-list-listp (car S-M)))))
  (let ((Fc (car S-M))
	(lista-L (cdr S-M)))
    (if (contiene-clausula-vacia Fc)
	nil
	(let ((C (busca-clausula-unitaria Fc)))
	  (if C
	      (list (cons (reduce-forma-clausal-literal (car C) Fc)
			  (cons (car C) lista-L)))
	      (let ((L (seleccion Fc)))
		(if L
		    (list (cons (reduce-forma-clausal-literal L Fc)
				(cons L lista-L))
			  (cons (reduce-forma-clausal-literal-c L Fc)
				(cons (complementario L) lista-L)))
		    t)))))))

(defthm d&p-objeto-d&p-regla-computacion-2
  (implies (and (d&p-objeto S-M-1)
		(member-equal S-M-2 (d&p-regla-computacion-2 S-M-1)))
	   (d&p-objeto S-M-2)))

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) Propiedad P1
;;;
;;;----------------------------------------------------------------------------

(defthm P1-d&p-2
  (implies (member-equal S-M-2 (d&p-regla-computacion-2 S-M-1))
	   (e0-ord-< (d&p-medida S-M-2)
		     (d&p-medida S-M-1))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.3) Propiedad P3
;;;
;;;----------------------------------------------------------------------------

(defthm modelo-forma-clausal-modelo-clausula-unitaria
  (implies (and (es-forma-clausal Fc)
		(busca-clausula-unitaria Fc)
		(modelo-forma-clausal sigma Fc))
	   (modelo sigma (car (busca-clausula-unitaria Fc)))))

(defthm P3-d&p-2
  (implies (and (d&p-objeto S-M)
		(not (equal (d&p-regla-computacion-2 S-M) t)))
	   (iff (d&p-asignacion-distinguida-lista 
		 sigma (d&p-regla-computacion-2 S-M))
		(d&p-asignacion-distinguida sigma S-M)))
  :hints (("Goal" 
	   :in-theory (disable complementario es-literal
			       modelo-forma-clausal))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.4) Propiedad P4
;;;
;;;----------------------------------------------------------------------------

(defthm P4-d&p-2
  (implies (and (d&p-objeto S-M)
		(equal (d&p-regla-computacion-2 S-M) t))
	   (d&p-asignacion-distinguida (d&p-modelo S-M) S-M)))

;;;----------------------------------------------------------------------------
;;;
;;; 2.5) Instancia
;;;
;;;----------------------------------------------------------------------------

;;;   A continuacin instanciamos la teora (definiciones y teoremas) del
;;; demostrador sat genrico con la siguiente correspondencia entre las
;;; funciones:

;;;   gen-objeto                       -- d&p-objeto
;;;   gen-representacion               -- d&p-representacion
;;;   gen-asignacion-distinguida       -- d&p-asignacion-distinguida
;;;   gen-asignacion-distinguida-lista -- d&p-asignacion-distinguida-lista
;;;   gen-regla-computacion            -- d&p-regla-computacion-2
;;;   gen-seleccion                    -- car
;;;   gen-medida                       -- d&p-medida
;;;   gen-modelo                       -- d&p-modelo

;;;   Falta por demostrar la propiedad asociada a la funcin obj-seleccion,
;;; que aplicada a un lista de ramas no vaca devuelva un elemento de dicha
;;; lista. No es necesario establecerla explcitamente dado la simplicidad de
;;; la funcin que usamos en su lugar.

(encapsulate
 ()

 (local
  (in-theory (disable d&p-objeto
		      d&p-representacion
		      d&p-asignacion-distinguida
		      d&p-regla-computacion-2
		      d&p-medida
		      (:type-prescription d&p-medida)
		      d&p-modelo)))

 (definstance-*sat-generico*
   ((gen-objeto                       d&p-objeto)
    (gen-representacion               d&p-representacion)
    (gen-asignacion-distinguida       d&p-asignacion-distinguida)
    (gen-asignacion-distinguida-lista d&p-asignacion-distinguida-lista)
    (gen-regla-computacion            d&p-regla-computacion-2)
    (gen-seleccion                    car)
    (gen-medida                       d&p-medida)
    (gen-modelo                       d&p-modelo))
   "-d&p-2")

 )

;;;============================================================================
;;;
;;; 3) Ejemplos
;;;
;;;============================================================================

#|

(include-book "SAT-davisputnam")

;;;   Un ejemplo de utilizacin de este ATP es el siguiente:

;;;   Necesitamos una funcin de seleccin concreta, para ello indicamos que se
;;; va a permitir redefiniciones en el sistema.

(set-ld-redefinition-action '(:warn . :overwrite) state)

(defun seleccion (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  
  (cond ((endp Fc) nil)
	((endp (car Fc))
	 (seleccion (cdr Fc)))
	(t (car (car Fc)))))

;;;   Cargamos el archivo de evaluacin de ejemplos:

(value :q)

(compile 'seleccion)

(load "../0-ejemplos/ejemplos.lisp")

(load "../0-ejemplos/reinas.lisp")

(evalua-ejemplos 'sat-generico-d&p-1 *n-reinas*)

;;; +==========+============+
;;; | Frmula  | Tiempo     |
;;; +----------+------------+
;;; | Reinas-2 |      0.000 |
;;; | Reinas-3 |      0.010 |
;;; | Reinas-4 |      0.040 |
;;; | Reinas-5 |      0.140 |
;;; | Reinas-6 |      0.250 |
;;; | Reinas-7 |      0.570 |
;;; | Reinas-8 |      1.600 |
;;; +==========+============+

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(lp)

(include-book "../0-ejemplos/pelletier")

(value :q)

(evalua-ejemplos 'dat-generico-d&p-1 *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.010 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.010 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.010 |
;;; +---------------------------------------------------------------+--------+

;;;   Cargamos el archivo con las frmulas de Urquhart y evaluamos los
;;; ejemplos: 

(lp)

(include-book "../0-ejemplos/urquhart")

(value :q)

(evalua-ejemplos 'sat-generico-d&p-1 *urquhart-v*)

;;;   Los resultados obtenidos en tiempo con las frmulas de Urquhart son los
;;; siguientes:

;;; +---------+------------+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.000 |
;;; | U_8     |      0.000 |
;;; | U_9     |      0.000 |
;;; | U_10    |      0.000 |
;;; | U_11    |      0.000 |
;;; | U_12    |      0.000 |
;;; | U_13    |      0.000 |
;;; | U_14    |      0.000 |
;;; | U_15    |      0.010 |
;;; | U_16    |      0.000 |
;;; | U_17    |      0.000 |
;;; | U_18    |      0.000 |
;;; | U_19    |      0.000 |
;;; | U_20    |      0.010 |
;;; +---------+------------+

(evalua-ejemplos 'sat-generico-d&p-2 *urquhart-f* t)

;;; +---------+-----+------------+
;;; | Frmula | Sol | Tiempo     |
;;; +---------+-----+------------+
;;; |  U_1   | NIL |      0.000 |
;;; |  U_2   | NIL |      0.000 |
;;; |  U_3   | NIL |      0.000 |
;;; |  U_4   | NIL |      0.000 |
;;; |  U_5   | NIL |      0.010 |
;;; |  U_6   | NIL |      0.000 |
;;; |  U_7   | NIL |      0.020 |
;;; |  U_8   | NIL |      0.030 |
;;; |  U_9   | NIL |      0.100 |
;;; |  U_10  | NIL |      0.220 |
;;; |  U_11  | NIL |      0.510 |
;;; |  U_12  | NIL |      1.300 |
;;; |  U_13  | NIL |      2.950 |
;;; |  U_14  | NIL |      6.830 |
;;; |  U_15  | NIL |     15.220 |
;;; |  U_16  | NIL |     34.210 |
;;; |  U_17  | NIL |     76.610 |
;;; |  U_18  | NIL |    170.380 |
;;; |  U_19  | NIL |    376.120 |
;;; |  U_20  | NIL |    828.710 |
;;; +---------+-----+------------+

(set-ld-redefinition-action '(:warn . :overwrite) state)

|#

;;;============================================================================