;;;============================================================================
;;;
;;; Ttulo: Clusulas y formas clausales (I) (Kleene)
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../8-davisputnam/clausulas-K")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin.
;;;
;;;============================================================================

(include-book "../3-logica/semantica-K")

(in-theory (disable definiciones-conectivas))

;;;============================================================================
;;;
;;; 1) Sintaxis y semntica
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Literales
;;;
;;;----------------------------------------------------------------------------

(defun es-literal-positivo (F)
  (es-atomica F))

(defun es-literal-negativo (F)
  (and (es-negacion F)
       (es-atomica (arg1 F))))

(defun es-literal (F)
  (or (es-literal-positivo F)
      (es-literal-negativo F)))

(defthm es-literal-es-proposicional
  (implies (es-literal F)
	   (es-proposicional F)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) clusulas
;;;
;;;----------------------------------------------------------------------------

(defun es-clausula (C)
  (declare (xargs :guard (true-listp C)))
  (or (endp C)
      (and (es-literal (car C))
	   (es-clausula (cdr C)))))

(defthm member-clausula-es-literal
  (implies (and (member-equal L C) 
		(es-clausula C))
	   (es-literal L)))

(defthm es-clausula-append
  (iff (es-clausula (append C-1 C-2))
       (and (es-clausula C-1)
	    (es-clausula C-2))))

(defun valor-clausula-K (C sigma)
  (declare (xargs :guard (and (true-listp C)
			      (alistp sigma))))
  (cond ((endp C) *F*)
	(t (funcion-de-verdad-de-disyuncion-K
	    (valor-K (car C) sigma)
	    (valor-clausula-K (cdr C) sigma)))))

(defthm valor-clausula-K-es-0-o-1
  (implies (valor-clausula-K C sigma)
	   (or (equal (valor-clausula-K C sigma) *F*)
	       (equal (valor-clausula-K C sigma) *V*)))
  :rule-classes ((:forward-chaining 
		  :trigger-terms ((valor-clausula-K C sigma)))))

(defthm valor-clausula-K-cons
  (equal (valor-clausula-K (cons F C) sigma)
	 (funcion-de-verdad-de-disyuncion-K
	  (valor-K F sigma)
	  (valor-clausula-K C sigma))))

(defthm valor-clausula-K-append
  (equal (valor-clausula-K (append C-1 C-2) sigma)
	 (funcion-de-verdad-de-disyuncion-K
	  (valor-clausula-K C-1 sigma)
	  (valor-clausula-K C-2 sigma))))

(defthm valor-clausula-K-miembro-valor-V
  (implies (and (member-equal L C)
		(equal (valor-K L sigma) *V*))
	   (equal (valor-clausula-K C sigma) *V*)))

(defun modelo-clausula-K (sigma C)
  (declare (xargs :guard (and (true-listp C) 
			      (alistp sigma))))
  (if (consp C)	     
      (or (modelo-K sigma (car c))
	  (modelo-clausula-K sigma (cdr C)))
      nil))

(defthm modelo-clausula-K-cons
  (iff (modelo-clausula-K sigma (cons L C))
       (or (modelo-K sigma L)
	   (modelo-clausula-K sigma C))))

(defthm modelo-clausula-K-append
  (iff (modelo-clausula-K sigma (append C-1 C-2))
       (or (modelo-clausula-K sigma C-1)
	   (modelo-clausula-K sigma C-2))))

(defthm modelo-clausula-K-member
  (implies (and (member-equal L C)
		(modelo-K sigma L))
	   (modelo-clausula-K sigma C)))

(defthm modelo-clausula-K-valor-clausula-K
  (iff (modelo-clausula-K sigma C)
       (equal (valor-clausula-K C sigma) *V*)))

(in-theory (disable modelo-clausula-K-valor-clausula-K))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Extensin de una asignacin a un literal
;;;
;;;----------------------------------------------------------------------------

(defthm complementario-complementario
  (implies (es-literal F)
           (equal (complementario (complementario F)) F))
  :hints (("Goal" :in-theory (enable es-literal))))

(in-theory (disable complementario))

(defun asume-literal (L sigma)
  (declare (xargs :guard (alistp sigma)))
  (cond ((es-simbolo-proposicional L)
         (asume-valor L *V* sigma))
        (t (asume-valor (arg1 L) *F* sigma))))

(defthm asume-literal-es-no-nulo
  (consp (asume-literal L sigma)))

(defthm asume-literal-es-modelo
  (implies (es-literal L)
           (modelo-K (asume-literal L sigma) L))
  :hints (("Goal" :in-theory (enable es-literal modelo-K))))

(defthm asume-literal-alistp
  (implies (alistp sigma)
           (alistp (asume-literal L sigma))))

(defthm negacion-no-es-simbolo-proposicional
  (not (es-simbolo-proposicional (negacion F)))
  :hints (("Goal" :in-theory (enable es-simbolo-proposicional negacion))))

(defthm asume-literal-no-es-modelo-K-c-1
  (implies (es-literal L)
           (not (modelo-K (asume-literal L sigma) (complementario L))))
  :hints (("Goal" :in-theory (enable es-literal complementario modelo-K))))

(defthm asume-literal-no-es-modelo-K-c-2
  (implies (es-literal L)
           (not (modelo-K (asume-literal (complementario L) sigma) L)))
  :hints (("Goal" :in-theory (enable es-literal complementario modelo-K))))

(in-theory (disable asume-literal))

(defthm complementario-literal-es-literal
  (implies (es-literal L)
           (es-literal (complementario L)))
  :hints (("Goal" :in-theory (enable es-literal complementario))))

(defthm literales-iguales-1
  (implies (and (es-literal L-1)
                (es-literal L-2))
           (iff (equal (complementario L-1) (complementario L-2))
                (equal L-1 L-2)))
  :hints (("Goal" :in-theory (disable complementario es-literal))
          ("Subgoal 1"
           :use ((:instance complementario-complementario
                            (F L-1))
                 (:instance complementario-complementario
                            (F L-2))))))

(defthm literales-iguales-2
  (implies (and (es-literal L-1)
                (es-literal L-2))
           (iff (equal L-1 (complementario L-2))
                (equal (complementario L-1) L-2))))

(defthm literal-mantiene-valor
  (implies (and (es-literal L-1)
                (es-literal L-2)
                (not (equal L-1 L-2))
                (not (equal (complementario L-1) L-2)))
           (iff (modelo-K (asume-literal L-1 sigma) L-2)
                (modelo-K sigma L-2)))
  :hints (("Goal" 
           :in-theory (enable asume-literal complementario))))

(defthm modelo-clausula-K-asume-literal-member
  (implies (and (member-equal L C)
                (es-literal L))
           (modelo-clausula-K (asume-literal L sigma) C)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Caracterizacin de clusulas vlidas
;;;
;;;----------------------------------------------------------------------------

(in-theory (disable complementario modelo-K es-literal))

(defun tiene-literales-complementarios (C)
  (declare (xargs :guard (true-listp C)))
  (cond ((endp C) nil)
        ((member-equal (complementario (car C)) (cdr C)) t)
        (t (tiene-literales-complementarios (cdr C)))))

(defthm tiene-literales-complementarios-es-clausula-valida-aux
  (implies (and (tiene-literales-complementarios C)
		(valor-clausula-K C sigma)
		(es-clausula C))
	   (equal (valor-clausula-K C sigma) *V*))
  :hints (("Goal" :induct (tiene-literales-complementarios C))))

(defthm tiene-literales-complementarios-es-clausula-valida
  (implies (and (tiene-literales-complementarios C)
		(valor-clausula-K C sigma)
		(es-clausula C))
	   (modelo-clausula-K sigma C))
  :hints (("Goal" :in-theory (enable modelo-clausula-K-valor-clausula-K))))

(defun contramodelo-clausula (C)
  (declare (xargs :guard (true-listp C)))
  (cond ((endp C) nil)
        (t (asume-literal (complementario (car C))
                          (contramodelo-clausula (cdr C))))))

(defthm no-tiene-literales-complementarios-correcto
  (implies (and (es-clausula C)
		(not (tiene-literales-complementarios C))
		(member-equal L C))
	   (not (member-equal (complementario L) C))))

(defthm contramodelo-clausula-cons
  (implies (and (es-clausula C)
		(es-literal L)
		(not (member-equal (complementario L) C)))
	   (not (modelo-K (contramodelo-clausula (cons L C)) L))))

(defthm contramodelo-clausula-complementario
  (implies (and (es-clausula C)
		(member-equal L C)
		(not (member-equal (complementario L) C)))
	   (not (modelo-K (contramodelo-clausula C) L)))
  :hints (("Goal" :induct (member-equal L C))))

(defthm contramodelo-clausula-funciona
  (implies (and (es-clausula C)
		(not (tiene-literales-complementarios C))
		(member-equal L C))
	   (not (modelo-K (contramodelo-clausula C) L))))

(include-book "../6-extensiones/listas")

(defthm clausula-valida-tiene-literales-complementarios
  (implies (and (not (tiene-literales-complementarios C))
                (es-clausula C))
           (not (modelo-clausula-K (contramodelo-clausula C) C)))
  :hints ((extension-conjuntiva-a-listas
	   C
	   (lambda (elt)
	     (not (modelo-K (contramodelo-clausula C) elt)))
	   (lambda (lst)
	     (not (modelo-clausula-K (contramodelo-clausula C) lst))))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Formas clausales
;;;
;;;----------------------------------------------------------------------------

(defun es-forma-clausal (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (or (endp Fc)
      (and (es-clausula (car Fc))
	   (es-forma-clausal (cdr Fc)))))

(defthm es-forma-clausal-append
  (iff (es-forma-clausal (append Fc-1 Fc-2))
       (and (es-forma-clausal Fc-1)
	    (es-forma-clausal Fc-2))))

(defthm member-es-forma-clausal-es-clausula
  (implies (and (es-forma-clausal Fc)
		(member-equal C Fc))
	   (es-clausula C)))

(defun valor-forma-clausal-K (Fc sigma)
  (declare (xargs :guard (and (true-list-listp Fc)
			      (alistp sigma))))
  (cond ((endp Fc) *V*)
	(t (funcion-de-verdad-de-conjuncion-K
	    (valor-clausula-K (car Fc) sigma)
	    (valor-forma-clausal-K (cdr Fc) sigma)))))

(defthm valor-forma-clausal-K-es-0-o-1
  (implies (valor-forma-clausal-K Fc sigma)
	   (or (equal (valor-forma-clausal-K Fc sigma) *F*)
	       (equal (valor-forma-clausal-K Fc sigma) *V*)))
  :rule-classes ((:forward-chaining 
		  :trigger-terms ((valor-forma-clausal-K Fc sigma)))))

(defthm valor-forma-clausal-K-cons
  (equal (valor-forma-clausal-K (cons C Fc) sigma)
	 (funcion-de-verdad-de-conjuncion-K
	  (valor-clausula-K C sigma)
	  (valor-forma-clausal-K Fc sigma))))

(defthm valor-forma-clausal-K-append
  (equal (valor-forma-clausal-K (append Fc-1 Fc-2) sigma)
	 (funcion-de-verdad-de-conjuncion-K
	  (valor-forma-clausal-K Fc-1 sigma)
	  (valor-forma-clausal-K Fc-2 sigma))))

(defun modelo-forma-clausal-K (sigma Fc)
  (declare (xargs :guard (and (true-list-listp Fc) 
			      (alistp sigma))))
  (if (consp Fc)	     
      (and (modelo-clausula-K sigma (car Fc))
	   (modelo-forma-clausal-K sigma (cdr Fc)))
      t))

(defthm modelo-forma-clausal-K-cons
  (iff (modelo-forma-clausal-K sigma (cons C Fc))
       (and (modelo-clausula-K sigma C)
	    (modelo-forma-clausal-K sigma Fc))))

(defthm modelo-forma-clausal-K-append
  (iff (modelo-forma-clausal-K sigma (append Fc-1 Fc-2))
       (and (modelo-forma-clausal-K sigma Fc-1)
	    (modelo-forma-clausal-K sigma Fc-2))))

(defthm no-modelo-forma-clausal-K-clausula-falsa
  (implies (and (not (modelo-clausula-K sigma C))
		(member-equal C Fc)
		(es-forma-clausal Fc))
	   (not (modelo-forma-clausal-K sigma Fc))))

(defthm modelo-forma-clausal-K-valor-forma-clausal-K
  (iff (modelo-forma-clausal-K sigma Fc)
       (equal (valor-forma-clausal-K Fc sigma) *V*))
  :hints (("Goal" :in-theory (enable modelo-clausula-K-valor-clausula-K))))

(in-theory (disable modelo-forma-clausal-K-valor-forma-clausal-K))

;;;----------------------------------------------------------------------------
;;;
;;; 1.6) Forma clausal con clusula vaca es inconsistente
;;;
;;;----------------------------------------------------------------------------

(defun contiene-clausula-vacia (Fc)
  (declare (xargs :guard (true-list-listp Fc)))

  (cond ((endp Fc) nil)
        ((consp (car Fc))
         (contiene-clausula-vacia (cdr Fc)))
        (t t)))

(defthm contiene-clausula-vacia-es-inconsistente
  (implies (contiene-clausula-vacia Fc)
           (not (modelo-forma-clausal-K sigma Fc))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.7) Conjunto de smbolos de un literal, una clusula y una forma clausal
;;;
;;;----------------------------------------------------------------------------

(defun variable-literal (L)
  (if (es-literal-negativo L)
      (arg1 L)
      L))

(defun variables-clausula (C)
  (declare (xargs :guard (true-listp C)))
  (if (endp C) 
      nil
      (union-equal (list (variable-literal (car C)))
		   (variables-clausula (cdr C)))))

(defun variables-forma-clausal (F)
  (declare (xargs :guard (true-list-listp F)))
  (if (endp F) 
      nil
      (union-equal (variables-clausula (car F))
		   (variables-forma-clausal (cdr F)))))

;;;============================================================================
