;;;============================================================================
;;;
;;; Ttulo: Procedimiento de Davis & Putnam (Kleene)
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../8-davisputnam/davis-putnam-K")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../8-davisputnam/clausulas-K")

(in-theory (disable definiciones-conectivas
                    otra-definicion-de-es-proposicional
                    es-literal))

;;;============================================================================
;;;
;;; 1) El procedimiento de Davis & Putnam
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Eliminacin de clusulas con literales complementarios
;;;
;;;----------------------------------------------------------------------------

(in-theory (disable complementario))

(defun tiene-clausulas-con-complementarios (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
        ((tiene-literales-complementarios (car Fc)) t)
        (t (tiene-clausulas-con-complementarios (cdr Fc)))))

(defun elimina-clausulas-con-complementarios (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
        ((tiene-literales-complementarios (car Fc))
         (elimina-clausulas-con-complementarios (cdr Fc)))
        (t (cons (car Fc)
                 (elimina-clausulas-con-complementarios (cdr Fc))))))

(defthm elimina-clausulas-con-complementarios-funciona
  (not (tiene-clausulas-con-complementarios
        (elimina-clausulas-con-complementarios Fc))))

;;;----------------------------------------------------------------------------

(encapsulate
 ()

 (local
  (defthm elimina-clausulas-con-complementarios-preserva-valor
    (implies (and (valor-forma-clausal-K Fc sigma)
		  (es-forma-clausal Fc))
	     (equal (valor-forma-clausal-K 
		     (elimina-clausulas-con-complementarios Fc) sigma)
		    (valor-forma-clausal-K Fc sigma)))))
  
 (defthm elimina-clausulas-con-complementarios-preserva-modelos
   (implies (and (valor-forma-clausal-K Fc sigma)
		 (es-forma-clausal Fc))
	    (iff (modelo-forma-clausal-K 
		  sigma (elimina-clausulas-con-complementarios Fc))
		 (modelo-forma-clausal-K sigma Fc)))
   :hints (("Goal" 
	    :in-theory (enable modelo-forma-clausal-K-valor-forma-clausal-K))))
 
 )

;;;----------------------------------------------------------------------------

(defthm elimina-clausulas-con-complementarios-nil-forma-clausal-valida
  (implies (and (es-forma-clausal Fc)
		(equal (elimina-clausulas-con-complementarios Fc) nil)
		(valor-forma-clausal-K Fc sigma))
	   (modelo-forma-clausal-K sigma Fc))
  :hints (("Goal" :induct (elimina-clausulas-con-complementarios Fc))
	  ("Subgoal *1/2" :cases ((valor-clausula-K (car Fc) sigma))
	   :in-theory (enable modelo-forma-clausal-K-valor-forma-clausal-K))))

(defthm forma-clausal-valida-elimina-clausulas-con-complementarios-nil
  (implies (and (es-forma-clausal Fc)
		(not (equal (elimina-clausulas-con-complementarios Fc) nil)))
	   (not (modelo-forma-clausal-K
		 (contramodelo-clausula
		  (car (elimina-clausulas-con-complementarios Fc)))
		 Fc))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Reduccin de una forma clausal por un literal
;;;
;;;----------------------------------------------------------------------------

(defun elimina-literal (L C)
  (declare (xargs :guard (true-listp C)))
  (cond ((endp C) C)
        ((equal L (car C))
         (elimina-literal L (cdr C)))
        (t (cons (car C) (elimina-literal L (cdr C))))))

(defthm member-equal-elimina-literal
  (iff (member-equal L-1 (elimina-literal L-2 C))
       (and (member-equal L-1 C)
            (not (equal L-1 L-2)))))

(defthm elimina-literal-es-clausula
  (implies (es-clausula C)
           (es-clausula (elimina-literal L C))))

(defthm elimina-literal-true-listp
  (implies (true-listp C)
           (true-listp (elimina-literal L C))))

(defun reduce-forma-clausal-literal (L Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
        ((member-equal L (car Fc))
         (reduce-forma-clausal-literal L (cdr Fc)))
        (t (cons (elimina-literal (complementario L) (car Fc))
                 (reduce-forma-clausal-literal L (cdr Fc))))))

(defun reduce-forma-clausal-literal-C (L Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
        ((member-equal (complementario L) (car Fc))
         (reduce-forma-clausal-literal-C L (cdr Fc)))
        (t (cons (elimina-literal L (car Fc))
                 (reduce-forma-clausal-literal-C L (cdr Fc))))))

;;;----------------------------------------------------------------------------

(defthm reduce-forma-clausal-literal-C-equivalente
  (implies (es-literal L)
           (equal (reduce-forma-clausal-literal-C L Fc)
                  (reduce-forma-clausal-literal (complementario L) Fc)))
  :rule-classes :definition)

(in-theory (disable (:definition reduce-forma-clausal-literal-C-equivalente)))

;;;----------------------------------------------------------------------------

(defthm reduce-fc-literal-sin-clausulas-con-complementarios
  (implies (not (tiene-clausulas-con-complementarios Fc))
           (not (tiene-clausulas-con-complementarios
                 (reduce-forma-clausal-literal L Fc)))))

(defthm reduce-fc-literal-C-sin-clausulas-con-complementarios
  (implies (not (tiene-clausulas-con-complementarios Fc))
           (not (tiene-clausulas-con-complementarios
                 (reduce-forma-clausal-literal-C L Fc)))))

;;;----------------------------------------------------------------------------

(defthm es-forma-clausal-reduce-forma-clausal-literal
  (implies (es-forma-clausal Fc)
           (es-forma-clausal (reduce-forma-clausal-literal L Fc))))

(defthm es-forma-clausal-reduce-forma-clausal-literal-C
  (implies (es-forma-clausal Fc)
           (es-forma-clausal (reduce-forma-clausal-literal-C L Fc))))

;;;----------------------------------------------------------------------------

(defthm true-list-listp-reduce-forma-clausal-literal
  (implies (true-list-listp Fc)
           (true-list-listp (reduce-forma-clausal-literal L Fc))))

(defthm true-list-listp-reduce-forma-clausal-literal-C
  (implies (true-list-listp Fc)
           (true-list-listp (reduce-forma-clausal-literal-C L Fc))))

;;;----------------------------------------------------------------------------

(defthm completitud-reduce-forma-clausal-literal
  (implies (and (modelo-forma-clausal-K sigma Fc)
                (modelo-K sigma L)
                (es-literal L))
           (modelo-forma-clausal-K sigma 
				   (reduce-forma-clausal-literal L Fc))))

(defthm completitud-reduce-forma-clausal-literal-C
  (implies (and (modelo-forma-clausal-K sigma Fc)
                (not (modelo-K sigma L))
                (es-literal L))
           (modelo-forma-clausal-K sigma 
				   (reduce-forma-clausal-literal-C L Fc))))

(defthm modelo-clausula-K-elimina-literal-C
  (implies (and (not (member-equal L C))
                (es-literal L)
                (es-clausula C)
                (modelo-clausula-K sigma 
				   (elimina-literal (complementario L) C)))
           (modelo-clausula-K (asume-literal L sigma) C)))

(defthm correccion-reduce-forma-clausal-literal
  (implies (and (es-literal L)
                (es-forma-clausal Fc)
                (modelo-forma-clausal-K
                 sigma (reduce-forma-clausal-literal L Fc)))
           (modelo-forma-clausal-K (asume-literal L sigma) Fc))
  :hints (("Goal" :induct (reduce-forma-clausal-literal L Fc))))

(defthm correccion-reduce-forma-clausal-literal-C
  (implies (and (es-literal L)
                (es-forma-clausal Fc)
                (modelo-forma-clausal-K 
                 sigma (reduce-forma-clausal-literal-C L Fc)))
           (modelo-forma-clausal-K
	    (asume-literal (complementario L) sigma) Fc))
  :hints (("Goal" :induct (reduce-forma-clausal-literal-C L Fc))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Eliminacin de clusulas unitarias
;;;
;;;----------------------------------------------------------------------------

(defun clausula-unitaria (C)
  (declare (xargs :guard (true-listp C)))
  (and (consp C)
       (not (consp (elimina-literal (car C) C)))))

(defthm clausula-unitaria-no-contiene-otro
  (implies (and (clausula-unitaria C)
                (member-equal L-1 C)
                (not (equal L-1 L-2)))
           (not (member-equal L-2 C))))

(defthm clausula-unitaria-contiene-uno
  (implies (and (clausula-unitaria C)
                (member-equal L C))
           (not (consp (elimina-literal L C)))))

(defthm reduce-forma-clausal-literal-C-clausula-unitaria-contiene-nula
  (implies (and (member-equal C Fc)
                (clausula-unitaria C)
                (member-equal L C))
           (contiene-clausula-vacia (reduce-forma-clausal-literal-C L Fc)))
  :hints (("Goal" :in-theory (disable clausula-unitaria))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Eliminacin de literales puros
;;;
;;;----------------------------------------------------------------------------

(defun literal-en-forma-clausal (L Fc)
  (declare (xargs :guard (true-list-listp Fc)))

  (cond ((endp Fc) nil)
        (t (or (member-equal L (car Fc))
               (literal-en-forma-clausal L (cdr Fc))))))

(defun es-literal-puro (L Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (not (literal-en-forma-clausal (complementario L) Fc)))

;;;----------------------------------------------------------------------------

(defthm member-equal-reduce-forma-clausal-literal-C
  (implies (and (member-equal C Fc)
                (not (member-equal L C))
                (not (member-equal (complementario L) C)))
           (member-equal C (reduce-forma-clausal-literal-C L Fc))))

(defthm member-equal-reduce-forma-clausal-literal-2
  (implies (member-equal C (reduce-forma-clausal-literal L Fc))
           (and (not (member-equal L C))
                (not (member-equal (complementario L) C))))
  :hints (("Goal"
           :induct (reduce-forma-clausal-literal L Fc))))

(defthm elimina-literal-not-member
  (implies (not (member-equal L C))
           (equal (elimina-literal L C) C)))

(defthm reduce-forma-clausal-literal-literal-puro
  (implies (and (es-literal-puro L Fc)
                (member-equal C (reduce-forma-clausal-literal L Fc)))
           (member-equal C Fc)))

(defthm reduce-forma-clausal-literal-literal-puro-member
  (implies (and (es-literal-puro L Fc)
                (member-equal C (reduce-forma-clausal-literal L Fc)))
           (member-equal C (reduce-forma-clausal-literal-C L Fc)))
  :hints (("Goal" :in-theory (disable es-literal-puro))))

(include-book "../6-extensiones/listas")

(defthm reduce-forma-clausal-literal-C-literal-puro
  (implies (es-literal-puro L Fc)
           (subsetp-equal (reduce-forma-clausal-literal L Fc)
                          (reduce-forma-clausal-literal-C L Fc)))
  :hints ((extension-conjuntiva-a-listas
           (reduce-forma-clausal-literal L Fc)
           (lambda (elt)
             (member-equal elt (reduce-forma-clausal-literal-C L Fc)))
           (lambda (lst)
             (subsetp-equal lst (reduce-forma-clausal-literal-C L Fc)))))
  :rule-classes :forward-chaining)

;;;----------------------------------------------------------------------------

(defthm modelo-forma-clausal-K-member
  (implies (and (modelo-forma-clausal-K sigma Fc)
                (member-equal C Fc))
           (modelo-clausula-K sigma C)))

(defthm modelo-forma-clausal-K-subsetp-equal
  (implies (and (subsetp-equal Fc-1 Fc-2)
                (modelo-forma-clausal-K sigma Fc-2))
           (modelo-forma-clausal-K sigma Fc-1)))

;;;============================================================================
;;;
;;; 2) Decisin de satisfacibilidad por bifurcacin
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Funcin de seleccin de literal
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 (((seleccion *) => *))
 
 (local
  (defun seleccion (Fc)
    (declare (xargs :guard (true-list-listp Fc)))
    
    (cond ((endp Fc) nil)
          ((endp (car Fc))
           (seleccion (cdr Fc)))
          (t (car (car Fc))))))

 (defthm seleccion-esta-en-conjunto
   (implies (seleccion Fc)
            (literal-en-forma-clausal (seleccion Fc) Fc)))

 (defthm seleccion-es-literal
   (implies (and (es-forma-clausal Fc)
                 (seleccion Fc))
            (es-literal (seleccion Fc))))

 (defthm seleccion-existe
   (implies (and (es-forma-clausal Fc)
                 (consp Fc)
                 (consp (car Fc)))
            (seleccion Fc)))

 )

(defthm literal-en-forma-clausal-consp
  (implies (literal-en-forma-clausal L Fc)
	   (consp Fc))
  :rule-classes nil)

(defthm seleccion-consp
  (implies (seleccion Fc)
	   (consp Fc))
  :hints (("Goal" :in-theory (disable literal-en-forma-clausal)
	   :use (:instance literal-en-forma-clausal
			   (L (seleccion Fc))))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) Medida de terminacin
;;;
;;;----------------------------------------------------------------------------

(defun medida-forma-clausal (Fc)
  (declare (xargs :guard (true-list-listp Fc)))

  (cond ((endp Fc) 0)
        (t (+ (len (car Fc)) 
              (medida-forma-clausal (cdr Fc))))))

(defthm reduce-forma-clausal-literal-disminuye-medida
  (implies (literal-en-forma-clausal L Fc)
           (< (medida-forma-clausal (reduce-forma-clausal-literal L Fc))
              (medida-forma-clausal Fc))))

(defthm reduce-forma-clausal-literal-C-disminuye-medida
  (implies (literal-en-forma-clausal L Fc)
           (< (medida-forma-clausal (reduce-forma-clausal-literal-C L Fc))
              (medida-forma-clausal Fc))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.3) Definicin de SAT-bifurcacion
;;;
;;;----------------------------------------------------------------------------

(defun SAT-bifurcacion (Fc)
  (declare (xargs :guard (true-list-listp Fc)
                  :measure (medida-forma-clausal Fc)))
  (if (contiene-clausula-vacia Fc) 
      nil
      (let ((L (seleccion Fc)))
        (if L
            (or (SAT-bifurcacion (reduce-forma-clausal-literal L Fc))
		(SAT-bifurcacion (reduce-forma-clausal-literal-C L Fc)))
            t))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.4) SAT-bifurcacion es completo
;;;
;;;----------------------------------------------------------------------------

(in-theory (disable modelo-K complementario))

(defthm modelo-clausula-K-elimina-literal
  (implies (and (modelo-clausula-K sigma C)
                (not (modelo-K sigma L)))
           (modelo-clausula-K sigma (elimina-literal L C))))

(defthm completitud-SAT-bifurcacion
  (implies (and (modelo-forma-clausal-K sigma Fc)
                (es-forma-clausal Fc))
           (SAT-bifurcacion Fc))
  :hints (("Subgoal *1/3" :cases ((modelo-K sigma (seleccion Fc))))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.5) Construccin de un modelo: MOD-bifurcacion
;;;
;;;----------------------------------------------------------------------------

(defun MOD-bifurcacion (Fc sigma)
  (declare (xargs :guard (and (true-list-listp Fc) (alistp sigma))
                  :measure (medida-forma-clausal Fc)))
  (if (contiene-clausula-vacia Fc) 
      nil
      (let ((L (seleccion Fc)))
        (if L
            (or (MOD-bifurcacion
                 (reduce-forma-clausal-literal L Fc)
                 (asume-literal L sigma))
                (MOD-bifurcacion
                 (reduce-forma-clausal-literal-C L Fc)
                 (asume-literal (complementario L) sigma)))
            sigma))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.6) SAT-bifurcacion es completo
;;;
;;;----------------------------------------------------------------------------

(defthm SAT-bifurcacion-MOD-bifurcacion
  (implies (consp sigma)
	   (iff (MOD-bifurcacion Fc sigma)
		(SAT-bifurcacion Fc))))

(defthm modelo-forma-clausal-K-vacia
  (implies (and (es-forma-clausal Fc)
                (not (seleccion Fc))
                (not (contiene-clausula-vacia Fc)))
           (modelo-forma-clausal-K simga Fc)))

(defun literal-no-esta-en-conjunto (L Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) t)
        (t (and (not (member-equal L (car Fc)))
                (not (member-equal (complementario L) (car Fc)))
                (literal-no-esta-en-conjunto L (cdr Fc))))))

(defthm literal-no-esta-en-conjunto-reduce-forma-clausal-literal
  (implies (literal-no-esta-en-conjunto L-1 Fc)
           (literal-no-esta-en-conjunto 
            L-1 (reduce-forma-clausal-literal L-2 Fc))))

(defthm literal-no-esta-en-conjunto-reduce-forma-clausal-literal-C
  (implies (literal-no-esta-en-conjunto L-1 Fc)
           (literal-no-esta-en-conjunto 
            L-1 (reduce-forma-clausal-literal-C L-2 Fc))))

(defthm literal-no-esta-en-conjunto-distinto-miembro
  (implies (and (literal-no-esta-en-conjunto L-1 Fc)
                (literal-en-forma-clausal L-2 Fc)
		(es-literal L-1)
		(es-literal L-2))
           (and (not (equal L-2 L-1))
                (not (equal (complementario L-2) L-1)))))

(defthm reduce-forma-clausal-literal-no-contiene-literal
  (literal-no-esta-en-conjunto L (reduce-forma-clausal-literal L Fc)))

(defthm reduce-forma-clausal-literal-C-no-contiene-literal
  (literal-no-esta-en-conjunto L (reduce-forma-clausal-literal-C L Fc)))

(defthm literal-no-esta-en-conjunto-no-es-asignado-modelo
  (implies (and (es-forma-clausal Fc)
                (es-literal L)
                (literal-no-esta-en-conjunto L Fc)
                (MOD-bifurcacion Fc sigma)
                (modelo-K sigma L))
           (modelo-K (MOD-bifurcacion Fc sigma) L))
  :hints (("Goal"
           :induct (MOD-bifurcacion Fc sigma))))

(defthm modelo-asume-literal-cierto
  (implies (and (modelo-K sigma L-2)
                (es-literal L-1)
                (es-literal L-2))
           (iff (modelo-K (asume-literal L-2 sigma) L-1)
                (modelo-K sigma L-1)))
  :hints (("Goal" :cases ((not (equal l-1 l-2))))
          ("Subgoal 1" :cases ((not (equal l-1 (complementario l-2)))))))

(defthm modelo-clausula-K-asume-literal-cierto
  (implies (and (modelo-K sigma L)
                (es-literal L)
                (es-clausula C))
           (iff (modelo-clausula-K (asume-literal L sigma) C)
                (modelo-clausula-K sigma C))))

(defthm modelo-forma-clausal-K-asume-literal-cierto
  (implies (and (modelo-K sigma L)
                (es-literal L)
                (es-forma-clausal Fc))
           (iff (modelo-forma-clausal-K (asume-literal L sigma) Fc)
                (modelo-forma-clausal-K sigma Fc))))

(defthm modelo-forma-clausal-K-reduccion-literal
  (implies (and (SAT-bifurcacion (reduce-forma-clausal-literal L Fc))
                (es-literal L)
                (es-forma-clausal Fc))
           (iff (modelo-forma-clausal-K 
                 (MOD-bifurcacion (reduce-forma-clausal-literal L Fc)
                                       (asume-literal L sigma))
                 Fc)
                (modelo-forma-clausal-K
                 (asume-literal 
                  L (MOD-bifurcacion (reduce-forma-clausal-literal L Fc)
                                          (asume-literal L sigma)))
                 Fc))))

(in-theory (enable (:definition reduce-forma-clausal-literal-C-equivalente)))

(defthm correccion-SAT-bifurcacion
  (implies (and (es-forma-clausal Fc)
                (SAT-bifurcacion Fc))
           (modelo-forma-clausal-K (MOD-bifurcacion Fc sigma) Fc))
  :hints (("Goal" 
           :in-theory (disable modelo-forma-clausal-K-asume-literal-cierto))))

;;;============================================================================
;;;
;;; 3) Decisin de satisfacibilidad por Davis y Putnam
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;; 
;;; 3.1) Bsqueda de una clusula unitaria
;;;
;;;----------------------------------------------------------------------------

(defun busca-clausula-unitaria (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
        ((clausula-unitaria (car Fc))
         (car Fc))
        (t (busca-clausula-unitaria (cdr Fc)))))

(defthm busca-clausula-unitaria-esta-en-conjunto
  (implies (busca-clausula-unitaria Fc)
           (member-equal (busca-clausula-unitaria Fc) Fc)))

(defthm busca-clausula-unitaria-es-consp
  (implies (busca-clausula-unitaria Fc)
           (consp (busca-clausula-unitaria Fc))))

(defthm busca-clausula-unitaria-es-unitaria
  (implies (busca-clausula-unitaria Fc)
           (clausula-unitaria (busca-clausula-unitaria Fc))))

(defthm literal-de-clausula-unitaria-esta-en-conjunto
  (implies (busca-clausula-unitaria Fc)
           (literal-en-forma-clausal (car (busca-clausula-unitaria Fc)) Fc)))

(defthm literal-de-clausula-unitaria-existe
  (implies (and (busca-clausula-unitaria Fc)
                (es-forma-clausal Fc))
           (car (busca-clausula-unitaria Fc))))

(defthm literal-de-clausula-unitaria-es-literal
  (implies (and (busca-clausula-unitaria Fc)
                (es-forma-clausal Fc))
           (es-literal (car (busca-clausula-unitaria Fc)))))

;;;----------------------------------------------------------------------------
;;; 
;;; 3.2) Bsqueda de un literal puro
;;;
;;;----------------------------------------------------------------------------

(defun append-lista (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
        (t (append (car Fc)
                   (append-lista (cdr Fc))))))

(defun elimina-par-literal (L C)
  (declare (xargs :guard (true-listp C)))
  (cond ((endp C) nil)
        ((or (equal (car C) L)
             (equal (complementario (car C)) L))
         (elimina-par-literal L (cdr C)))
        (t (cons (car C) (elimina-par-literal L (cdr C))))))

(defthm elimina-par-literal-es-clausula
  (implies (es-clausula C)
           (es-clausula (elimina-par-literal L C))))

(defthm pertenencia-elimina-par-literal
  (iff (member-equal L-1 (elimina-par-literal L-2 C))
       (and (member-equal L-1 C)
            (not (equal L-1 L-2))
            (not (equal (complementario L-1) L-2))))
  :hints (("Goal" :induct (elimina-par-literal L-2 C))))

(defun busca-literal-sin-complementario (C)
  (declare (xargs :guard (true-listp C)
                  :measure (len C)))
  (cond ((endp C) nil)
        ((member-equal (complementario (car C)) (cdr C))
         (busca-literal-sin-complementario
          (elimina-par-literal (car C) (cdr C))))
        (t (car C))))

(defun busca-literal-puro (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (busca-literal-sin-complementario (append-lista Fc)))

(defthm busca-literal-sin-complementario-esta-en-original
  (implies (busca-literal-sin-complementario C)
           (member-equal (busca-literal-sin-complementario C) C))
  :rule-classes :forward-chaining)

(defthm busca-literal-sin-complementario-funciona
 (implies (and (busca-literal-sin-complementario C)
               (es-clausula C))
          (not (member-equal 
                (complementario (busca-literal-sin-complementario C))
                C)))
 :hints (("Subgoal *1/4"
          :use (:instance busca-literal-sin-complementario-esta-en-original
                          (C (elimina-par-literal (car C)
                                                  (cdr C)))))))

(defthm busca-literal-sin-complementario-es-literal
  (implies (and (es-clausula C)
                (busca-literal-sin-complementario C))
           (es-literal (busca-literal-sin-complementario C))))

(defthm append-lista-es-clausula
  (implies (es-forma-clausal Fc)
           (es-clausula (append-lista Fc))))

(defthm literal-puro-es-literal
  (implies (and (busca-literal-puro Fc)
                (es-forma-clausal Fc))
           (es-literal (busca-literal-puro Fc)))
  :hints (("Goal" :in-theory (disable es-literal))))

(defthm member-equal-append
  (iff (member-equal elemento (append lista1 lista2))
       (or (member-equal elemento lista1)
           (member-equal elemento lista2))))

(defthm literal-en-forma-clausal-append-lista
  (implies (member-equal L (append-lista Fc))
           (literal-en-forma-clausal L Fc)))

(defthm literal-puro-esta-en-conjunto
  (implies (busca-literal-puro Fc)
           (literal-en-forma-clausal (busca-literal-puro Fc) Fc)))

(defthm literal-en-forma-clausal-miembro-append-lista
  (iff (literal-en-forma-clausal (complementario L) Fc)
       (member-equal (complementario L) (append-lista Fc))))

(defthm busca-literal-puro-devuelve-literal-puro
  (implies (and (busca-literal-puro Fc)
                (es-forma-clausal Fc))
           (es-literal-puro (busca-literal-puro Fc) Fc)))

(in-theory (disable busca-literal-puro))

;;;----------------------------------------------------------------------------
;;;
;;; 3.3) Instancia de SAT-bifurcacion
;;;
;;;----------------------------------------------------------------------------

(defun seleccion-instancia (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (let ((C (busca-clausula-unitaria Fc)))
    (if C
        (car C)
        (let ((L (busca-literal-puro Fc)))
          (or L
              (if (endp Fc)
                  nil
                  (seleccion Fc)))))))

(defthm seleccion-instancia-esta-en-conjunto
  (implies (seleccion-instancia Fc)
           (literal-en-forma-clausal (seleccion-instancia Fc) Fc)))

(defthm seleccion-instancia-es-literal
  (implies (and (es-forma-clausal Fc)
                (seleccion-instancia Fc))
           (es-literal (seleccion-instancia Fc)))
  :hints (("Goal" :in-theory (disable es-literal))))

(defthm seleccion-instancia-existe
  (implies (and (es-forma-clausal Fc)
                (consp Fc)
                (consp (car Fc)))
           (seleccion-instancia Fc)))

(in-theory (disable seleccion-instancia))

(defun SAT-bifurcacion-instancia (Fc)
  (declare (xargs :guard (true-list-listp Fc)
                  :measure (medida-forma-clausal Fc)))
  (if (contiene-clausula-vacia Fc) 
      nil
      (let ((L (seleccion-instancia Fc)))
        (if L
            (or (SAT-bifurcacion-instancia 
		 (reduce-forma-clausal-literal L Fc))
                (SAT-bifurcacion-instancia 
		 (reduce-forma-clausal-literal-C L Fc)))
            t))))

(in-theory (disable es-literal))

(defthm completitud-SAT-bifurcacion-instancia
  (implies (and (es-forma-clausal Fc)
                (modelo-forma-clausal-K sigma Fc))
           (SAT-bifurcacion-instancia Fc))
  :hints (("Goal" 
           :by (:functional-instance 
                completitud-SAT-bifurcacion
                (SAT-bifurcacion SAT-bifurcacion-instancia)
                (seleccion seleccion-instancia)))))

(defun MOD-bifurcacion-instancia (Fc sigma)
  (declare (xargs :guard (and (true-list-listp Fc) (alistp sigma))
                  :measure (medida-forma-clausal Fc)))
  (if (contiene-clausula-vacia Fc) 
      nil
      (let ((L (seleccion-instancia Fc)))
        (if L
            (or (MOD-bifurcacion-instancia
                 (reduce-forma-clausal-literal L Fc)
                 (asume-literal L sigma))
                (MOD-bifurcacion-instancia
                 (reduce-forma-clausal-literal-C L Fc)
                 (asume-literal (complementario L) sigma)))
            sigma))))

(defthm correccion-SAT-bifurcacion-instancia
  (implies (and (es-forma-clausal Fc)
                (SAT-bifurcacion-instancia Fc))
           (modelo-forma-clausal-K (MOD-bifurcacion-instancia Fc sigma)
                                 Fc))
  :hints (("Goal" 
           :by (:functional-instance 
                correccion-SAT-bifurcacion
                (SAT-bifurcacion SAT-bifurcacion-instancia)
                (MOD-bifurcacion MOD-bifurcacion-instancia)
                (seleccion seleccion-instancia)))))

;;;---------------------------------------------------------------------------
;;;
;;; 3.4) Definicin de SAT-davis-putnam
;;;
;;;----------------------------------------------------------------------------

(defthm literal-en-forma-clausal-seleccion
  (implies (and (not (contiene-clausula-vacia Fc))
                (not (busca-clausula-unitaria Fc))
                (not (busca-literal-puro Fc))
                (seleccion Fc))
           (literal-en-forma-clausal (seleccion Fc) Fc)))

(defun SAT-davis-putnam (Fc)
  (declare (xargs :guard (true-list-listp Fc)
                  :measure (medida-forma-clausal Fc)))
  (cond 
   ((contiene-clausula-vacia Fc) nil)
   (t (let ((C (busca-clausula-unitaria Fc)))
        (if C 
            (SAT-davis-putnam (reduce-forma-clausal-literal (car C) Fc))
            (let ((L (busca-literal-puro Fc)))
              (if L
                  (SAT-davis-putnam (reduce-forma-clausal-literal L Fc))
                  (let ((L (seleccion Fc)))
                    (if L
                        (or (SAT-davis-putnam 
			     (reduce-forma-clausal-literal L Fc))
                            (SAT-davis-putnam 
			     (reduce-forma-clausal-literal-C L Fc)))
                        t)))))))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.5) Eliminacin de clusulas unitarias
;;;
;;;----------------------------------------------------------------------------

(defthm SAT-bifurcacion-instancia-clausula-unitaria
  (implies (and (es-forma-clausal Fc)
                (member-equal C Fc)
                (clausula-unitaria C))
           (not (SAT-bifurcacion-instancia 
                 (reduce-forma-clausal-literal-C (car C) Fc)))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.6) Eliminacin de literales puros
;;;
;;;----------------------------------------------------------------------------

(defthm SAT-bifurcacion-instancia-subsetp-equal
  (implies (and (subsetp-equal Fc-1 Fc-2)
                (es-forma-clausal Fc-1)
                (es-forma-clausal Fc-2)
                (not (SAT-bifurcacion-instancia Fc-1)))
           (not (SAT-bifurcacion-instancia Fc-2)))
  :hints (("Goal"
           :use ((:instance 
                  completitud-SAT-bifurcacion-instancia
                  (Fc Fc-1)
                  (sigma (MOD-bifurcacion-instancia Fc-2 nil)))))))

(defthm SAT-bifurcacion-instancia-literal-puro
  (implies (and (es-forma-clausal Fc)
                (es-literal-puro L Fc)
                (SAT-bifurcacion-instancia 
		 (reduce-forma-clausal-literal-C L Fc)))
           (SAT-bifurcacion-instancia (reduce-forma-clausal-literal L Fc)))
  :hints (("Goal" :in-theory (disable es-literal-puro))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.7) Correccin y completitud de SAT-davis-putnam
;;;
;;;----------------------------------------------------------------------------

(in-theory (enable seleccion-instancia))

(defthm SAT-davis-putnam-y-SAT-bifurcacion-equivalentes
  (implies (es-forma-clausal Fc)
           (iff (SAT-davis-putnam Fc)
                (SAT-bifurcacion-instancia Fc)))
  :hints (("Goal" 
           :in-theory (disable contiene-clausula-vacia
                               clausula-unitaria
                               reduce-forma-clausal-literal
                               reduce-forma-clausal-literal-c
                               reduce-forma-clausal-literal-c-equivalente))
          ("Subgoal *1/3''" :expand (SAT-bifurcacion-instancia Fc))))

(defthm completitud-SAT-davis-putnam
  (implies (and (es-forma-clausal Fc)
                (modelo-forma-clausal-K sigma Fc))
           (SAT-davis-putnam Fc)))

(defthm correccion-SAT-davis-putnam
  (implies (and (es-forma-clausal Fc)
                (SAT-davis-putnam Fc))
           (modelo-forma-clausal-K (MOD-bifurcacion-instancia Fc sigma) Fc)))

;;;============================================================================
;;;
;;; 4) Ejemplos de uso
;;;
;;;============================================================================

#|

;;;   Un ejemplo de utilizacin de este DAT es el siguiente:

(include-book "../8-davisputnam/davis-putnam")

;;;   Necesitamos una funcin de seleccin concreta, para ello indicamos que se
;;; va a permitir redefiniciones en el sistema.

(set-ld-redefinition-action '(:warn . :overwrite) state)

(defun seleccion (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  
  (cond ((endp Fc) nil)
	((endp (car Fc))
	 (seleccion (cdr Fc)))
	(t (car (car Fc)))))

|#

;;;============================================================================