;;;============================================================================
;;;
;;; Ttulo: Clusulas y formas clausales (II)
;;; Parte del proyecto de tesis:
;;;    "Verificacin de sistemas de razonamiento automtico"
;;; FJMM 21-05-2001
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(defpkg "MUL" (union-eq *acl2-exports*
			*common-lisp-symbols-from-main-lisp-package*))

(certify-book "../8-davisputnam/forma-clausal" 1)

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin.
;;;
;;;============================================================================

(include-book "../4-tableros/uniforme")

(include-book "../8-davisputnam/clausulas")

(in-theory (disable alfa-formula beta-formula doble-negacion
		    componente-1 componente-2 componente-neg-neg
		    otra-definicion-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) Transformacin a forma clausal
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Listas de frmulas y listas de listas de frmulas
;;;
;;;----------------------------------------------------------------------------

(defun lista-formulas (Gamma)
  (declare (xargs :guard (true-listp Gamma)))
  (or (endp Gamma)
      (and (es-proposicional (car Gamma))
	   (lista-formulas (cdr Gamma)))))

(defun lista-lista-formulas (S)
  (declare (xargs :guard (true-list-listp S)))
  (or (endp S)
      (and (lista-formulas (car S))
	   (lista-lista-formulas (cdr S)))))

(defthm member-lista-formulas-es-proposicional 
  (implies (and (lista-formulas Gamma)
		(member-equal F Gamma))
	   (es-proposicional F)))

(defthm member-lista-lista-formulas-es-proposicional 
  (implies (and (lista-lista-formulas S)
		(member-equal Gamma S))
	   (lista-formulas Gamma)))

(defthm lista-lista-formulas-append
  (iff (lista-lista-formulas (append S-1 S-2))
       (and (lista-lista-formulas S-1)
	    (lista-lista-formulas S-2))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Funcin de seleccin
;;;
;;;----------------------------------------------------------------------------

(defun selecciona-no-literal (Gamma)
  (declare (xargs :guard (true-listp Gamma)))
  (cond ((endp Gamma) nil)
	((es-literal (car Gamma))
	 (selecciona-no-literal (cdr Gamma)))
	(t (car Gamma))))

(defthm selecciona-no-literal-member-equal
  (implies (selecciona-no-literal Gamma)
	   (member-equal (selecciona-no-literal Gamma) Gamma)))

(defthm selecciona-no-literal-es-no-literal
  (implies (selecciona-no-literal Gamma)
	   (not (es-literal (selecciona-no-literal Gamma)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Un paso de transformacin
;;;
;;;----------------------------------------------------------------------------

(defun elimina-una (elt lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) lista)
	((equal (car lista) elt) (cdr lista))
	(t (cons (car lista) (elimina-una elt (cdr lista))))))

(defthm true-listp-elimina-una
  (implies (true-listp Gamma)
	   (true-listp (elimina-una F Gamma))))

(defthm lista-formulas-elimina-una
  (implies (lista-formulas Gamma)
	   (lista-formulas (elimina-una F Gamma))))

(defun aade-formula (F Gamma)
  (declare (xargs :guard (true-listp Gamma)))
  (cond ((member-equal F Gamma) (list Gamma))
	((member-equal (complementario F) Gamma) nil)
	(t (list (cons F Gamma)))))

(defun aade-formulas (Fs Gamma)
  (declare (xargs :guard (and (true-listp Fs)
			      (true-listp Gamma))))
  (cond ((endp Fs) (list Gamma))
	(t (let ((aade-resto (aade-formulas (cdr Fs) Gamma)))
	     (if aade-resto
		 (aade-formula (car Fs) (car aade-resto))
		 nil)))))

(defun forma-clausal-procesa-una (F Gamma)
  (declare (xargs :guard (true-listp Gamma)))
  (let ((Gamma-F (elimina-una F Gamma)))
    (cond ((doble-negacion F)
	   (aade-formula (componente-neg-neg F) 
			  Gamma-F))
	  ((beta-formula F)
	   (aade-formulas (list (componente-1 F)
				 (componente-2 F)) Gamma-F))
	  ((alfa-formula F)
	   (append (aade-formula (componente-1 F) Gamma-F)
		   (aade-formula (componente-2 F) Gamma-F)))
	  (t '(nil)))))

(defthm true-list-listp-forma-clausal-procesa-una
  (implies (true-listp Gamma)
	   (and (true-list-listp (forma-clausal-procesa-una F Gamma))
		(true-listp (forma-clausal-procesa-una F Gamma))))
  :hints (("Goal"
	   :expand ((forma-clausal-procesa-una F Gamma)))))

(defthm lista-lista-formulas-forma-clausal-procesa-una
  (implies (and (lista-formulas Gamma)
		(member-equal F Gamma))
	   (lista-lista-formulas (forma-clausal-procesa-una F Gamma)))
  :hints (("Goal"
	   :in-theory (disable definicion-uniforme-de-es-proposicional
			       es-literal)
	   :expand ((forma-clausal-procesa-una F Gamma)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Transformacin a forma clausal: forma-clausal
;;;
;;;----------------------------------------------------------------------------

(defun forma-clausal-aux (S)
  (declare (xargs :mode :program
		  :guard (true-list-listp S)))
  (if (endp S)
      nil
      (let ((F (selecciona-no-literal (car S))))
	(if F
	    (forma-clausal-aux
	     (append (forma-clausal-procesa-una F (car S))
		     (cdr S)))
	    (cons (car S) (forma-clausal-aux (cdr S)))))))

(defun forma-clausal (F)
  (declare (xargs :mode :program))
  (forma-clausal-aux (list (list F))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Prueba de terminacin
;;;
;;;----------------------------------------------------------------------------

(defun medida-uniforme-lista (Gamma)
  (declare (xargs :guard (true-listp Gamma)))
  (cond ((endp Gamma) nil)
	(t (cons (medida-uniforme (car Gamma))
		 (medida-uniforme-lista (cdr Gamma))))))

(include-book "../6-extensiones/defmul")

; (defmul-components e0-ord-<)
; => (E0-ORD-< NIL E0-ORDINALP E0-ORD-<-FN NIL NIL)

(defmul (E0-ORD-< NIL E0-ORDINALP E0-ORD-<-FN NIL NIL)
  :verify-guards t)

(defthm e0-ordinalp-true-listp-medida-uniforme-lista
  (e0-ordinalp-true-listp (medida-uniforme-lista Gamma)))

(encapsulate
 ()

 (local
  (defthm mul-e0-ord-<-nil
    (implies (consp lista)
	     (mul-e0-ord-< nil lista))))

 (local
  (in-theory (disable mul-e0-ord-< complementario)))

 (local
  (defthm medida-uniforme-lista-forma-clausal-procesa-una-aux
    (implies (and (member-equal F Gamma)
		  (member-equal N-Gamma (forma-clausal-procesa-una F Gamma)))
	     (mul-e0-ord-< (medida-uniforme-lista N-Gamma)		  
			   (cons (medida-uniforme F)
				 (medida-uniforme-lista 
				  (elimina-una F Gamma)))))
    :rule-classes nil))

 (defthm medida-uniforme-lista-elimina-una
   (implies (member-equal F Gamma)
	    (mul::multiset-equal 
	     (cons (medida-uniforme F)
		   (medida-uniforme-lista (elimina-una F Gamma)))
	     (medida-uniforme-lista Gamma))))

 (defthm medida-uniforme-lista-forma-clausal-procesa-una-instance 
   (implies (and (selecciona-no-literal Gamma)
		 (member-equal N-Gamma 
			       (forma-clausal-procesa-una
				(selecciona-no-literal Gamma) Gamma)))
	    (mul-e0-ord-< (medida-uniforme-lista N-Gamma)		  
			  (medida-uniforme-lista Gamma)))
   :hints (("Goal"
	    :use
	    (:instance medida-uniforme-lista-forma-clausal-procesa-una-aux
		       (F (selecciona-no-literal Gamma)))
	    :in-theory (disable selecciona-no-literal))))

 )

(defun medida-uniforme-forma-clausal (S)
  (declare (xargs :guard (true-list-listp S)))
  (cond ((endp S) nil)
	(t (cons (medida-uniforme-lista (car S))
		 (medida-uniforme-forma-clausal (cdr S))))))

(defthm medida-uniforme-forma-clausal-append
  (equal (medida-uniforme-forma-clausal (append S-1 S-2))
	 (append (medida-uniforme-forma-clausal S-1)
		 (medida-uniforme-forma-clausal S-2))))

; (defmul-components mul-e0-ord-<)
; => (MUL-E0-ORD-< MULTISET-EXTENSION-OF-E0-ORD-<-WELL-FOUNDED
;                  E0-ORDINALP-TRUE-LISTP MAP-E0-ORD-<-FN-E0-ORD X Y)

(defmul (MUL-E0-ORD-< MULTISET-EXTENSION-OF-E0-ORD-<-WELL-FOUNDED
		      E0-ORDINALP-TRUE-LISTP
		      MAP-E0-ORD-<-FN-E0-ORD X Y)
  :verify-guards t)

(defthm e0-ordinalp-true-listp-true-listp-medida-uniforme-forma-clausal
  (e0-ordinalp-true-listp-true-listp (medida-uniforme-forma-clausal S)))

(include-book "../6-extensiones/listas")

(in-theory (disable forma-clausal-procesa-una))

(defthm forall-exists-mul-e0-ord-<-bigger-forma-clausal-procesa-una
  (implies (selecciona-no-literal Gamma)
	   (forall-exists-mul-e0-ord-<-bigger
	    (medida-uniforme-forma-clausal
	     (forma-clausal-procesa-una (selecciona-no-literal Gamma) Gamma))
	    (list (medida-uniforme-lista Gamma))))
  :hints ((extension-conjuntiva-a-listas
	   (forma-clausal-procesa-una (selecciona-no-literal Gamma) Gamma)
	   (lambda (elt)
	     (exists-mul-e0-ord-<-bigger 
	      (medida-uniforme-lista elt)
	      (list (medida-uniforme-lista Gamma))))
	   (lambda (lst)
	     (forall-exists-mul-e0-ord-<-bigger
	      (medida-uniforme-forma-clausal lst)
	      (list (medida-uniforme-lista Gamma)))))))

(encapsulate
 ()

 (local
  (defthm mul::multiset-diff-append-cons
    (mul::multiset-equal (mul::multiset-diff (append l1 l2) (cons e l2))
			 (mul::multiset-diff l1 (list e)))
    :hints (("Goal"
	     :in-theory (disable mul::multiset-diff-append-2-multiset-equal)
	     :use (:instance mul::multiset-diff-append-2-multiset-equal
			     (mul::m1 l1)
			     (mul::m2 l2)
			     (mul::m3 (list e)))))))
  
 (local
  (defthm mul::multiset-diff-cons-append
    (mul::multiset-equal (mul::multiset-diff (cons e l2) (append l1 l2))
			 (mul::multiset-diff (list e) l1))
    :hints (("Goal"
	     :in-theory (disable mul::multiset-diff-append-2-multiset-equal)
	     :use (:instance mul::multiset-diff-append-2-multiset-equal
			     (mul::m3 l1)
			     (mul::m2 l2)
			     (mul::m1 (list e)))))))

 (defthm mul-mul-e0-ord-<-append-cons
   (iff (mul-mul-e0-ord-< (append lm1 lm2) (cons e lm2))
	(mul-mul-e0-ord-< lm1 (list e)))
   :hints (("Goal" :in-theory (disable mul::multiset-diff))))

 )

(defthm true-list-listp-append
  (implies (and (true-list-listp lista-1)
		(true-list-listp lista-2))
	   (true-list-listp (append lista-1 lista-2))))

(verify-termination forma-clausal-aux
  (declare (xargs :measure (medida-uniforme-forma-clausal S)
		  :well-founded-relation mul-mul-e0-ord-<
		  :hints (("Goal" :in-theory (disable mul-mul-e0-ord-<))))))

(verify-termination forma-clausal)

;;;----------------------------------------------------------------------------
;;;
;;; 1.6) La forma-clausal genera formas clausales
;;;
;;;----------------------------------------------------------------------------

(defthm forma-clausal-aux-es-forma-clausal
  (implies (lista-lista-formulas S)
	   (es-forma-clausal (forma-clausal-aux S))))

(defthm forma-clausal-es-forma-clausal
  (implies (es-proposicional F)
	   (es-forma-clausal (forma-clausal F))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.7) Equivalencia lgica de la forma clausal
;;;
;;;----------------------------------------------------------------------------

(defthm valor-forma-clausal-aade-formula
  (implies (es-proposicional F)
	   (equal (valor-forma-clausal (aade-formula F Gamma) sigma)
		  (valor-clausula (cons F Gamma) sigma)))
  :hints (("Goal" 
	   :in-theory (disable definicion-uniforme-de-es-proposicional))))

(encapsulate
 ()

 (local
  (defthm valor-forma-clausal-aade-formulas-valor-clausula-car
    (implies 
     (and (lista-formulas lista-F)
	  (aade-formulas lista-F Gamma))
     (equal (valor-forma-clausal (aade-formulas lista-F Gamma) sigma)
	    (valor-clausula (car (aade-formulas lista-F Gamma)) sigma)))))

 (defthm valor-forma-clausal-aade-formulas
   (implies 
    (lista-formulas lista-F)
    (equal (valor-forma-clausal (aade-formulas lista-F Gamma) sigma)
	   (valor-clausula (append lista-F Gamma) sigma)))
   :hints (("Goal" :in-theory (disable aade-formula))))

 )

(defthm es-proposicional-casos
  (implies (and (es-proposicional F)
		(not (es-literal F))
		(not (doble-negacion F))
		(not (beta-formula F))
		(not (alfa-formula F)))
	   nil)
  :rule-classes :forward-chaining)

(defthm valor-forma-clausal-forma-clausal-procesa-una
  (implies (and (es-proposicional F)
		(not (es-literal F))
		(member-equal F Gamma))
	   (equal (valor-forma-clausal 
		   (forma-clausal-procesa-una F Gamma) sigma)
		  (valor-clausula (cons F (elimina-una F Gamma)) sigma)))
  :hints (("Goal"
 	   :in-theory (disable definicion-uniforme-de-es-proposicional
 			       es-literal aade-formula aade-formulas)
 	   :expand ((forma-clausal-procesa-una F Gamma)))))

(defthm valor-clasula-elimina-una
  (implies (member-equal F Gamma)
	   (equal (valor-clausula (cons F (elimina-una F Gamma)) sigma)
		  (valor-clausula Gamma sigma))))
	   
(defthm valor-forma-clausal-forma-clausal-aux
  (implies (lista-lista-formulas S)
	   (equal (valor-forma-clausal (forma-clausal-aux S) sigma)
		  (valor-forma-clausal S sigma))))

(defthm valor-forma-clausal-forma-clausal
  (implies (es-proposicional F)
	   (equal (valor F sigma)
		  (valor-forma-clausal (forma-clausal F) sigma)))
  :hints (("Goal"
	   :in-theory (disable definicion-uniforme-de-es-proposicional))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.8) En la forma clausal no hay clusulas con literales complementarios
;;;
;;;----------------------------------------------------------------------------

(defthm tiene-literales-complementarios-cons
  (implies (tiene-literales-complementarios Gamma)
	   (tiene-literales-complementarios (cons F Gamma))))

(defthm member-equal-elimina-una
  (implies (member-equal F-1 (elimina-una F-2 Gamma))
	   (member-equal F-1 Gamma)))

(defthm tiene-formulas-complementarias-elimina-una
  (implies (not (tiene-literales-complementarios Gamma))
	   (not (tiene-literales-complementarios (elimina-una F Gamma)))))

(defthm miembro-forma-clausal-procesa-una-no-tiene-literales-complementarios
  (implies (and (not (tiene-literales-complementarios Gamma-1))
		(member-equal F Gamma-1)
		(member-equal Gamma-2 (forma-clausal-procesa-una F Gamma-1)))
	   (not (tiene-literales-complementarios Gamma-2)))
  :hints (("Goal"
	   :expand ((forma-clausal-procesa-una F Gamma-1)))))

(defun tiene-clausulas-con-complementarios (Fc)
  (declare (xargs :guard (true-list-listp Fc)))
  (cond ((endp Fc) nil)
        ((tiene-literales-complementarios (car Fc)) t)
        (t (tiene-clausulas-con-complementarios (cdr Fc)))))

(defthm not-tiene-clausulas-con-complementarios-member
  (implies (and (member-equal C Fc)
		(not (tiene-clausulas-con-complementarios Fc)))
	   (not (tiene-literales-complementarios C))))

(defthm forma-clausal-procesa-una-no-tiene-clausulas-con-complementarios
  (implies (and (not (tiene-literales-complementarios Gamma))
		(member-equal F Gamma))
	   (not (tiene-clausulas-con-complementarios
		 (forma-clausal-procesa-una F Gamma))))
  :hints ((extension-conjuntiva-a-listas
	   (forma-clausal-procesa-una F Gamma)
	   (lambda (elt)
	     (not (tiene-literales-complementarios elt)))
	   (lambda (lst)
	     (not (tiene-clausulas-con-complementarios lst))))))

(defthm tiene-clausulas-con-complementarios-append
  (iff (tiene-clausulas-con-complementarios (append Fc-1 Fc-2))
       (or (tiene-clausulas-con-complementarios Fc-1)
	   (tiene-clausulas-con-complementarios Fc-2))))

(defthm forma-clausal-aux-no-tiene-clausulas-con-complementarios
  (implies (not (tiene-clausulas-con-complementarios Fc))
	   (not (tiene-clausulas-con-complementarios 
		 (forma-clausal-aux Fc)))))

(defthm forma-clausal-no-tiene-clausulas-con-complementarios
  (implies (es-proposicional F)
	   (not (tiene-clausulas-con-complementarios (forma-clausal F)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.9) La forma clausal obtenida caracteriza frmulas vlidas
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-validez-forma-clausal
  (implies (and (es-proposicional F)
		(equal (forma-clausal F) nil))
	   (modelo sigma F))
  :hints (("Goal" :in-theory (enable modelo))))

(defthm member-forma-clausal-no-tiene-literales-complementarios
  (implies (and (es-proposicional F)
		(member-equal C (forma-clausal F)))
	   (not (tiene-literales-complementarios C))))

(in-theory (disable forma-clausal definicion-uniforme-de-es-proposicional))

(defthm member-forma-clausal-es-clausula
  (implies (and (es-proposicional F)
		(member-equal C (forma-clausal F)))
	   (es-clausula C))
  :hints (("Goal" :use (:instance member-es-forma-clausal-es-clausula
				  (Fc (forma-clausal F))))))

(defthm member-forma-clausal-tiene-contramodelo
  (implies (and (es-proposicional F)
		(member-equal C (forma-clausal F)))
	   (not (modelo-clausula (contramodelo-clausula C) C))))

(defthm forma-clausal-tiene-contramodelo
  (implies (and (es-proposicional F)
		(forma-clausal F))
	   (not (modelo-forma-clausal 
		 (contramodelo-clausula (car (forma-clausal F))) 
		 (forma-clausal F)))))

(defthm modelo-forma-clausal-forma-clausal
  (implies (es-proposicional F)
	   (iff (modelo sigma F)
		(modelo-forma-clausal sigma (forma-clausal F))))
  :hints (("Goal"
	   :in-theory (enable modelo-forma-clausal-valor-forma-clausal
			      modelo))))

(defthm completitud-validez-forma-clausal
  (implies (and (es-proposicional F)
		(forma-clausal F))
	   (not (modelo (contramodelo-clausula (car (forma-clausal F))) F))))

;;;============================================================================