;;;============================================================================
;;;
;;; Ttulo: Resolucin: Conjuntos representantes
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../9-resolucion/representantes")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../9-resolucion/conjuntos")

;;;============================================================================
;;;
;;; 1) Conjuntos representantes
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Conjuntos representantes
;;;
;;;----------------------------------------------------------------------------

(defun conjunto-representante (R S)
  (declare (xargs :guard (and (true-listp R)
                              (true-list-listp S))))
  (cond ((endp S) t)
        ((not (conjunto-vacio (interseccion R (car S))))
         (conjunto-representante R (cdr S)))
        (t nil)))

(encapsulate
 ()

 (local
  (defthm conjunto-representante-propiedad-aux
    (implies (and (conjunto-representante R S)
                  (conjunto-miembro C S))
             (not (conjunto-vacio 
                   (interseccion R (conjunto-miembro-testigo C S)))))
    :rule-classes nil))

 (defthm conjunto-representante-propiedad
   (implies (and (conjunto-representante R S)
                 (conjunto-miembro C S))
            (not (conjunto-vacio (interseccion R C))))
   :hints (("Goal" :use conjunto-representante-propiedad-aux)))
 
 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Conjunto tstigo
;;;
;;;----------------------------------------------------------------------------

(defun conjunto-testigo (R S L)
  (declare (xargs :guard (and (true-listp R)
                              (true-list-listp S))))
  (cond ((endp S) nil)
        ((and (member-equal L (car S))
              (conjunto-unitario (interseccion R (car S))))
         (car S))
        (t (conjunto-testigo R (cdr S) L))))

(defthm conjunto-testigo-pertenece-a-la-coleccion
  (implies (conjunto-testigo R S L)
           (conjunto-miembro (conjunto-testigo R S L) S))
  :hints (("Goal" :in-theory (disable conjunto-unitario))))

(defthm conjunto-testigo-propiedades
  (implies (and (conjunto-testigo R S L)
                (member-equal L R))
           (and (member-equal L (conjunto-testigo R S L))
                (conjunto-unitario 
                 (interseccion R (conjunto-testigo R S L)))))
  :hints (("Goal" :in-theory (disable conjunto-unitario))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Conjunto representante minimal
;;;
;;;----------------------------------------------------------------------------

(defun conjunto-representante-minimal-aux (R-aux R S)
  (declare (xargs :guard (and (true-listp R-aux)
                              (true-listp R)
                              (true-list-listp S))))
  (cond ((endp R-aux) t)
        ((conjunto-testigo R S (car R-aux))
         (conjunto-representante-minimal-aux (cdr R-aux) R S))
        (t nil)))

(defthm conjunto-representante-minimal-aux-tiene-testigos
  (implies (and (conjunto-representante-minimal-aux R-aux R S)
                (member-equal L R-aux))
           (conjunto-testigo R S L)))

(defun conjunto-representante-minimal (R S)
  (declare (xargs :guard (and (true-listp R)
                              (true-list-listp S))))
  (and (conjunto-representante R S)
       (conjunto-representante-minimal-aux R R S)))

(defthm conjunto-representante-minimal-tiene-testigos
  (implies (and (conjunto-representante-minimal R S)
                (member-equal L R))
           (conjunto-testigo R S L)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Caracterizacin de los conjuntos representantes minimales
;;;
;;;----------------------------------------------------------------------------

(defthm conjunto-representante-minimal-member
  (implies (and (conjunto-representante-minimal R S)
                (subsetp-equal R-aux R)
                (conjunto-representante R-aux S)
                (member-equal L R))
           (member-equal L R-aux))
  :hints (("Goal"
           :use (:instance
                 unicidad-en-conjunto-unitario-interseccion
                 (elemento L)
                 (conjunto1 R)
                 (conjunto2 R-aux)
                 (conjunto3 (conjunto-testigo R S L))))))

(include-book "../6-extensiones/listas")

(in-theory (disable conjunto-representante-minimal))

(defthm conjunto-representante-minimal-subsetp
  (implies (and (conjunto-representante-minimal R S)
                (subsetp-equal R-aux R)
                (conjunto-representante R-aux S))
            (subsetp-equal R R-aux))
  :hints ((extension-conjuntiva-a-listas
           R
           (lambda (elt)
             (member-equal elt R-aux))
           (lambda (lst)
             (subsetp-equal lst R-aux)))))

(defthm conjunto-representante-minimal-equivalencia-2->1
  (implies (and (conjunto-representante-minimal R S)
                (subsetp-equal R-aux R)
                (not (subsetp-equal R R-aux)))
           (not (conjunto-representante R-aux S))))

(in-theory (enable conjunto-representante-minimal))

;;;----------------------------------------------------------------------------

(encapsulate
 (((hip-R) => *)
  ((hip-S) => *))

 (local
  (defun hip-R ()
    'nil))
 
 (local
  (defun hip-S ()
    'nil))

 (defthm hip-R-es-conjunto-representante
   (conjunto-representante (hip-R) (hip-S)))

 (defthm hip-R-no-tiene-subconjuntos-propios-representantes
   (implies (and (subsetp-equal R-aux (hip-R))
                 (not (subsetp-equal (hip-R) R-aux)))
            (not (conjunto-representante R-aux (hip-S)))))

 )

(defthm conjunto-representante-reducido
  (implies (and (conjunto-representante R S)
		(member-equal L R)
		(not (conjunto-testigo R S L)))
	   (conjunto-representante (elimina-elemento L R) S))
  :hints (("Goal" 
	   :induct (conjunto-testigo R S L))))

(defthm conjunto-representante-minimal-tiene-conjunto-testigo
  (implies (member-equal L (hip-R))
	   (conjunto-testigo (hip-R) (hip-S) L))
  :hints (("Goal"
	   :in-theory (disable conjunto-representante-reducido)
	   :use (:instance conjunto-representante-reducido
			   (R (hip-R))
			   (S (hip-S))))))

(defthm conjunto-representante-minimal-equivalencia-1->2
  (conjunto-representante-minimal (hip-R) (hip-S))
  :hints ((extension-conjuntiva-a-listas
	   (hip-R)
	   (lambda (elt)
	     (conjunto-testigo (hip-R) (hip-S) elt))
	   (lambda (lst)
	     (conjunto-representante-minimal-aux
	      lst (hip-R) (hip-S))))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Obtencin de un conjunto de representantes
;;;
;;;----------------------------------------------------------------------------

(defun union-coleccion (S)
  (declare (xargs :guard (true-list-listp S)))
  (cond ((endp S) nil)
        (t (union-equal (car S)
                        (union-coleccion (cdr S))))))

(defun coleccion-de-conjuntos-no-vacios (S)
  (declare (xargs :guard (true-list-listp S)))
  (cond ((endp S) t)
        (t (and (not (conjunto-vacio (car S)))
                (coleccion-de-conjuntos-no-vacios (cdr S))))))

(defthm member-coleccion-de-conjuntos-no-vacios
  (implies (and (coleccion-de-conjuntos-no-vacios S)
                (member-equal conjunto S))
           (not (conjunto-vacio conjunto))))

(defthm member-equal-union-coleccion
  (implies (and (member-equal C S)
                (member-equal L C))
           (member-equal L (union-coleccion S))))

(defthm consp-interseccion-union-coleccion
  (implies (and (member-equal C S)
                (not (conjunto-vacio C)))
           (not (conjunto-vacio (interseccion (union-coleccion S) C))))
  :hints (("Goal"
           :use (:instance interseccion-existe
                           (elemento (car C))
                           (conjunto1 (union-coleccion S))
                           (conjunto2 C)))))

(in-theory (disable conjunto-vacio))

(defthm union-coleccion-es-conjunto-representante
  (implies (coleccion-de-conjuntos-no-vacios S)
           (conjunto-representante (union-coleccion S) S))
  :hints ((extension-conjuntiva-a-listas
           S
           (lambda (elt)
             (not (conjunto-vacio (interseccion (union-coleccion S) elt))))
           (lambda (lst)
             (conjunto-representante (union-coleccion S) lst)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Construccin de un conjunto representante minimal
;;;
;;;----------------------------------------------------------------------------

(defun elemento-sin-conjunto-testigo (R-aux R S)
  (declare (xargs :guard (and (true-listp R-aux)
                              (true-listp R)
                              (true-list-listp S))))
  (cond ((endp R-aux) nil)
        ((conjunto-testigo R S (car R-aux))
         (elemento-sin-conjunto-testigo (cdr R-aux) R S))
        (t (list (car R-aux)))))

(defthm elemento-sin-conjunto-testigo-member-equal
  (implies (elemento-sin-conjunto-testigo R-aux R S)
           (member-equal (car (elemento-sin-conjunto-testigo R-aux R S))
                         R-aux)))

(defthm elemento-sin-conjunto-testigo-no-tiene-conjunto-testigo
  (implies (elemento-sin-conjunto-testigo R-aux R S)
           (not (conjunto-testigo 
                 R S (car (elemento-sin-conjunto-testigo R-aux R S))))))

(defun reduccion-conjunto-representante (R S)
  (declare (xargs :guard (and (true-listp R)
                              (true-list-listp S))))
  (if (elemento-sin-conjunto-testigo R R S)
      (reduccion-conjunto-representante
       (elimina-elemento (car (elemento-sin-conjunto-testigo R R S)) R) S)
      R))

(defthm reduccion-conjunto-representante-es-conjunto-representante
  (implies (conjunto-representante R S)
           (conjunto-representante (reduccion-conjunto-representante R S) S)))

(defthm existe-elemento-sin-conjunto-testigo
  (implies (and (member-equal L R-aux)
                (not (conjunto-testigo R S L)))
           (elemento-sin-conjunto-testigo R-aux R S)))

(defthm reduccion-conjunto-representante-es-minimal-elemento
  (implies (and (conjunto-representante R S)
                (member-equal L (reduccion-conjunto-representante R S)))
           (conjunto-testigo (reduccion-conjunto-representante R S) S L)))

(defthm reduccion-conjunto-representante-es-minimal
  (implies (conjunto-representante R S)
           (conjunto-representante-minimal
            (reduccion-conjunto-representante R S) S))
  :hints ((extension-conjuntiva-a-listas
           (reduccion-conjunto-representante R S)
           (lambda (elt)
             (conjunto-testigo (reduccion-conjunto-representante R S) S elt))
           (lambda (lst)
             (conjunto-representante-minimal-aux
              lst (reduccion-conjunto-representante R S) S)))))

(defun construye-conjunto-representante-minimal (S)
  (declare (xargs :guard (true-list-listp S)))
  (reduccion-conjunto-representante (union-coleccion S) S))

(defthm construye-conjunto-representante-minimal-es-minimal
  (implies (coleccion-de-conjuntos-no-vacios S)
           (conjunto-representante-minimal
            (construye-conjunto-representante-minimal S) S)))

;;;============================================================================