;;;============================================================================
;;;
;;; Ttulo: Resolucin: Propiedades
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../9-resolucion/resolucion-aux")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;; 
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../9-resolucion/resolucion-sat")

;;;============================================================================
;;;
;;; 1) El proceso devuelve una forma clausal
;;;
;;;============================================================================

(encapsulate
 ()
 
 (local
  (in-theory (enable resolucion-cond-clausulas
                     resolucion-cond-clausula-conjunto)))

 (local
  (defthm resolucion-cond-clausulas-es-clausula
    (implies (and (es-clausula C-1)
                  (es-clausula C-2))
             (es-clausula (car (resolucion-cond-clausulas C-1 C-2))))))

 (defthm resolucion-cond-clausula-conjunto-es-forma-clausal
   (implies (and (es-clausula C)
                 (es-forma-clausal S-2)
                 (es-forma-clausal S-N))
            (es-forma-clausal
             (resolucion-cond-clausula-conjunto C S-2 S-N S-1-2))))

 (local
  (in-theory (disable resolucion-cond-clausulas
                      resolucion-cond-clausula-conjunto)))

 (local
  (defthm INSAT-resolucion-es-forma-clausal-aux
    (implies (and (es-forma-clausal S-1)
                  (es-forma-clausal S-2)
                  (not (equal (resolucion-cond-saturacion S-1 S-2) t)))
             (es-forma-clausal (resolucion-cond-saturacion S-1 S-2)))))

 (defthm INSAT-resolucion-es-forma-clausal
   (implies (and (es-forma-clausal S)
                 (not (INSAT-resolucion S)))
            (es-forma-clausal (saturacion-resolucion S))))

 )

;;;============================================================================
;;;
;;; 2) El proceso devuelve un conjunto que no contiene la clusula vaca
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (defthm contiene-clausula-vacia-append
    (iff (contiene-clausula-vacia (append S-1 S-2))
         (or (contiene-clausula-vacia S-1)
             (contiene-clausula-vacia S-2)))))

 (local
  (defthm INSAT-resolucion-no-contiene-clausula-vacia-aux
    (implies (and (not (contiene-clausula-vacia S-1))
                  (not (contiene-clausula-vacia S-2))
                  (not (equal (resolucion-cond-saturacion S-1 S-2) t)))
             (not (contiene-clausula-vacia
                   (resolucion-cond-saturacion S-1 S-2))))))
 
 (defthm INSAT-resolucion-no-contiene-clausula-vacia
   (implies (and (not (contiene-clausula-vacia S))
                 (not (INSAT-resolucion S)))
            (not (contiene-clausula-vacia
                  (saturacion-resolucion S)))))

 )

;;;============================================================================
;;;
;;; 3) El proceso devuelve una forma clausal que contiene a la original
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (defthm INSAT-resolucion-contiene-acumulador
    (implies (not (equal (resolucion-cond-saturacion S-1 S-2) t))
             (subsetp-equal S-2 (resolucion-cond-saturacion S-1 S-2)))
    :rule-classes nil))

 (local
  (defthm INSAT-resolucion-contiene-acumulador-cons
    (implies 
     (not (equal (resolucion-cond-saturacion S-1 (cons C S-2)) t))
     (member-equal C (resolucion-cond-saturacion S-1 (cons C S-2))))
    :hints (("Goal"
             :use (:instance INSAT-resolucion-contiene-acumulador
                             (S-2 (cons C S-2)))))))

 (local
  (defthm subsetp-equal-append
    (iff (subsetp-equal (append S-1 S-2) S-3)
         (and (subsetp-equal S-1 S-3)
              (subsetp-equal S-2 S-3)))))

 (local
  (defthm INSAT-resolucion-contiene-original-aux
    (implies (not (equal (resolucion-cond-saturacion S-1 S-2) t))
             (subsetp-equal S-1 (resolucion-cond-saturacion S-1 S-2)))))
 
 (defthm INSAT-resolucion-contiene-original
   (implies (not (INSAT-resolucion S))
            (subsetp-equal S (saturacion-resolucion S)))
   :rule-classes nil)
                
 )

;;;============================================================================
;;;
;;; 4) El proceso devuelve un conjunto saturado por resolucin condicionada
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (in-theory (disable resolucion-cond-clausulas
                      resolucion-cond-clausula-conjunto
                      igual-conjunto)))

 (local
  (defthm conjunto-miembro-append
    (iff (conjunto-miembro C (append S1 S2))
         (or (conjunto-miembro C S1)
             (conjunto-miembro C S2)))))

 (local
  (defthm conjunto-miembro-cons
    (iff (conjunto-miembro C1 (cons C2 S))
         (or (igual-conjunto C1 C2)
             (conjunto-miembro C1 S)))))

 (local
  (defthm resolucion-cond-incosistencia-conjunto-miembro-2
    (implies (and (conjunto-miembro C S2)
                  (not (equal (resolucion-cond-saturacion S1 S2) t)))
             (conjunto-miembro C (resolucion-cond-saturacion S1 S2)))))

 (local
  (defthm resolucion-cond-incosistencia-conjunto-miembro-car
    (implies (and (consp S1)
                  (not (equal (resolucion-cond-saturacion S1 S2) t)))
             (conjunto-miembro (car S1)
                               (resolucion-cond-saturacion S1 S2)))))

 (local
  (defthm resolucion-cond-incosistencia-conjunto-miembro-1
    (implies (and (conjunto-miembro C S1)
                  (not (equal (resolucion-cond-saturacion S1 S2) t)))
             (conjunto-miembro C (resolucion-cond-saturacion S1 S2)))))

 (local
  (defthm resolucion-cond-clausulas-igual-conjunto
    (implies (and (igual-conjunto C1 C2)
                  (es-clausula C1))
             (not (consp (resolucion-cond-clausulas C1 C2))))))

 (local
  (defthm conjunto-miembro-resolucion-cond-saturacion-not-t
    (implies (conjunto-miembro C (resolucion-cond-saturacion S1 S2))
             (not (equal (resolucion-cond-saturacion S1 S2) t)))
    :rule-classes :forward-chaining))

 (local
  (defthm resolucion-cond-clausula-conjunto-conjunto-miembro-acumulador
    (implies (conjunto-miembro C-1 S-N)
             (conjunto-miembro 
              C-1 (resolucion-cond-clausula-conjunto C-2 S-2 S-N S-1-2)))
    :hints (("Goal" :in-theory (enable resolucion-cond-clausula-conjunto)))))

 (local
  (defthm resolucion-cond-clausula-conjunto-conjunto-miembro-acumulador-inst
    (implies (and (igual-conjunto C-2 C-3)
                  (es-clausula C-2)
                  (es-clausula C-1))
             (conjunto-miembro
              (car (resolucion-cond-clausulas C-1 C-2))
              (resolucion-cond-clausula-conjunto
               C-1 S-2 (cons (car (resolucion-cond-clausulas C-1 C-3)) S-N)
               S-1-2)))
    :hints (("Goal" :use
             (:instance 
              resolucion-cond-clausula-conjunto-conjunto-miembro-acumulador
              (C-1 (car (resolucion-cond-clausulas C-1 C-2)))
              (C-2 C-1)
              (S-N (cons (car (resolucion-cond-clausulas C-1 C-3)) S-N)))))))

 (local
  (defthm resolucion-cond-clausula-conjunto-conjunto-miembro-3
    (implies 
     (and (or (and (igual-conjunto C2 C) (conjunto-miembro C1 S-2))
              (and (igual-conjunto C1 C) (conjunto-miembro C2 S-2)))
          (es-clausula C1)
          (es-clausula C2)
          (consp (resolucion-cond-clausulas C1 C2))
          (not (conjunto-miembro (car (resolucion-cond-clausulas C1 C2)) 
                                 S-1-2)))
     (conjunto-miembro (car (resolucion-cond-clausulas C1 C2))
                       (resolucion-cond-clausula-conjunto C S-2 S-N S-1-2)))
    :hints (("Goal" :in-theory (enable resolucion-cond-clausula-conjunto)))
    :rule-classes nil))

 (local
  (defthm resolucion-cond-saturacion-satura-aux
    (implies (and (es-clausula C1)
                  (es-clausula C2)
                  (conjunto-miembro C1 (resolucion-cond-saturacion S-1 S-2))
                  (conjunto-miembro C2 (resolucion-cond-saturacion S-1 S-2))
                  (implies (and (conjunto-miembro C1 S-2)
                                (conjunto-miembro C2 S-2)
                                (consp (resolucion-cond-clausulas C1 C2)))
                           (conjunto-miembro
                            (car (resolucion-cond-clausulas C1 C2)) S-1))
                  (consp (resolucion-cond-clausulas C1 C2)))
             (conjunto-miembro (car (resolucion-cond-clausulas C1 C2))
                               (resolucion-cond-saturacion S-1 S-2)))
    :hints (("Subgoal *1/3"
             :use (:instance 
                   resolucion-cond-clausula-conjunto-conjunto-miembro-3
                   (C (car S-1))
                   (S-N nil)
                   (S-1-2 (cons (car S-1) (append (cdr S-1) S-2))))))))

 (local
  (defthm saturacion-resolucion-funciona
    (implies (and (es-clausula C1)
                  (es-clausula C2)
                  (conjunto-miembro C1 (saturacion-resolucion S))
                  (conjunto-miembro C2 (saturacion-resolucion S))
                  (consp (resolucion-cond-clausulas C1 C2)))
             (conjunto-miembro (car (resolucion-cond-clausulas C1 C2))
                               (saturacion-resolucion S)))))

 (local
  (defthm resolucion-cond-saturacion-clausulas
    (implies (and (conjunto-miembro C (saturacion-resolucion S))
                  (es-forma-clausal S))
             (es-clausula C))
    :otf-flg t
    :hints (("Goal"
             :in-theory (disable INSAT-resolucion-es-forma-clausal)
             :use INSAT-resolucion-es-forma-clausal))))

 (defthm INSAT-resolucion-saturado-cond-conjunto
   (implies (and (es-forma-clausal S)
                 (not (INSAT-resolucion S)))
            (saturado-cond-conjunto (saturacion-resolucion S)))
   :hints (("Goal"
	    :in-theory (disable saturacion-resolucion)
            :use (:FUNCTIONAL-INSTANCE
                  SATURADO-COND-CONJUNTO-EXTENSION
                  (SATURADO-COND-H
                   (LAMBDA NIL
                           (AND (ES-FORMA-CLAUSAL S)
                                (NOT (INSAT-RESOLUCION S)))))
                  (SATURADO-COND-S 
                   (LAMBDA NIL
                           (SATURACION-RESOLUCION S)))))))

 )

;;;============================================================================
;;;
;;; 5) El proceso es correcto
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (defthm correccion-resolvente
    (implies (and (es-clausula C-1)
                  (es-clausula C-2)
                  (es-literal L)
                  (modelo-clausula sigma C-1)
                  (modelo-clausula sigma C-2))
             (modelo-clausula sigma (resolvente L C-1 C-2)))))

 (local
  (in-theory (disable resolvente)))

 (local
  (in-theory (enable resolucion-cond-clausulas
                     resolucion-cond-clausula-conjunto)))
 (local
  (defthm correccion-resolucion-cond-clausulas
    (implies (and (es-clausula C-1)
                  (es-clausula C-2)
                  (modelo-clausula sigma C-1)
                  (modelo-clausula sigma C-2)
                  (resolucion-cond-clausulas C-1 C-2))
             (modelo-clausula
              sigma (car (resolucion-cond-clausulas C-1 C-2))))))

 (local
  (defthm correccion-resolucion-cond-clausula-conjunto
    (implies (and (es-clausula C)
                  (es-forma-clausal S-2)
                  (es-forma-clausal S-N)
                  (es-forma-clausal S-1-2)
                  (modelo-clausula sigma C)
                  (modelo-forma-clausal sigma S-2)
                  (modelo-forma-clausal sigma S-N))
             (modelo-forma-clausal 
              sigma (resolucion-cond-clausula-conjunto C S-2 S-N S-1-2)))))

 (local
  (in-theory (disable resolucion-cond-clausulas
                      resolucion-cond-clausula-conjunto
                      contiene-clausula-vacia-es-inconsistente)))

 (local
  (defthm correccion-INSAT-resolucion-aux
    (implies (and (es-forma-clausal S-1)
                  (es-forma-clausal S-2)
                  (not (equal S-2 t))
                  (modelo-forma-clausal sigma S-2)
                  (modelo-forma-clausal sigma S-1))
             (not (equal (resolucion-cond-saturacion S-1 S-2) t)))
    :rule-classes nil
    :hints (("Subgoal *1/2"
             :use (:instance
                   contiene-clausula-vacia-es-inconsistente
                   (Fc (resolucion-cond-clausula-conjunto 
                        (car S-1) S-2
                        nil (append S-1 S-2))))))))
 
 (defthm correccion-INSAT-resolucion
   (implies (and (es-forma-clausal S)
                 (INSAT-resolucion S))
            (not (modelo-forma-clausal sigma S)))
   :hints (("Goal"
            :use (:instance correccion-INSAT-resolucion-aux
                            (S-1 S)
                            (S-2 nil)))))
 )

;;;============================================================================
;;;
;;; 6) El proceso es completo
;;;
;;;============================================================================

(include-book "../9-resolucion/resolucion-thm")

(encapsulate
 ()

 (local
  (defthm coleccion-clausulas-falsas-no-vacias
    (implies (and (es-forma-clausal S)
                  (not (contiene-clausula-vacia S)))
             (coleccion-de-conjuntos-no-vacios
              (coleccion-clausulas-falsas-asignacion sigma S)))
    :hints (("Goal" :in-theory (enable conjunto-vacio)))))

 (local
  (defthm completitud-INSAT-resolucion-aux
    (implies (and (es-forma-clausal S)
                  (not (contiene-clausula-vacia S))
                  (not (equal (saturacion-resolucion S) t)))
             (modelo-forma-clausal
              (nueva-asignacion
               (construye-conjunto-representante-minimal
                (coleccion-clausulas-falsas-asignacion
                 (asignacion) (saturacion-resolucion S)))
               (asignacion))
              (saturacion-resolucion S)))
    :hints (("Goal" :in-theory (disable saturacion-resolucion)))
    :rule-classes nil))

 (local
  (defthm modelo-forma-clausal-subsetp-equal
    (implies (and (subsetp-equal S-1 S-2)
                  (modelo-forma-clausal sigma S-2))
             (modelo-forma-clausal sigma S-1))))

 (local
  (defthm modelo-saturacion-resolucion
    (implies (and (es-forma-clausal S)
                  (not (contiene-clausula-vacia S))
                  (not (equal (saturacion-resolucion S) t))
                  (modelo-forma-clausal
                   sigma (saturacion-resolucion S)))
             (modelo-forma-clausal sigma S))
    :hints (("Goal" :use INSAT-resolucion-contiene-original))))

 (defthm completitud-INSAT-resolucion
   (implies (and (es-forma-clausal S)
                 (not (contiene-clausula-vacia S))
                 (not (INSAT-resolucion S)))
            (modelo-forma-clausal
             (nueva-asignacion
              (construye-conjunto-representante-minimal
               (coleccion-clausulas-falsas-asignacion
                (asignacion) (saturacion-resolucion S)))
              (asignacion))
             S))
   :hints (("Goal" 
	    :in-theory (disable saturacion-resolucion)
	    :use completitud-INSAT-resolucion-aux)))

 )

;;;============================================================================