;;;============================================================================
;;;
;;; Ttulo: Resolucin: Teora genrica e Instancias
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(defpkg "MUL" (union-eq *acl2-exports*
			*common-lisp-symbols-from-main-lisp-package*))

(certify-book "../9-resolucion/resolucion-gen" 1)

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../9-resolucion/resolucion-aux")

;;;============================================================================
;;;
;;; 1) Teora genrica
;;;
;;;============================================================================

(defthm true-list-car-resolucion-cond-clausulas
  (implies (and (true-listp C1)
		(true-listp C2)
		(consp (resolucion-cond-clausulas C1 C2)))
	   (true-listp (car (resolucion-cond-clausulas C1 C2))))
  :hints (("Goal" :in-theory (enable resolucion-cond-clausulas))))

(defthm resolucion-cond-clausulas-built-in-clause
  (EQUAL (RESOLUCION-COND-CLAUSULAS C1 C2)
	 (IF (CONDICION C1 C2)
	     (LET ((L (LITERAL-RESOLUCION-CLAUSULAS C1 C2)))
		  (IF L
		      (LET ((RES (RESOLVENTE L C1 C2)))
			   (IF (TIENE-LITERALES-COMPLEMENTARIOS RES)
			       NIL (LIST RES)))
		      NIL))
	     NIL))
  :rule-classes :built-in-clause
  :hints (("Goal" :in-theory (enable resolucion-cond-clausulas))))

(defthm resolucion-cond-clausula-conjunto-built-in-clause
  (EQUAL (RESOLUCION-COND-CLAUSULA-CONJUNTO C S-2 S-N S-1-2)
	 (IF
	  (CONSP S-2)
	  (LET
	   ((RES (RESOLUCION-COND-CLAUSULAS C (CAR S-2))))
	   (IF
	    (CONSP RES)
	    (COND ((CONJUNTO-MIEMBRO (CAR RES) S-1-2)
		   (RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2)
						      S-N S-1-2))
		  ((CONJUNTO-MIEMBRO (CAR RES) S-N)
		   (RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2)
						      S-N S-1-2))
		  (T (RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2)
							(CONS (CAR RES) S-N)
							S-1-2)))
	    (RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2)
					       S-N S-1-2)))
	  S-N))
  :INSTRUCTIONS ((:DIVE 1) :X :TOP :BASH)
  :RULE-CLASSES :BUILT-IN-CLAUSE)

(include-book "../6-extensiones/generic-theory")

;;;   La constante que define la teora genrica es la siguiente:

(defconst *resolucion*
  '((DEFUN RESOLUCION-COND-CLAUSULAS (C1 C2)
      (DECLARE (XARGS :GUARD (AND (TRUE-LISTP C1) (TRUE-LISTP C2))))
      (IF (CONDICION C1 C2)
	  (LET ((L (LITERAL-RESOLUCION-CLAUSULAS C1 C2)))
	       (IF L
		   (LET ((RES (RESOLVENTE L C1 C2)))
			(IF (TIENE-LITERALES-COMPLEMENTARIOS RES)
			    NIL (LIST RES)))
		   NIL))
	  NIL))

    (defthm resolucion-cond-clausulas-built-in-clause
      (EQUAL 
       (RESOLUCION-COND-CLAUSULAS C1 C2)
       (AND (CONDICION C1 C2)
	    (LET ((L (LITERAL-RESOLUCION-CLAUSULAS C1 C2)))
		 (AND L
		      (LET ((RES (RESOLVENTE L C1 C2)))
			   (AND (NOT (TIENE-LITERALES-COMPLEMENTARIOS RES))
				(LIST RES)))))))
      :rule-classes :built-in-clause)

    (in-theory (disable resolucion-cond-clausulas))
    
    (defthm true-list-car-resolucion-cond-clausulas
      (implies (and (true-listp C1)
		    (true-listp C2)
		    (consp (resolucion-cond-clausulas C1 C2)))
	       (true-listp (car (resolucion-cond-clausulas C1 C2)))))

    (DEFUN RESOLUCION-COND-CLAUSULA-CONJUNTO (C S-2 S-N S-1-2)
        (DECLARE (XARGS :GUARD (AND (TRUE-LISTP C)
				    (TRUE-LIST-LISTP S-2)
				    (TRUE-LIST-LISTP S-N)
				    (TRUE-LIST-LISTP S-1-2))))
        (IF (ENDP S-2)
            S-N
            (LET ((RES (RESOLUCION-COND-CLAUSULAS C (CAR S-2))))
                 (IF (OR (NOT (CONSP RES))
                         (CONJUNTO-MIEMBRO (CAR RES) S-1-2)
                         (CONJUNTO-MIEMBRO (CAR RES) S-N))
                     (RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2) S-N S-1-2)
                     (RESOLUCION-COND-CLAUSULA-CONJUNTO
		      C (CDR S-2) (CONS (CAR RES) S-N) S-1-2)))))

    (defthm resolucion-cond-clausula-conjunto-built-in-clause
      (EQUAL (RESOLUCION-COND-CLAUSULA-CONJUNTO C S-2 S-N S-1-2)
	     (IF
	      (CONSP S-2)
	      (LET
	       ((RES (RESOLUCION-COND-CLAUSULAS C (CAR S-2))))
	       (IF
		(CONSP RES)
		(COND ((CONJUNTO-MIEMBRO (CAR RES) S-1-2)
		       (RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2)
							  S-N S-1-2))
		      ((CONJUNTO-MIEMBRO (CAR RES) S-N)
		       (RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2)
							  S-N S-1-2))
		      (T (RESOLUCION-COND-CLAUSULA-CONJUNTO
			  C (CDR S-2) (CONS (CAR RES) S-N) S-1-2)))
		(RESOLUCION-COND-CLAUSULA-CONJUNTO C (CDR S-2)
						   S-N S-1-2)))
	      S-N))
      :RULE-CLASSES :BUILT-IN-CLAUSE)

    (in-theory (disable RESOLUCION-COND-CLAUSULA-CONJUNTO))

    (DEFTHM TRUE-LIST-LISTP-RESOLUCION-COND-CLAUSULA-CONJUNTO
      (IMPLIES (TRUE-LIST-LISTP S-N)
	       (TRUE-LIST-LISTP (RESOLUCION-COND-CLAUSULA-CONJUNTO 
				 C S-2 S-N S-1-2))))

    (DEFTHM TRUE-LISTP-RESOLUCION-COND-CLAUSULA-CONJUNTO
      (IMPLIES (TRUE-LIST-LISTP S-N)
               (TRUE-LISTP (RESOLUCION-COND-CLAUSULA-CONJUNTO 
			    C S-2 S-N S-1-2))))

    (DEFTHM MEDIDA-RESOLUCION-COTA-INVARIANTE
      (EQUAL (MEDIDA-RESOLUCION-COTA
	      (CONS C (APPEND (APPEND (RESOLUCION-COND-CLAUSULA-CONJUNTO
				       C S-2
				       NIL (CONS C (APPEND S-1 S-2)))
				      S-1)
			      S-2)))
	     (MEDIDA-RESOLUCION-COTA (CONS C (APPEND S-1 S-2)))))

    (DEFTHM MEDIDA-RESOLUCION-AUMENTA
      (IMPLIES
       (AND (NOT (CONTIENE-CLAUSULA-VACIA
		  (RESOLUCION-COND-CLAUSULA-CONJUNTO
		   C S-2 NIL (CONS C (APPEND S-1 S-2)))))
	    (CONSP (RESOLUCION-COND-CLAUSULA-CONJUNTO
		    C S-2 NIL (CONS C (APPEND S-1 S-2)))))
       (< (MEDIDA-RESOLUCION (CONS C (APPEND S-1 S-2)))
	  (MEDIDA-RESOLUCION
	   (CONS C (APPEND (APPEND (RESOLUCION-COND-CLAUSULA-CONJUNTO
				    C S-2 NIL (CONS C (APPEND S-1 S-2)))
				   S-1)
			   S-2)))))
      :RULE-CLASSES :LINEAR)

    (DEFUN RESOLUCION-COND-SATURACION (S-1 S-2)
      (DECLARE 
       (XARGS :GUARD (AND (TRUE-LIST-LISTP S-1)
			  (TRUE-LIST-LISTP S-2))
 	      :MEASURE (medida-saturacion S-1 S-2)
	      :hints (("Goal" 
		       :in-theory
		       (disable resolucion-cond-clausula-conjunto
				medida-resolucion-cota
				medida-resolucion
				e0-ordinalp)
		       :expand
		       (medida-saturacion S-1 S-2))
		      ("Subgoal 1"
		       :cases ((consp (resolucion-cond-clausula-conjunto 
				       (car S-1) S-2
				       nil (append S-1 S-2))))))))
      (IF (ENDP S-1)
 	  S-2
 	  (LET
 	   ((RESOLVENTES (RESOLUCION-COND-CLAUSULA-CONJUNTO 
			  (CAR S-1) S-2 NIL (APPEND S-1 S-2))))
 	   (IF (CONTIENE-CLAUSULA-VACIA RESOLVENTES)
 	       T
 	       (RESOLUCION-COND-SATURACION 
 		(APPEND RESOLVENTES (CDR S-1)) (CONS (CAR S-1) S-2))))))

    (DEFUN SATURACION-RESOLUCION (S)
      (DECLARE (XARGS :GUARD (TRUE-LIST-LISTP S)))
      (RESOLUCION-COND-SATURACION S NIL))

    (DEFUN INSAT-RESOLUCION (S)
      (DECLARE (XARGS :GUARD (TRUE-LIST-LISTP S)))
      (EQUAL (SATURACION-RESOLUCION S) T))

    (DEFTHM CORRECCION-INSAT-RESOLUCION
      (IMPLIES (AND (ES-FORMA-CLAUSAL S)
		    (INSAT-RESOLUCION S))
	       (NOT (MODELO-FORMA-CLAUSAL SIGMA S))))

    (DEFTHM COMPLETITUD-INSAT-RESOLUCION
      (IMPLIES (AND (ES-FORMA-CLAUSAL S)
		    (NOT (CONTIENE-CLAUSULA-VACIA S))
		    (NOT (INSAT-RESOLUCION S)))
	       (MODELO-FORMA-CLAUSAL
		(NUEVA-ASIGNACION (CONSTRUYE-CONJUNTO-REPRESENTANTE-MINIMAL
				   (COLECCION-CLAUSULAS-FALSAS-ASIGNACION
                                    (ASIGNACION)
                                    (SATURACION-RESOLUCION S)))
				  (ASIGNACION))
		S)))
  
  ))

(make-generic-theory *resolucion*)

;;;============================================================================
;;;
;;; 2) Instancias
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Resolucin binaria
;;;
;;;----------------------------------------------------------------------------

;;;   La condicin asociada a la resolucin binaria es siempre cierta:

(defun cond-binaria (C1 C2)
  (declare (ignore C1 C2))
  t)

(defcong igual-conjunto iff (cond-binaria c1 c2) 1)

(defcong igual-conjunto iff (cond-binaria c1 c2) 2)

(defthm cond-binaria-conmuta
  (iff (cond-binaria c1 c2)
       (cond-binaria c2 c1)))

(defthm cond-binaria-boolean-p
  (booleanp (cond-binaria c1 c2)))

(defun asig-binaria ()
  nil)

(defthm propiedad-fundamental-cond-binaria
  (implies (and (es-clausula D)
		(es-clausula C)
		(not (modelo-clausula (asig-binaria) D)))
	   (cond-binaria C D)))

(in-theory (disable igual-conjunto
		    cond-binaria
		    (:type-prescription cond-binaria)
		    asig-binaria
		    (asig-binaria)
		    (:type-prescription asig-binaria)
		    resolvente
		    resolucion-cond-clausulas))

(definstance-*resolucion*
  ((condicion cond-binaria)
   (asignacion asig-binaria))
  "-binaria")

;;;   Veamos algunos ejemplos de uso:

#|

(include-book "../9-resolucion/resolucion-gen")

(include-book "../0-ejemplos/plaisted")

(include-book "../8-davisputnam/forma-clausal")

(defun INSAT-resolucion-binaria-formula (F)
  (INSAT-resolucion-binaria (forma-clausal F)))

(value :q)

(compile 'INSAT-resolucion-binaria-formula)

(load "../0-ejemplos/ejemplos.lisp")

(evalua-ejemplos 'INSAT-resolucion-binaria *plaisted*)

;;; +=============+============+
;;; | Frmula     | Tiempo     |
;;; +-------------+------------+
;;; | Plaisted 5  |      8.720 |
;;; | Plaisted 6  |    250.380 |
;;; +=============+============+

|#	    

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) Resolucin positiva
;;;
;;;----------------------------------------------------------------------------

;;;   Una clusula es positiva si todos sus elementos son literales positivos:

(defun es-clausula-positiva (C)
  (declare (xargs :guard (true-listp C)))
  (if (endp C)
      t
      (and (es-literal-positivo (car C))
	   (es-clausula-positiva (cdr C)))))

(in-theory (enable igual-conjunto))

(defthm es-clausula-positiva-subsetp-equal
  (implies (and (subsetp-equal c1 c2)
		(es-clausula-positiva c2))
	   (es-clausula-positiva c1)))

(defcong igual-conjunto iff (es-clausula-positiva c) 1)

;;;   La condicin asociada a la resolucin positiva es que al menos alguna de
;;; las clusulas que en ella intervienen sea positiva:

(defun cond-positiva (C1 C2)
  (declare (xargs :guard (and (true-listp C1)
			      (true-listp C2))))

  (or (es-clausula-positiva C1)
      (es-clausula-positiva C2)))

(defcong igual-conjunto iff (cond-positiva c1 c2) 1)

(defcong igual-conjunto iff (cond-positiva c1 c2) 2)

(defthm cond-positiva-conmuta
  (iff (cond-positiva c1 c2)
       (cond-positiva c2 c1)))

(defthm cond-positiva-boolean-p
  (booleanp (cond-positiva c1 c2)))

(defun asig-positiva ()
  nil)

(defthm modelo-asig-positiva-literal-no-positivo
  (implies (and (not (es-literal-positivo L))
		(es-literal L))
	   (modelo (asig-positiva) L))
  :hints (("Goal" :in-theory (enable modelo))))

(in-theory (disable asig-positiva
		    (asig-positiva)
		    (:type-prescription asig-positiva)))

(defthm no-modelo-asig-positiva-clausula-positiva
  (implies (and (es-clausula C)
		(not (modelo-clausula (asig-positiva) C)))
	   (es-clausula-positiva C)))

(defthm propiedad-fundamental-cond-positiva
  (implies (and (es-clausula C)
		(es-clausula D)
		(not (modelo-clausula (asig-positiva) D)))
	   (cond-positiva C D))
  :hints (("Goal" :in-theory (disable igual-conjunto))))

(in-theory (disable igual-conjunto
		    cond-positiva
		    (:type-prescription cond-positiva)
		    asig-positiva
		    (asig-positiva)
		    (:type-prescription asig-positiva)
		    resolvente
		    resolucion-cond-clausulas))

(definstance-*resolucion*
  ((condicion cond-positiva)
   (asignacion asig-positiva))
  "-positiva")

;;;   Veamos algunos ejemplos de uso:

#|

(include-book "../9-resolucion/resolucion-gen")

(include-book "../0-ejemplos/plaisted")

(include-book "../8-davisputnam/forma-clausal")

(defun INSAT-resolucion-positiva-formula (F)
  (INSAT-resolucion-positiva (forma-clausal F)))

(value :q)

(compile 'INSAT-resolucion-positiva-formula)

(load "../0-ejemplos/ejemplos.lisp")

(evalua-ejemplos 'INSAT-resolucion-positiva *plaisted* t)

;;; +=============+=====+============+
;;; | Frmula     | Sol | Tiempo     |
;;; +-------------+-----+------------+
;;; | Plaisted 5  | T   |      0.000 |
;;; | Plaisted 6  | T   |      0.000 |
;;; | Plaisted 7  | T   |      0.010 |
;;; | Plaisted 8  | T   |      0.000 |
;;; | Plaisted 9  | T   |      0.010 |
;;; | Plaisted 10 | T   |      0.000 |
;;; +=============+=====+============+

|#	    

;;;----------------------------------------------------------------------------
;;;
;;; 2.3) Resolucin negativa
;;;
;;;----------------------------------------------------------------------------

;;;   Una clusula es negativa si todos sus elementos son literales negativos:

(defun es-clausula-negativa (C)
  (declare (xargs :guard (true-listp C)))
  (if (endp C)
      t
      (and (es-literal-negativo (car C))
	   (es-clausula-negativa (cdr C)))))

(in-theory (enable igual-conjunto))

(defthm es-clausula-negativa-subsetp-equal
  (implies (and (subsetp-equal c1 c2)
		(es-clausula-negativa c2))
	   (es-clausula-negativa c1))
  :hints (("Goal" :in-theory (disable es-literal-negativo))))

(defcong igual-conjunto iff (es-clausula-negativa c) 1)

;;;   La condicin asociada a la resolucin negativa es que al menos alguna de
;;; las clusulas que en ella intervienen sea negativa:

(defun cond-negativa (C1 C2)
  (declare (xargs :guard (and (true-listp C1)
			      (true-listp C2))))

  (or (es-clausula-negativa C1)
      (es-clausula-negativa C2)))

(defcong igual-conjunto iff (cond-negativa c1 c2) 1)

(defcong igual-conjunto iff (cond-negativa c1 c2) 2)

(defthm cond-negativa-conmuta
  (iff (cond-negativa c1 c2)
       (cond-negativa c2 c1)))

(defthm cond-negativa-boolean-p
  (booleanp (cond-negativa c1 c2)))

;;; !*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!
                                          ;*!
;;; Asignacin identicamente positiva !!! ;*!
                                          ;*!
(defstub asig-negativa () => *)           ;*!
                                          ;*!
(defaxiom asig-negativa-es-positiva       ;*!
  (equal (valor F (asig-negativa)) *V*))  ;*!
                                          ;*!
;;; !*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!*!

(defthm modelo-asig-negativa-literal-no-negativo
  (implies (and (not (es-literal-negativo L))
		(es-literal L))
	   (modelo (asig-negativa) L))
  :hints (("Goal" :in-theory (enable modelo))))

(defthm no-modelo-asig-negativa-clausula-negativa
  (implies (and (es-clausula C)
		(not (modelo-clausula (asig-negativa) C)))
	   (es-clausula-negativa C)))

(defthm propiedad-fundamental-cond-negativa
  (implies (and (es-clausula D)
		(es-clausula C)
		(not (modelo-clausula (asig-negativa) D)))
	   (cond-negativa C D))
  :hints (("Goal" :in-theory (disable igual-conjunto))))

(in-theory (disable igual-conjunto
		    cond-negativa
		    (:type-prescription cond-negativa)
		    resolvente
		    resolucion-cond-clausulas))

(definstance-*resolucion*
  ((condicion cond-negativa)
   (asignacion asig-negativa))
  "-negativa")

;;;   Veamos algunos ejemplos de uso:

#|

(include-book "../9-resolucion/resolucion-gen")

(include-book "../0-ejemplos/plaisted")

(include-book "../8-davisputnam/forma-clausal")

(defun INSAT-resolucion-negativa-formula (F)
  (INSAT-resolucion-negativa (forma-clausal F)))

(value :q)

(compile 'INSAT-resolucion-negativa-formula)

(load "../0-ejemplos/ejemplos.lisp")

(evalua-ejemplos 'INSAT-resolucion-negativa *plaisted* t)

;;; +=============+=====+============+
;;; | Frmula     | Sol | Tiempo     |
;;; +-------------+-----+------------+
;;; | Plaisted 5  | T   |      0.040 |
;;; | Plaisted 6  | T   |      0.220 |
;;; | Plaisted 7  | T   |      0.420 |
;;; | Plaisted 8  | T   |      0.810 |
;;; | Plaisted 9  | T   |      2.770 |
;;; | Plaisted 10 | T   |      3.810 |
;;; +=============+=====+============+

;;; +==========+============+
;;; | Frmula  | Tiempo     |
;;; +----------+------------+
;;; | Reinas-2 |      0.000 |
;;; | Reinas-3 |   1655.490 |
;;; +==========+============+

|#	    

;;;----------------------------------------------------------------------------
;;;
;;; 2.4) Resolucin semntica
;;;
;;;----------------------------------------------------------------------------

;;;   Primero consideramos una asignacin cualquiera:

(encapsulate
 (((asig-semantica) => *))

 (local
  (defun asig-semantica ()
    nil))

 (defthm alistp-asig-semantica
   (alistp (asig-semantica)))

 )

;;;   La condicin asociada a la resolucin semntica es que una clusula sea
;;; cierta en una asignacin predeterminada y la otra falsa:

(defun cond-semantica (C1 C2)
  (declare (xargs :guard (and (true-listp C1)
			      (true-listp C2))))

  (or (and (modelo-clausula (asig-semantica) C1)
	   (not (modelo-clausula (asig-semantica) C2)))
      (and (modelo-clausula (asig-semantica) C2)
	   (not (modelo-clausula (asig-semantica) C1)))))

(defcong igual-conjunto iff (cond-semantica c1 c2) 1)

(defcong igual-conjunto iff (cond-semantica c1 c2) 2)

(defthm cond-semantica-conmuta
  (iff (cond-semantica c1 c2)
       (cond-semantica c2 c1)))

(defthm cond-semantica-boolean-p
  (booleanp (cond-semantica c1 c2)))

(defthm propiedad-fundamental-cond-semantica
  (implies (and (es-clausula C)
		(es-clausula D)
		(modelo-clausula (asig-semantica) C)
		(not (modelo-clausula (asig-semantica) D)))
	   (cond-semantica C D))
  :hints (("Goal" :in-theory (disable igual-conjunto))))

(in-theory (disable igual-conjunto
		    cond-semantica
		    (:type-prescription cond-semantica)
		    resolvente
		    resolucion-cond-clausulas))

(definstance-*resolucion*
  ((condicion cond-semantica)
   (asignacion asig-semantica))
  "-semantica")

;;;   Veamos algunos ejemplos de uso:

#|

(include-book "resolucion-sat")

(value :q)

(time (resolucion-cond-saturacion-semantica
       '((p) ((- p) q) ((- q))) nil))
; => T

(time (resolucion-cond-saturacion-semantica
       '((p) ((- p) q)) nil))
; => NIL

(time (resolucion-cond-saturacion-semantica
       '((p q) ((- p) q) (p (- q)) ((- p) (- q))) nil))
; => T

(time (resolucion-cond-saturacion-semantica
       '((p q) (p r) ((- q) (- r)) ((- p))) nil))
; => T

(time (resolucion-cond-saturacion-semantica
       '((h) ((- h) p q) ((- p) c) ((- q) c) ((- c))) nil))
; => T

(time (resolucion-cond-saturacion-semantica
       '( ( 16 23 42 )             ( (- 16) 23 42 )     
          ( 26 41 (- 42) )         ( (- 26) 41 (- 42) )       ; 0.000
          ( 32 (- 41) (- 42) )     ( 6 15 (- 41) )            ; 0.030
          ( (- 6) 15 (- 32) )      ( 1 (- 32) 46 )            ; 1.250
          ( (- 1) (- 32) 46 )      ( (- 15) (- 41) (- 46) )   ; 10.500
;          ( (- 15) (- 21) (- 46) ) ( (- 23) 33 38 ) 
          ) nil))

|#	    

;;;----------------------------------------------------------------------------
;;;
;;; 2.5) Resolucin soporte
;;;
;;;----------------------------------------------------------------------------

;;;   Primero consideramos una forma clausal satisfacible cualquiera a la que
;;; llamaremos soporte:

(encapsulate
 (((S-soporte) => *)
  ((asig-soporte) => *))

 (local
  (defun S-soporte ()
    nil))

 (local
  (defun asig-soporte ()
    nil))

 (defthm alistp-asig-soporte
   (alistp (asig-soporte)))

 (defthm soporte-es-forma-clausal
   (es-forma-clausal (S-soporte)))

 (defthm true-list-listp-soporte
   (true-list-listp (S-soporte)))

 (defthm soporte-es-satisfacible
   (modelo-forma-clausal (asig-soporte) (S-soporte)))

 )

;;;   La condicin asociada a la resolucin soporte es que ambas clusulas no
;;; esten simultneamente en el soporte.

(defun cond-soporte (C1 C2)
  (declare (xargs :guard (and (true-listp C1)
			      (true-listp C2))))
  (or (not (conjunto-miembro C1 (S-soporte)))
      (not (conjunto-miembro C2 (S-soporte)))))

(defcong igual-conjunto iff (cond-soporte c1 c2) 1)

(defcong igual-conjunto iff (cond-soporte c1 c2) 2)

(defthm cond-soporte-conmuta
  (iff (cond-soporte c1 c2)
       (cond-soporte c2 c1)))

(defthm cond-soporte-boolean-p
  (booleanp (cond-soporte c1 c2)))

(defthm propiedad-fundamental-cond-soporte
  (implies (and (es-clausula C)
		(es-clausula D)
		(not (modelo-clausula (asig-soporte) D)))
	   (cond-soporte C D))
  :hints (("Goal" :in-theory (disable igual-conjunto))))

(in-theory (disable igual-conjunto
		    cond-soporte
		    (:type-prescription cond-soporte)
		    resolvente
		    resolucion-cond-clausulas))

(definstance-*resolucion*
  ((condicion cond-soporte)
   (asignacion asig-soporte))
  "-soporte")

;;;   Veamos algunos ejemplos de uso:

#|

(include-book "resolucion-sat")

(value :q)

(time (resolucion-cond-saturacion-soporte
       '((p) ((- p) q) ((- q))) nil))
; => T

(time (resolucion-cond-saturacion-soporte
       '((p) ((- p) q)) nil))
; => NIL

(time (resolucion-cond-saturacion-soporte
       '((p q) ((- p) q) (p (- q)) ((- p) (- q))) nil))
; => T

(time (resolucion-cond-saturacion-soporte
       '((p q) (p r) ((- q) (- r)) ((- p))) nil))
; => T

(time (resolucion-cond-saturacion-soporte
       '((h) ((- h) p q) ((- p) c) ((- q) c) ((- c))) nil))
; => T

(time (resolucion-cond-saturacion-soporte
       '( ( 16 23 42 )             ( (- 16) 23 42 )     
          ( 26 41 (- 42) )         ( (- 26) 41 (- 42) )       ; 0.000
          ( 32 (- 41) (- 42) )     ( 6 15 (- 41) )            ; 0.030
          ( (- 6) 15 (- 32) )      ( 1 (- 32) 46 )            ; 1.250
          ( (- 1) (- 32) 46 )      ( (- 15) (- 41) (- 46) )   ; 10.500
;          ( (- 15) (- 21) (- 46) ) ( (- 23) 33 38 ) 
          ) nil))

|#	    

;;;============================================================================
