;;;============================================================================
;;;
;;; Ttulo: Resolucin: Teorema de completitud de Bezem
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../9-resolucion/resolucion-thm")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 1) Introduccin
;;;
;;;============================================================================

(include-book "../9-resolucion/resolucion-sat")

(in-theory (disable resolucion-cond-clausulas))

;;;============================================================================
;;;
;;; 1) Modelo de un conjunto saturado por resolucin condicionada
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Clusulas falsas en una asignacin
;;;
;;;----------------------------------------------------------------------------

(defun coleccion-clausulas-falsas-asignacion (sigma S)
  (declare (xargs :guard (and (alistp sigma)
                              (true-list-listp S))))
  (cond ((endp S) S)
        ((not (modelo-clausula sigma (car S)))
         (cons (car S)
               (coleccion-clausulas-falsas-asignacion sigma (cdr S))))
        (t (coleccion-clausulas-falsas-asignacion sigma (cdr S)))))

(defthm coleccion-clausulas-falsas-asignacion-member-equal
  (implies (member-equal C (coleccion-clausulas-falsas-asignacion sigma S))
           (not (modelo-clausula sigma C))))

(defthm caracterizacion-coleccion-clausulas-falsas-asignacion
  (iff (member-equal C (coleccion-clausulas-falsas-asignacion sigma S))
       (and (member-equal C S)
            (not (modelo-clausula sigma C)))))

(defthm subsetp-equal-modelo-clausula
  (implies (and (subsetp-equal C1 C2)
                (modelo-clausula sigma C1))
           (modelo-clausula sigma C2)))

(defcong igual-conjunto iff (modelo-clausula sigma C) 2
  :hints (("Goal" :in-theory (enable igual-conjunto))))

(defthm coleccion-clausulas-falsas-asignacion-conjunto-miembro
  (implies (conjunto-miembro C (coleccion-clausulas-falsas-asignacion sigma S))
           (not (modelo-clausula sigma C)))
  :hints (("Goal" 
           :use (:instance 
                 coleccion-clausulas-falsas-asignacion-member-equal
                 (C (conjunto-miembro-testigo
                     C (coleccion-clausulas-falsas-asignacion sigma S)))))))

(defthm coleccion-clausulas-falsas-asignacion-subsetp-equal
  (subsetp-equal (coleccion-clausulas-falsas-asignacion sigma S) S))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Nueva asignacin
;;;
;;;----------------------------------------------------------------------------

(defun literales-positivos (R)
  (declare (xargs :guard (true-listp R)))
  (cond ((endp R) nil)
        ((es-simbolo-proposicional (car R))
         (cons (car R) (literales-positivos (cdr R))))
        (t (literales-positivos (cdr R)))))

(defthm literales-positivos-true-listp
  (true-listp (literales-positivos R)))

(defthm literales-positivos-member-equal
  (iff (member-equal var (literales-positivos R))
       (and (es-simbolo-proposicional var)
            (member-equal var R))))

(defun literales-representante-asignacion (R alist sigma)
  (declare (xargs :guard (and (true-listp R)
                              (alistp alist)
                              (alistp sigma))))
  (cond ((endp alist) nil)
        ((and (es-simbolo-proposicional (car (car alist)))
              (modelo sigma (car (car alist)))
              (not (member-equal (complementario (car (car alist))) R)))
         (cons (car (car alist)) 
               (literales-representante-asignacion R (cdr alist) sigma)))
        (t (literales-representante-asignacion R (cdr alist) sigma))))

(defthm literales-representante-asignacion-true-listp
  (true-listp (literales-representante-asignacion R alist sigma)))

(defun variable-en-dominio-asignacion (var sigma)
  (declare (xargs :guard (alistp sigma)))
  (cond ((endp sigma) nil)
        ((equal var (car (car sigma))) t)
        (t (variable-en-dominio-asignacion var (cdr sigma)))))

(defthm modelo-simbolo-es-variable-en-dominio
  (implies (and (modelo sigma var)
                (es-simbolo-proposicional var))
           (variable-en-dominio-asignacion var sigma))
  :hints (("Goal" :in-theory (enable modelo))))

(defthm literales-representante-asignacion-member-equal
  (iff (member-equal var (literales-representante-asignacion R alist sigma))
       (and (es-simbolo-proposicional var)
            (variable-en-dominio-asignacion var alist)
            (modelo sigma var)
            (not (member-equal (complementario var) R))))
  :hints (("Goal" :in-theory (disable modelo es-simbolo-proposicional)
           :induct (literales-representante-asignacion R alist sigma))))

(defun genera-asignacion-positiva (literales)
  (declare (xargs :guard (true-listp literales)))
  (cond ((endp literales) nil)
        ((es-simbolo-proposicional (car literales))
         (asume-valor (car literales) *V*
                      (genera-asignacion-positiva (cdr literales))))
        (t (genera-asignacion-positiva (cdr literales)))))

(defthm true-listp-append
  (implies (and (true-listp l1)
                (true-listp l2))
           (true-listp (append l1 l2))))

(defthm modelo-genera-asignacion-positiva
  (implies (es-simbolo-proposicional var)
           (iff (modelo (genera-asignacion-positiva literales) var)
                (member-equal var literales)))
  :hints (("Goal" :in-theory (enable modelo))))

(defun nueva-asignacion (R sigma)
  (declare (xargs :guard (and (true-listp R)
                              (alistp sigma))))
  (genera-asignacion-positiva 
   (append (literales-positivos R)
           (literales-representante-asignacion R sigma sigma))))

(defthm member-equal-append
  (iff (member-equal e (append l1 l2))
       (or (member-equal e l1)
           (member-equal e l2))))

(defthm nueva-asignacion-modelo
  (implies (es-simbolo-proposicional var)
           (iff (modelo (nueva-asignacion R sigma) var)
                (or (member-equal var R)
                    (and (not (member-equal (complementario var) R))
                         (modelo sigma var)))))
  :hints (("Goal" :in-theory (disable modelo es-simbolo-proposicional))))

(in-theory (disable nueva-asignacion))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Esquema de induccin
;;;
;;;----------------------------------------------------------------------------
        
(defun reduce-coleccion-literal (S L)
  (declare (xargs :guard (true-list-listp S)))
  (cond ((endp S) S)
        ((member-equal L (car S))
         (reduce-coleccion-literal (cdr S) L))
        (t (cons (car S)
                 (reduce-coleccion-literal (cdr S) L)))))

(defthm reduce-coleccion-literal-preserva-forma-clausal
  (implies (es-forma-clausal S)
           (es-forma-clausal (reduce-coleccion-literal S L))))

(defthm caracterizacion-pertenencia-reduce-coleccion-literal
  (iff (conjunto-miembro C (reduce-coleccion-literal S L))
       (and (conjunto-miembro C S)
            (not (member-equal L C))))
  :hints (("Goal" :induct (reduce-coleccion-literal S L))))

(defthm reduce-coleccion-literal-preserva-conjunto-saturado-aux
  (implies (and (es-forma-clausal S)
                (saturado-cond-conjunto S)
                (conjunto-miembro C1 (reduce-coleccion-literal S L))
                (conjunto-miembro C2 (reduce-coleccion-literal S L))
                (consp (resolucion-cond-clausulas C1 C2)))              
           (conjunto-miembro (car (resolucion-cond-clausulas C1 C2))
                             (reduce-coleccion-literal S L)))
  :hints (("Goal''" :in-theory (enable resolucion-cond-clausulas))))

(defthm reduce-coleccion-literal-preserva-conjunto-saturado
  (implies (and (es-forma-clausal S)
                (saturado-cond-conjunto S))
           (saturado-cond-conjunto (reduce-coleccion-literal S L)))
  :hints (("Goal"
           :use (:functional-instance
                 saturado-cond-conjunto-extension
                 (saturado-cond-S
                  (lambda () (reduce-coleccion-literal S L)))
                 (saturado-cond-H 
                  (lambda () (and (es-forma-clausal S)
                                  (saturado-cond-conjunto S)))))
           :in-theory (disable resolvente complementario))))

(defthm reduce-coleccion-literal-e0-ord-<
  (implies (and (member-equal C S)
                (member-equal L C))
           (e0-ord-< (acl2-count (reduce-coleccion-literal S L))
                     (acl2-count S)))
  :rule-classes nil)

(defthm reduce-coleccion-literal-preserva-true-list-listp
  (implies (true-list-listp S)
           (true-list-listp (reduce-coleccion-literal S L))))

(defthm reduce-coleccion-literal-preserva-clausula-nula
  (iff (contiene-clausula-vacia (reduce-coleccion-literal S L))
       (contiene-clausula-vacia S)))

(defthm reduce-coleccion-literal-coleccion-clausulas-falsas-asignacion
  (equal (coleccion-clausulas-falsas-asignacion 
          sigma (reduce-coleccion-literal S L))
         (reduce-coleccion-literal
          (coleccion-clausulas-falsas-asignacion sigma S) L)))

(include-book "../9-resolucion/representantes")

(defthm reduce-coleccion-literal-conjunto-representante
  (implies (conjunto-representante R S)
           (conjunto-representante R (reduce-coleccion-literal S L))))

(defun literal-no-existe-en-coleccion (L S)
  (declare (xargs :guard (true-list-listp S)))
  (cond ((endp S) t)
        ((member-equal L (car S)) nil)
        (t (literal-no-existe-en-coleccion L (cdr S)))))

(defthm reduce-coleccion-literal-no-existe
  (implies (literal-no-existe-en-coleccion L S)
           (equal (reduce-coleccion-literal S L) S)))

(defthm literal-cierto-no-existe-en-coleccion-clausulas-falsas
  (implies (modelo sigma L)
           (literal-no-existe-en-coleccion
            L (coleccion-clausulas-falsas-asignacion sigma S))))

(defthm reduce-coleccion-literal-literal-cierto
  (implies (modelo sigma L)
           (equal (reduce-coleccion-literal 
                   (coleccion-clausulas-falsas-asignacion sigma S) L) 
                  (coleccion-clausulas-falsas-asignacion sigma S))))

(defun literal-cierto-clausula (sigma C)
  (declare (xargs :guard (and (alistp sigma)
                              (true-listp C))))
  (cond ((endp C) nil)
        ((modelo sigma (car C))
         (car C))
        (t (literal-cierto-clausula sigma (cdr C)))))

(defthm literal-cierto-clausula-pertenece-a-la-clausula
  (implies (literal-cierto-clausula sigma C)
           (member-equal (literal-cierto-clausula sigma C) C)))

(defthm literal-cierto-clausula-es-cierto
  (implies (literal-cierto-clausula sigma C)
           (modelo sigma (literal-cierto-clausula sigma C)))
  :hints (("Goal" :in-theory (disable modelo))))

(defthm literal-cierto-clausula-existe
  (implies (and (member-equal l c)
                (modelo sigma l))
           (literal-cierto-clausula sigma c)))

(defun clausula-con-literal-cierto (sigma S)
  (declare (xargs :guard (and (alistp sigma)
                              (true-list-listp S))))
  (cond ((endp S) nil)
        ((literal-cierto-clausula sigma (car S))
         (car S))
        (t (clausula-con-literal-cierto sigma (cdr S)))))

(defthm clausula-con-literal-cierto-tiene-literal-cierto
  (implies (clausula-con-literal-cierto sigma S)
           (literal-cierto-clausula
            sigma (clausula-con-literal-cierto sigma S))))

(defthm clausula-con-literal-cierto-pertenece-a-la-coleccion
  (implies (clausula-con-literal-cierto sigma S)
           (member-equal (clausula-con-literal-cierto sigma S) S)))

(defun induccion-literal-cierto (sigma S)
  (declare (xargs :hints (("Goal"
                           :use (:instance 
                                 reduce-coleccion-literal-e0-ord-<
                                 (C (clausula-con-literal-cierto sigma S))
                                 (L (literal-cierto-clausula 
                                     sigma 
                                     (clausula-con-literal-cierto sigma S))))))
                  :guard (and (alistp sigma)
                              (true-list-listp S))))
  (if (clausula-con-literal-cierto sigma S)
      (induccion-literal-cierto
       sigma
       (reduce-coleccion-literal 
        S (literal-cierto-clausula sigma 
                                   (clausula-con-literal-cierto sigma S))))
      t))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Caso base de la prueba de completitud
;;;
;;;----------------------------------------------------------------------------

(defthm coleccion-clasulas-positivas-sin-clausula-con-literal-negativo
  (implies (and (es-forma-clausal S)
                (not (clausula-con-literal-cierto sigma S)))
           (equal (coleccion-clausulas-falsas-asignacion sigma S) S)))

(defun elemento-en-interseccion (set1 set2)
  (declare (xargs :guard (and (true-listp set1)
                              (true-listp set2))))
  (cond ((endp set1) nil)
        ((member-equal (car set1) set2)
         (car set1))
        (t (elemento-en-interseccion (cdr set1) set2))))

(defthm interseccion-no-vacia-hay-elemento
  (implies (not (conjunto-vacio (interseccion set1 set2)))
           (and (member-equal (elemento-en-interseccion set1 set2) set1)
                (member-equal (elemento-en-interseccion set1 set2) set2)))
  :hints (("Goal"
           :in-theory (enable interseccion)
           :induct (elemento-en-interseccion set1 set2))))

(defthm modelo-literal-negativo
  (implies (and (es-literal-negativo L)
                (not (modelo sigma (arg1 L))))
           (modelo sigma L))
  :hints (("Goal" :in-theory (enable modelo))))

(defthm nueva-asignacion-literal-positivo
  (implies (and (member-equal L R)
                (member-equal L C)
                (es-literal-positivo L))
           (modelo (nueva-asignacion R sigma) L)))

(defthm not-clausula-con-literal-cierto-literal
  (implies (and (not (clausula-con-literal-cierto sigma S))
                (conjunto-miembro C S)
                (member-equal L C))
           (not (modelo sigma L)))
  :hints (("Goal" :in-theory (disable igual-conjunto))))

(defthm conjunto-representante-minimal-no-contiene-complementarios
  (implies (and (conjunto-representante-minimal R S)
                (not (clausula-con-literal-cierto sigma S))
                (member-equal L R)
                (es-literal-negativo L))
           (not (member-equal (arg1 L) R)))
  :hints (("goal" :in-theory (disable conjunto-representante-minimal
                                      es-literal-negativo)
           :use ((:instance not-clausula-con-literal-cierto-literal
                            (C (conjunto-testigo R S (arg1 L)))
                            (L (arg1 L)))
                 (:instance not-clausula-con-literal-cierto-literal
                            (C (conjunto-testigo R S L)))))))

(defthm complementario-literal-negativo
  (implies (es-literal-negativo L)
           (equal (complementario (arg1 L)) L))
  :hints (("Goal" :in-theory (enable complementario))))

(defthm nueva-asignacion-literal-negativo-aux
  (implies (and (member-equal L R)
                (es-literal-negativo L)
                (not (member-equal (arg1 L) R)))
           (not (modelo (nueva-asignacion R sigma) (arg1 L))))
  :hints (("Goal" :in-theory (disable es-simbolo-proposicional))))

(defthm nueva-asignacion-literal-negativo
  (implies (and (member-equal L R)
                (member-equal L C)
                (es-literal-negativo L)
                (conjunto-representante-minimal R S)
                (not (clausula-con-literal-cierto sigma S)))
           (modelo (nueva-asignacion R sigma) L))
  :hints (("Goal" 
           :in-theory (disable es-literal-negativo
                               conjunto-representante-minimal))))

(in-theory (enable es-literal))

(defthm nueva-asignacion-literal
  (implies (and (member-equal L R)
                (member-equal L C)
                (es-literal L)
                (conjunto-representante-minimal R S)
                (not (clausula-con-literal-cierto sigma S)))
           (modelo (nueva-asignacion R sigma) L))
  :hints (("Goal" :in-theory (disable conjunto-representante-minimal
                                      es-literal-positivo
                                      es-literal-negativo))))

(defthm elemento-en-interseccion-es-literal
  (implies (and (not (conjunto-vacio (interseccion R C)))
                (es-clausula C))
           (es-literal (elemento-en-interseccion R C)))
  :hints (("Goal" :use (:instance member-clausula-es-literal
                                  (L (elemento-en-interseccion R C))))))

(defthm nueva-asignacion-clausula
  (implies (and (es-forma-clausal S)
                (conjunto-representante-minimal R S)
                (not (clausula-con-literal-cierto sigma S))
                (not (contiene-clausula-vacia S))
                (member-equal C S))
           (modelo-clausula (nueva-asignacion R sigma) C))
   :hints (("Goal" :in-theory (disable es-literal igual-conjunto)
           :use ((:instance nueva-asignacion-literal
                            (L (elemento-en-interseccion R C)))
                 (:instance modelo-clausula-member
                            (L (elemento-en-interseccion R C))
                            (sigma (nueva-asignacion R sigma)))))))

(include-book "../6-extensiones/listas")

(in-theory (disable conjunto-representante-minimal))

(defthm nueva-asignacion-forma-clausal
  (implies (and (es-forma-clausal S)
                (conjunto-representante-minimal R S)
                (not (clausula-con-literal-cierto sigma S))
                (not (contiene-clausula-vacia S)))
           (modelo-forma-clausal (nueva-asignacion R sigma) S))
  :hints ((extension-conjuntiva-a-listas
           S
           (lambda (elt)
             (modelo-clausula (nueva-asignacion R sigma) elt))
           (lambda (lst)
             (modelo-forma-clausal (nueva-asignacion R sigma) lst)))))

(in-theory (enable conjunto-representante-minimal))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Caso inductivo de la prueba de completitud
;;;
;;;----------------------------------------------------------------------------

(defthm modelo-clausula-reduce-coleccion-literal
  (implies (and (member-equal C S)
                (not (member-equal L C))
                (modelo-forma-clausal sigma (reduce-coleccion-literal S L)))
           (modelo-clausula sigma C)))

(in-theory (enable complementario))

(defthm modelo-nueva-asignacion-literal-complementario-not-member
  (implies (and (es-literal L)
                (modelo sigma L)
                (not (member-equal (complementario L) R)))
           (modelo (nueva-asignacion R sigma) L))
  :rule-classes nil
  :hints (("Goal" 
           :in-theory (disable es-simbolo-proposicional)
           :use (:instance modelo-complementario
			   (F L)))))

(defthm modelo-nueva-asignacion-clausula-literal-complementario-not-member
  (implies (and (member-equal L C)
                (es-clausula C)
                (modelo sigma L)
                (not (member-equal (complementario L) R)))
           (modelo-clausula (nueva-asignacion R sigma) C))
  :hints (("Goal" 
           :use modelo-nueva-asignacion-literal-complementario-not-member)))

(in-theory (disable complementario))

(in-theory (disable es-simbolo-proposicional))

(defthm modelo-nueva-asignacion-literal-positivo-complementario-no-en-R
  (implies (and (es-literal-positivo L)
                (member-equal (complementario L) D)
                (not (member-equal (complementario L) R))
                (not (modelo (nueva-asignacion R sigma) L)))
           (not (modelo sigma L))))

(defthm complementario-literal-negativo-es-positivo
  (implies (es-literal-negativo L)
           (es-literal-positivo (complementario L)))
  :hints (("Goal" :in-theory (enable complementario))))

(defthm modelo-nueva-asignacion-literal-negativo-no-en-R
  (implies (and (es-literal-negativo L)
                (member-equal L D)
                (not (member-equal L R))
                (modelo (nueva-asignacion R sigma) L))
           (modelo sigma L))
  :hints (("Goal"
           :in-theory (disable es-literal-negativo es-literal-positivo)
           :use 
           (:instance
            modelo-nueva-asignacion-literal-positivo-complementario-no-en-R
            (L (complementario L))))))

(defthm modelo-nueva-asignacion-literal-no-en-R
  (implies (and (es-literal L)
                (member-equal L D)
                (not (member-equal L R))
                (modelo (nueva-asignacion R sigma) L))
           (modelo sigma L)))

(defun literal-testigo (sigma C)
  (declare (xargs :guard (and (alistp sigma)
                              (true-listp C))))
  (cond ((endp C) nil)
        ((modelo sigma (car C)) (car C))
        (t (literal-testigo sigma (cdr C)))))

(defthm literal-testigo-propiedades
  (implies (and (es-clausula C)
                (modelo-clausula sigma C))
           (and (es-literal (literal-testigo sigma C))
                (member-equal (literal-testigo sigma C) C)
                (modelo sigma (literal-testigo sigma C))))
  :hints (("Goal" :induct (literal-testigo sigma C))))

(defthm modelo-clausula-con-elemento-en-elimina-elemento
  (implies (and (member-equal L1 (elimina-elemento L2 D))
                (modelo sigma L1))
           (modelo-clausula sigma D))
  :rule-classes :forward-chaining)

(defthm modelo-clausula-con-elemento-en-elimina-elemento-instance
  (implies (modelo sigma (literal-testigo (nueva-asignacion R sigma)
                                          (elimina-elemento L D)))
           (modelo-clausula sigma D)))

(defthm literal-testigo-elimina-elemento-member
  (implies (modelo-clausula sigma (elimina-elemento l d))
           (member-equal (literal-testigo sigma (elimina-elemento l d)) d)))

(defthm conjunto-unitario-no-contiene-mas-de-uno
  (implies (and (member-equal L1 D)
                (member-equal L2 D)
                (not (equal L1 L2)))
           (not (conjunto-unitario D)))
  :rule-classes nil)

(defthm conjunto-unitario-interseccion-elimina-elemento
  (implies (and (member-equal L1 D)
                (member-equal L1 R)
                (member-equal L2 (elimina-elemento L1 D))
                (member-equal L2 R))
           (not (conjunto-unitario (interseccion R D))))
  :rule-classes nil
  :hints (("Goal"
           :in-theory (disable conjunto-unitario)
           :use (:instance conjunto-unitario-no-contiene-mas-de-uno
                           (D (interseccion R D))))))

(defthm modelo-clausula-nueva-asignacion-elimina-elemento
  (implies (and (es-literal L)
                (member-equal L D)
                (member-equal L R)
                (es-clausula D)
                (conjunto-unitario (interseccion R D))
                (not (modelo-clausula sigma D)))
           (not (modelo-clausula (nueva-asignacion R sigma)
                                 (elimina-elemento L D))))
  :hints (("Goal"
           :in-theory (disable es-literal conjunto-unitario)
           :use ((:instance conjunto-unitario-interseccion-elimina-elemento
                            (L1 L)
                            (L2 (literal-testigo (nueva-asignacion R sigma)
                                                 (elimina-elemento L D))))
                 (:instance modelo-nueva-asignacion-literal-no-en-R
                            (L (literal-testigo (nueva-asignacion R sigma)
                                                (elimina-elemento L D))))))))

(defthm modelo-clausula-union-equal
  (iff (modelo-clausula sigma (union-equal C1 C2))
       (or (modelo-clausula sigma C1)
           (modelo-clausula sigma C2))))

(defthm modelo-clausula-elimina-elemento
  (implies (modelo-clausula sigma (elimina-elemento L C))
           (modelo-clausula sigma C)))

(in-theory (enable resolvente resolucion-cond-clausulas))

(defthm literal-resolucion-clausulas-unico
  (implies (and (member-equal L C)
                (member-equal (complementario L) D)
                (es-clausula C)
                (resolucion-cond-clausulas C D))
           (equal (literal-resolucion-clausulas C D) L))
  :hints (("Goal"
           :in-theory (disable es-literal)
           :use (:instance tiene-literales-complementarios-member
                           (C (resolvente (literal-resolucion-clausulas C D)
                                          C D))))))

(defthm coleccion-clausulas-falsas-asignacion-modelo-clausula-instance
  (implies (and (member-equal L R)
                (conjunto-representante-minimal 
                 R (coleccion-clausulas-falsas-asignacion sigma S)))
           (not (modelo-clausula
                 sigma (conjunto-testigo 
                        R (coleccion-clausulas-falsas-asignacion sigma S) L))))
  :hints (("Goal"
           :use (:instance
                 coleccion-clausulas-falsas-asignacion-conjunto-miembro
                 (C (conjunto-testigo 
                     R (coleccion-clausulas-falsas-asignacion sigma S) L))))))

(defthm modelo-clausula-nueva-asignacion-caso-inductivo-aux-1
  (implies (and (conjunto-unitario (interseccion R D))
                (member-equal (complementario L) D)
                (member-equal (complementario L) R)
                (not (modelo-clausula sigma D))
                (es-literal L)
                (es-clausula C)
                (es-clausula D)
                (member-equal L C)
                (resolucion-cond-clausulas C D)
                (modelo-forma-clausal (nueva-asignacion R sigma)
                                      (resolucion-cond-clausulas C D)))
           (modelo-clausula (nueva-asignacion R sigma) C))
  :hints (("Goal" 
           :in-theory (disable 
                       es-literal
                       literal-resolucion-clausulas-suficiente-clausulas))))

(defthm resolucion-cond-clausulas-existe-aux
  (implies (and (es-literal L)
                (es-clausula C)
                (es-clausula D)
                (member-equal L C)
                (member-equal (complementario L) D)
                (not (equal (literal-resolucion-clausulas C D) L)))
           (tiene-literales-complementarios (resolvente L C D)))
  :rule-classes nil
  :INSTRUCTIONS
  (:BASH (:USE (:INSTANCE TIENE-LITERALES-COMPLEMENTARIOS-MEMBER
                          (L (LITERAL-RESOLUCION-CLAUSULAS C D))
                          (C (UNION-EQUAL (ELIMINA-ELEMENTO L C)
                                          (ELIMINA-ELEMENTO (COMPLEMENTARIO L)
                                                            D)))))
         (:IN-THEORY (DISABLE ES-LITERAL))
         :SPLIT :PROVE :PROVE :PROVE))
  
(defthm resolucion-cond-clausulas-existe
  (implies (and (not (tiene-literales-complementarios (resolvente L C D)))
                (es-literal L)
                (es-clausula C)
                (es-clausula D)
                (member-equal L C)
                (member-equal (complementario L) D)
                (condicion C D))
           (resolucion-cond-clausulas C D))
  :hints (("Goal" :use resolucion-cond-clausulas-existe-aux)))

(defthm tiene-literales-complementarios-union-equal
  (implies (and (tiene-literales-complementarios (union-equal C1 C2))
                (es-clausula C1)
                (es-clausula C2)
                (not (tiene-literales-complementarios C1))
                (not (tiene-literales-complementarios C2)))
           (literal-resolucion-clausulas C1 C2))
  :rule-classes nil)

(defthm tiene-literales-complementarios-elimina-elemento
  (implies (tiene-literales-complementarios (elimina-elemento L C))
           (tiene-literales-complementarios C)))

(defthm tiene-literales-complementarios-resolvente-aux
  (implies (and (es-literal L)
                (es-clausula C)
                (es-clausula D)
                (member-equal L C)
                (member-equal (complementario L) D)
                (tiene-literales-complementarios (resolvente L C D)))
           (or (tiene-literales-complementarios C)
               (tiene-literales-complementarios D)
               (literal-resolucion-clausulas
                (elimina-elemento L C) 
                (elimina-elemento (complementario L) D))))
  :hints (("Goal" :in-theory (enable resolvente)
           :use (:instance tiene-literales-complementarios-union-equal
                    (C1 (elimina-elemento L C))
                    (C2 (elimina-elemento (complementario L)
                                          D)))))
  :rule-classes nil)

(defthm literal-resolucion-clausulas-elimina-elemento-1
  (implies (literal-resolucion-clausulas (elimina-elemento L C) D)
           (member-equal
            (literal-resolucion-clausulas (elimina-elemento L C) D)
            C)))

(defthm literal-resolucion-clausulas-elimina-elemento-complementario
  (implies (literal-resolucion-clausulas 
            C (elimina-elemento (complementario L) D))
           (member-equal (complementario
                          (literal-resolucion-clausulas
                           C (elimina-elemento (complementario L) D)))
                         D))
  :hints (("Goal"
           :use (:instance literal-resolucion-clausulas-propiedades-1
                           (C1 C)
                           (C2 (elimina-elemento (complementario L) D))))))

(defthm literal-resolucion-clausulas-es-literal
  (implies (and (literal-resolucion-clausulas C D)
                (es-clausula C))
           (es-literal (literal-resolucion-clausulas C D))))

(defthm conjunto-unitario-no-contiene-literal
  (implies (and (conjunto-unitario (interseccion R D))
                (member-equal L1 R)
                (member-equal L1 D)
                (member-equal L2 D)
                (not (equal L1 L2)))
           (not (member-equal L2 R)))
  :hints (("Goal" :in-theory (disable conjunto-unitario))))

(defthm literal-resolucion-clausulas-elimina-elemento-2-aux
  (implies (literal-resolucion-clausulas C (elimina-elemento L D))
           (not (equal
                 (complementario
                  (literal-resolucion-clausulas C (elimina-elemento L D)))
                 L)))
  :hints (("Goal"
           :use (:instance literal-resolucion-clausulas-propiedades-1
                           (C1 C)
                           (C2 (elimina-elemento L D))))))

(defthm literal-resolucion-clausulas-elimina-elemento-2
  (implies (and (conjunto-unitario (interseccion R D))
                (member-equal (complementario L) R)
                (member-equal (complementario L) D)
                (literal-resolucion-clausulas
                 C (elimina-elemento (complementario L) D)))
           (not (member-equal (complementario
                               (literal-resolucion-clausulas
                                C (elimina-elemento (complementario L) D)))
                              R)))
  :hints (("Goal" :in-theory (disable conjunto-unitario))))

(defthm not-modelo-clausula-modelo-complementario
  (implies (and (not (modelo-clausula sigma C))
                (member-equal (complementario L) C)
                (es-literal L))
           (modelo sigma L)))

(defthm modelo-clausula-nueva-asignacion-caso-inductivo-aux-2
  (implies (and (conjunto-unitario (interseccion R D))
                (member-equal (complementario L) D)
                (member-equal (complementario L) R)
                (not (modelo-clausula sigma D))
                (es-literal L)
                (es-clausula C)
                (es-clausula D)
                (member-equal L C)
                (condicion C D)
                (not (resolucion-cond-clausulas C D)))
           (modelo-clausula (nueva-asignacion R sigma) C))
  :hints (("Goal"
           :in-theory 
           (disable
            resolvente resolucion-cond-clausulas
            modelo-nueva-asignacion-clausula-literal-complementario-not-member)
           :use (tiene-literales-complementarios-resolvente-aux
                 (:instance
                  modelo-nueva-asignacion-clausula-literal-complementario-not-member
                  (L (literal-resolucion-clausulas 
                      (elimina-elemento L C)
                      (elimina-elemento (complementario L) D))))))))

(defthm modelo-clausula-nueva-asignacion-caso-inductivo-aux
  (implies (and (conjunto-unitario (interseccion R D))
                (member-equal (complementario L) D)
                (member-equal (complementario L) R)
                (not (modelo-clausula sigma D))
                (es-clausula C)
                (es-clausula D)
                (member-equal L C)
                (condicion C D) ;; <- !!
                (modelo-forma-clausal (nueva-asignacion R sigma)
                                      (resolucion-cond-clausulas C D)))
           (modelo-clausula (nueva-asignacion R sigma) C))
  :hints (("Goal" 
           :in-theory (disable es-literal resolucion-cond-clausulas
                               conjunto-unitario)
           :cases ((resolucion-cond-clausulas C D)))))

(in-theory (disable resolvente resolucion-cond-clausulas))

(DEFTHM CONJUNTO-MIEMBRO-FORMA-CLAUSAL-ES-CLAUSULA
  (IMPLIES (AND (CONJUNTO-MIEMBRO C S)
                (ES-FORMA-CLAUSAL S))
           (ES-CLAUSULA C)))

(defthm miembro-reduce-coleccion-literal
  (implies (and (member-equal L C)
                (member-equal (complementario L) D)
                (not (member-equal L D))
                (conjunto-miembro C S)
                (conjunto-miembro D S)
                (condicion C D) ;; <- !!
                (saturado-cond-conjunto S)
                (es-forma-clausal S)
                (resolucion-cond-clausulas C D))
           (conjunto-miembro (car (resolucion-cond-clausulas C D))
                             (reduce-coleccion-literal S L)))
  :INSTRUCTIONS
  (:SPLIT (:REWRITE CARACTERIZACION-PERTENENCIA-REDUCE-COLECCION-LITERAL)
          :SPLIT :BASH
          (:IN-THEORY (ENABLE RESOLUCION-COND-CLAUSULAS))
          :BASH))

(defthm modelo-forma-clausal-conjunto-miembro
  (implies (and (conjunto-miembro c s)
                (modelo-forma-clausal sigma s))
           (modelo-clausula sigma c)))

(defthm modelo-forma-clausal-resolucion-en-forma-clausal-con-modelo
  (implies (and (conjunto-miembro (car (resolucion-cond-clausulas c d)) s)
                (modelo-forma-clausal sigma s))
           (modelo-forma-clausal sigma (resolucion-cond-clausulas c d)))
  :hints (("Goal" :in-theory (enable resolucion-cond-clausulas))))

(defthm nueva-asignacion-clausula-inductivo-2-auxiliar-1
  (implies (and (modelo-forma-clausal sigma (reduce-coleccion-literal S L))
                (member-equal L C)
                (member-equal (complementario L) D)
                (not (member-equal L D))
                (conjunto-miembro C S)
                (conjunto-miembro D S)
                (condicion C D) ;; <- !!
                (es-forma-clausal S)
                (saturado-cond-conjunto S))
           (modelo-forma-clausal sigma (resolucion-cond-clausulas C D)))
  :hints (("Goal" 
           :in-theory (DISABLE 
                       MIEMBRO-REDUCE-COLECCION-LITERAL
                       CARACTERIZACION-PERTENENCIA-REDUCE-COLECCION-LITERAL)
           :cases ((RESOLUCION-COND-CLAUSULAS C D)))
          ("Subgoal 1" :use miembro-reduce-coleccion-literal)))

(defthm literal-cierto-not-member-clausula-falsa
  (implies (and (clausula-con-literal-cierto sigma S)
                (conjunto-miembro 
                 C (coleccion-clausulas-falsas-asignacion sigma S)))
           (not (member-equal (literal-cierto-clausula 
                               sigma (clausula-con-literal-cierto sigma S))
                              C)))
  :hints (("Goal" 
           :use (:instance 
                 modelo-clausula-member
                 (L (literal-cierto-clausula 
                     sigma (clausula-con-literal-cierto sigma S)))))))

(defthm subsetp-equal-conjunto-miembro
  (implies (and (subsetp-equal S1 S2)
                (conjunto-miembro C S1))
           (conjunto-miembro C S2)))

(defthm conjunto-testigo-coleccion-clausulas-falsas-asignacion-conjunto-miembro
  (implies (and (member-equal L R)
                (conjunto-representante-minimal 
                 R (coleccion-clausulas-falsas-asignacion sigma S)))
           (conjunto-miembro
            (conjunto-testigo 
             R (coleccion-clausulas-falsas-asignacion sigma S) L) S))
  :hints (("Goal" 
           :use (:instance
                 subsetp-equal-conjunto-miembro
                 (S1 (coleccion-clausulas-falsas-asignacion sigma S))
                 (S2 S)
                 (C (conjunto-testigo 
                     R (coleccion-clausulas-falsas-asignacion sigma S) L))))))

(defthm conjunto-miembro-coleccion-clausulas-es-clausula
  (implies (and (conjunto-miembro
                 C (coleccion-clausulas-falsas-asignacion sigma S))
                (es-forma-clausal S))
           (es-clausula C)))

(defthm member-conjunto-testigo-coleccion-clausulas-falsas-asignacion
  (implies (conjunto-testigo 
            R (coleccion-clausulas-falsas-asignacion sigma S) C)
           (member-equal (conjunto-testigo
                          R (coleccion-clausulas-falsas-asignacion sigma S) C)
                         S)))

(defthm nueva-asignacion-clausula-inductivo-2
  (implies 
   (and (clausula-con-literal-cierto (asignacion) S)
        (modelo-forma-clausal
         (nueva-asignacion R (asignacion))
         (reduce-coleccion-literal
          S (literal-cierto-clausula
             (asignacion) (clausula-con-literal-cierto (asignacion) S))))
        (es-forma-clausal S)
        (saturado-cond-conjunto S)
        (not (contiene-clausula-vacia S))
        (conjunto-representante-minimal
         R (coleccion-clausulas-falsas-asignacion (asignacion) S))
        (member-equal C S)
        (member-equal
         (literal-cierto-clausula 
          (asignacion) (clausula-con-literal-cierto (asignacion) S)) C)
        (member-equal
         (complementario
          (literal-cierto-clausula
           (asignacion) (clausula-con-literal-cierto (asignacion) S))) R))
   (modelo-clausula (nueva-asignacion R (asignacion)) C))
  :hints (("Goal"
           :IN-THEORY (DISABLE ES-LITERAL)
           :use (
                 (:INSTANCE
                  MODELO-CLAUSULA-NUEVA-ASIGNACION-CASO-INDUCTIVO-AUX
                  (sigma (asignacion))
                  (L (LITERAL-CIERTO-CLAUSULA
                      (ASIGNACION)
                      (CLAUSULA-CON-LITERAL-CIERTO (ASIGNACION) S)))
                  (D (CONJUNTO-TESTIGO
                      R (COLECCION-CLAUSULAS-FALSAS-ASIGNACION (ASIGNACION) S)
                      (COMPLEMENTARIO
                       (LITERAL-CIERTO-CLAUSULA 
                        (ASIGNACION)
                        (CLAUSULA-CON-LITERAL-CIERTO (ASIGNACION) S))))))
                 (:INSTANCE
                  conjunto-miembro-coleccion-clausulas-es-clausula
                  (sigma (asignacion))
                  (C (CONJUNTO-TESTIGO
                      R (COLECCION-CLAUSULAS-FALSAS-ASIGNACION (ASIGNACION) S)
                      (COMPLEMENTARIO
                       (LITERAL-CIERTO-CLAUSULA 
                        (ASIGNACION)
                        (CLAUSULA-CON-LITERAL-CIERTO (ASIGNACION) S))))))
                 (:INSTANCE
                  PROPIEDAD-FUNDAMENTAL-CONDICION
                  (D (CONJUNTO-TESTIGO
                      R (COLECCION-CLAUSULAS-FALSAS-ASIGNACION (ASIGNACION) S)
                      (COMPLEMENTARIO
                       (LITERAL-CIERTO-CLAUSULA
                        (ASIGNACION)
                        (CLAUSULA-CON-LITERAL-CIERTO (ASIGNACION) S))))))
                 ))))

(defthm nueva-asignacion-clausula-inductivo
  (implies 
   (and (clausula-con-literal-cierto (asignacion) S)
        (modelo-forma-clausal
         (nueva-asignacion R (asignacion))
         (reduce-coleccion-literal
          S (literal-cierto-clausula
             (asignacion) (clausula-con-literal-cierto (asignacion) S))))
        (es-forma-clausal S)
        (saturado-cond-conjunto S)
        (not (contiene-clausula-vacia S))
        (conjunto-representante-minimal
         R (coleccion-clausulas-falsas-asignacion (asignacion) S))
        (member-equal C S))
   (modelo-clausula (nueva-asignacion R (asignacion)) C))
  :hints (("Goal" :cases
           ((member-equal 
             (literal-cierto-clausula 
              (asignacion) (clausula-con-literal-cierto (asignacion) S)) C)))
          ("Subgoal 1" :cases
           ((member-equal 
             (complementario
              (literal-cierto-clausula
               (asignacion) (clausula-con-literal-cierto (asignacion) S))) 
             R))))
  )

(defthm nueva-asignacion-forma-clausal-inductivo
  (implies 
   (and (clausula-con-literal-cierto (asignacion) S)
        (modelo-forma-clausal
         (nueva-asignacion R (asignacion))
         (reduce-coleccion-literal
          S (literal-cierto-clausula
             (asignacion) (clausula-con-literal-cierto (asignacion) S))))
        (es-forma-clausal S)
        (saturado-cond-conjunto S)
        (not (contiene-clausula-vacia S))
        (conjunto-representante-minimal
         R (coleccion-clausulas-falsas-asignacion (asignacion) S)))
   (modelo-forma-clausal (nueva-asignacion R (asignacion)) S))
  :hints ((extension-conjuntiva-a-listas
           S
           (lambda (elt)
             (modelo-clausula (nueva-asignacion R (asignacion)) elt))
           (lambda (lst)
             (modelo-forma-clausal (nueva-asignacion R (asignacion)) lst)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.6) Teorema de completitud de Bezem
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-resolucion-bezem
  (implies (and (es-forma-clausal S)
                (saturado-cond-conjunto S)
                (not (contiene-clausula-vacia S))
                (conjunto-representante-minimal 
                 R (coleccion-clausulas-falsas-asignacion (asignacion) S)))
           (modelo-forma-clausal (nueva-asignacion R (asignacion)) S))
  :hints (("Goal" 
           :induct (induccion-literal-cierto (asignacion) S))))

;;;============================================================================